/// \file
/// \author Matthew Allen
#ifndef __DRAG_AND_DROP
#define __DRAG_AND_DROP

#include "GVariant.h"

#if defined WIN32

#include <shlobj.h>
#include "GCom.h"

#else

#define S_OK								0
#define DROPEFFECT_NONE						0x0
#define DROPEFFECT_COPY						0x1
#define DROPEFFECT_MOVE						0x2
#define DROPEFFECT_LINK						0x4

#endif

// int <-> char* format conversions
#if defined WIN32

/// Convert a string d'n'd format to an OS dependant integer.
LgiFunc int FormatToInt(char *s);
/// Convert a Os dependant integer d'n'd format to a string.
LgiFunc char *FormatToStr(int f);

#elif defined LINUX

/// Convert a string d'n'd format to an OS dependant integer.
#define FormatToInt(c)		XInternAtom(XDisplay(), c, false)
/// Convert a Os dependant integer d'n'd format to a string.
#define FormatToStr(f)		XGetAtomName(XDisplay(), f)

#if !defined(XWIN) && !defined(YWIN)
	#error Either XWIN or YWIN must be defined.
#endif

#endif

//////////////////////////////////////////////////////////////////////////////
class LgiClass GDragDropSource
#if defined WIN32
	: public IDropSource, public IEnumFORMATETC
#elif defined XWIN
	: public QObject
#endif
{
	friend class GDataObject;

protected:
	char *CurrentFormat;
	
	#if defined WIN32
	int Index;

	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);

	// IEnumFORMATETC
	HRESULT STDMETHODCALLTYPE Next(ULONG celt, FORMATETC *rgelt, ULONG *pceltFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG celt);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumFORMATETC **ppenum);

	// IDropSource
	HRESULT STDMETHODCALLTYPE QueryContinueDrag(BOOL fEscapePressed, DWORD InputState);
	HRESULT STDMETHODCALLTYPE GiveFeedback(DWORD dwEffect);
	#endif

	/// Creates a file drop
	bool CreateFileDrop(GVariant *Var, GMouse &m, List<char> &Files);

	#ifdef XWIN
	void OnSelectionRequest(Display *Dsp, XEvent &Event, QWidget *W, Atom XdndSelection); 
	#endif
	
public:
	GDragDropSource();
	~GDragDropSource();

	/// Start a drag operation
	int Drag(GView *SourceWnd, int Effect);

	/// Called when window is registered
	virtual void OnRegister(bool Success) {}
	/// Called on start of drag op
	virtual void OnStartData() {}
	/// Called on end of drag op
	virtual void OnEndData() {}
	
	/// This is called when you are being asked for your
	/// data. You must provide it in the format specified
	/// and return it via the GVariant.
	virtual bool GetData(GVariant *Data,		// out
						 char *Format			// in
						 ) { return false; }

	/// This is called to see what formats your support
	/// Insert into the list dynamically allocated strings
	/// that describe the formats you can supply data in.
	/// One of these will be chosen by the target and passed
	/// to 'GetData'
	///
	/// You can define your own formats by making up a string
	/// or you can use OS defined ones to interact with other
	/// apps in the system. e.g. Win32 defines "CF_HDROP" for file
	/// d'n'd. You can use 'FormatToInt' and 'FormatToStr' to
	/// convert to and from the OS's integer id's. This works
	/// the same on Win32 as Linux. But the system formats will
	/// obviously be different.
	///
	/// The default file drop format for each OS is defined as
	/// #LGI_FileDropFormat.
	virtual bool GetFormats(List<char> &Formats) { return false; }
};

class LgiClass GDragDropTarget
#if defined WIN32
	: public IDropTarget
#elif defined XWIN
	: public QObject
#endif
{
private:
	GView *To;
	uchar *DragDropData;
	int DragDropLength;
	List<char> Formats;

	#if defined WIN32
	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);

	// IDropTarget
	HRESULT STDMETHODCALLTYPE DragEnter(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragLeave(void);
	HRESULT STDMETHODCALLTYPE Drop(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);

	#endif

protected:
	#if defined WIN32
	// Data elements
	IDataObject *DataObject;

	// Tools
	bool OnDropFileGroupDescriptor(FILEGROUPDESCRIPTOR *Data, List<char> &Files);
	#endif

public:
	GDragDropTarget();
	~GDragDropTarget();

	// call this when you have a HWND
	void SetWindow(GView *To);

	// Override these

	// Events
	virtual void OnDragInit(bool Success) {}
	virtual void OnDragEnter() {}
	virtual void OnDragExit() {}

	// 'WillAccept' is called to see whether this target
	// can cope with any of the data types being passed to it.
	// Once you have decided what format you want the data in
	// clear all the other formats from the list. The first
	// format left in the list will be passed to the 'OnDrop'
	// function.
	virtual int WillAccept(List<char> &Formats, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }

	// 'OnDrop' is called when the user releases the data over
	// your window. The data is going to be a binary GVariant
	// in the format you accepted earlier.
	virtual int OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }
};

#endif
