/**
	\file
	\author Matthew Allen
	\brief Variant class.\n
	Copyright (C), <a href="mailto:fret@memecode.com">Matthew Allen</a>
 */

#ifndef __GVARIANT_H__
#define __GVARIANT_H__

#include "GDom.h"
#undef Bool
#include "GDateTime.h"
#include "GContainers.h"

/// The different types the varient can be 
enum GVariantType
{
	// Main types
	
	/// Null type
	GV_NULL,
	/// 32-bit integer
	GV_INT32,
	/// 64-bit integer
	GV_INT64,
	/// true or false boolean.
	GV_BOOL,
	/// C++ double
	GV_DOUBLE,
	/// Null terminated string value
	GV_STRING,
	/// Block of binary data
	GV_BINARY,
	/// List of GVariant
	GV_LIST,
	/// GDom object
	GV_DOM,
	/// Untyped pointer
	GV_VOID_PTR,
	/// GDateTime class.
	GV_DATETIME,

	// Types of pointer (stored as void*)
	
	/// Pointer to GView (stored as a void*)
	GV_GVIEW,
	/// Pointer to GMouse (stored as a void*)
	GV_GMOUSE,
	/// Pointer to GKey (stored as a void*)
	GV_GKEY
};

/// A class that can be different types
class LgiClass GVariant
{
public:
	/// The type of the variant
    GVariantType Type;
	/// The indentifier
    short Id;
	/// A pointer to a bool to be set when the variant changes
    bool *Dirty;

    /// The value of the variant
	union
    {
		/// Valid when Type == GV_INT32
	    int Int;
		/// Valid when Type == GV_BOOL
	    bool Bool;
		/// Valid when Type == GV_INT64
	    int64 Int64;
		/// Valid when Type == GV_DOUBLE
	    double Dbl;
	    /// Valid when Type == GV_STRING
		char *String;
		/// Valid when Type == GV_DOM
		GDom *Dom;
		/// Valid when Type == GV_VOID_PTR
		void *Ptr;
		/// Valid when Type == GV_BINARY
	    struct _Binary
	    {
		    int Length;
		    void *Data;
	    } Binary;
		/// Valid when Type == GV_LIST
	    List<GVariant> *Lst;
		/// Valid when Type == GV_DATETIME
		GDateTime *Date;
    } Value;

    /// Constructor to null
    GVariant();
    /// Constructor for int
    GVariant(int i);
    /// Constructor for int
    GVariant(int64 i);
    /// Constructor for double
    GVariant(double i);
    /// Constructor for string
    GVariant(char *s);
    /// Constructor for ptr
    GVariant(void *p);
    /// Constructor for DOM ptr
    GVariant(GDom *p);
    /// Constructor for date
    GVariant(GDateTime *d);
    ~GVariant();

    /// Assign int value
    GVariant &operator =(int i);
    /// Assign bool value
    GVariant &operator =(bool i);
	/// Assign int value
    GVariant &operator =(int64 i);
	/// Assign double value
    GVariant &operator =(double i);
	/// Assign string value (makes a copy)
    GVariant &operator =(char *s);
	/// Assign another variant value
    GVariant &operator =(GVariant &i);
	/// Assign value to a void ptr
    GVariant &operator =(void *p);
	/// Assign value to DOM ptr
    GVariant &operator =(GDom *p);
	/// Assign value to be a date/time
    GVariant &operator =(GDateTime *d);

    GVariant &operator =(class GView *p);
    GVariant &operator =(class GMouse *p);
    GVariant &operator =(class GKey *k);

	/// Sets the value to block of binary data
    bool SetBinary(int Len, void *Data);
	/// Sets the value to a list of variants
    bool SetList(List<GVariant> *Lst = 0);

    /// Returns the string if valid
    char *Str();
	/// Sets the variant to NULL
    void Empty();

    /// Get the items in the list or 0 if not a list
    int GetItems();
    bool HasItem(GVariant *s);
    GVariant *ItemAt(int i);
    GVariant *First();
    GVariant *Next();
    GVariant *Prev();
    GVariant *Last();
    bool Insert(GVariant *s, int Where = -1);
    bool Delete(GVariant *s);

    /// True if currently a int
    bool IsInt();
    /// True if currently a bool
    bool IsBool();
    /// True if currently a double
    bool IsDouble();
    /// True if currently a string
    bool IsString();
    /// True if currently a binary block
    bool IsBinary();
    /// True if currently null
    bool IsNull();
    
    /// Casts the value to int, from whatever source type
    int32 CastInt32();
	/// Cast to a string from whatever source type
    char *CastString();
};

#endif
