/// \file
/// \author Matthew Allen <fret@memecode.com>
#ifndef _GXMLTREE_H_
#define _GXMLTREE_H_

/// Runtime option: Don't encode entities
#define GXT_NO_ENTITIES						0x0001
/// Runtime option: Use minimal whitespace
#define GXT_NO_PRETTY_WHITESPACE			0x0002

/// Xml attribute, a named value.
class LgiClass GXmlAttr
{
public:
	/// The name of the attribute
	char *Name;
	/// The value of the attribute
	char *Value;
	
	GXmlAttr()
	{
		Name = 0;
		Value = 0;
	}
	
	GXmlAttr(char *n, char *v)
	{
		Name = NewStr(n);
		Value = NewStr(v);
	}
	
	~GXmlAttr()
	{
		DeleteArray(Name);
		DeleteArray(Value);
	}
};

/// An XML element or tag. Contains optionally sub tags and a list of attributes.
/// C++ applications can inherit from this and have native member types mapped to
/// attributes using the SerializeAttr methods. All you have to do is override the
/// virtual member function Serialize and call SerializeAttr on each of your native
/// member variables. When loading and saving the attributes will be mapped to and
/// from your native types.
class LgiClass GXmlTag : public GDom
{
	friend class GXmlTree;

protected:
	bool Write;
	GXmlAttr *_Attr(char *Name, bool Write);
	bool GetVariant(char *Name, GVariant &Value, char *Array);
	bool SetVariant(char *Name, GVariant &Value, char *Array);

public:
	/// The name of the tag/element. This can be NULL in the case
	/// that the element is purely content.
	char *Tag;
	/// Any content following the tag.
	char *Content;
	/// The parent element/tag.
	GXmlTag *Parent;
	/// A list of attributes that this tag has
	List<GXmlAttr> Attr;
	/// A list of child tags. Don't edit this list yourself, use the
	/// InsertTag and RemoveTag methods.
	List<GXmlTag> Children;
	
	/// Construct the object
	GXmlTag
	(
		/// [Optional] Start with this name
		char *tag = 0
	);	
	virtual ~GXmlTag();

	/// For debugging.
	bool Dump(int Depth = 0);

	/// Get the string value of a named attribute
	char *GetAttr(char *Name);
	/// Get the value of a named attribute as an int
	int GetAsInt(char *Name);
	/// Set the value of a named attribute to a string
	bool SetAttr(char *Name, char *Value);
	/// Set the value of a named attribute to an int
	bool SetAttr(char *Name, int Value);

	/// Read/write a native C integer into an attribute
	bool SerializeAttr(char *Attr, int &Int);	
	/// Read/write a native C dynamically allocated string into an attribute
	bool SerializeAttr(char *Attr, char *&Str);
	/// Read/write a native C double into an attribute
	bool SerializeAttr(char *Attr, double &Dbl);
		
	/// Read/write all your native types in here
	virtual bool Serialize() { return false; }
	/// Returns a pointer to a child tag if present, or NULL if not.
	GXmlTag *GetTag(char *Name, bool Create = false);
	/// Creates a sub tag if it doesn't already exist.
	GXmlTag *CreateTag(char *Name) { return GetTag(Name, true); }
	/// Inserts a child tag.
	virtual void InsertTag(GXmlTag *t);
	/// Removes this tag from the DOM heirarchy.
	virtual void RemoveTag();
};

/// In the case your inheriting objects from GXmlTag you need to instantiate your
/// types as the XML is loaded and elements are encountered. Pass this class to the
/// GXmlTree::Read to get create events during the read process, create an object of
/// the right type and pass it back.
class GXmlFactory
{
public:
	/// Creates an object of type 'Tag' and returns it.
	virtual GXmlTag *Create(char *Tag) = 0;
};

/// The reader/writer for a tree of GXmlTag objects.
class LgiClass GXmlTree
{
	class GXmlTreePrivate *d;
	
	GXmlTag *Parse(GXmlTag *Tag, char *&t, bool &NoChildren, bool InTypeDef);
	void Output(GXmlTag *t, int Depth);

public:
	/// Constructor
	GXmlTree
	(
		/// Use #GXT_NO_ENTITIES and/or #GXT_NO_PRETTY_WHITESPACE
		int Flags = 0
	);
	~GXmlTree();
	
	/// Read an XML file into a DOM tree of GXmlTag objects from a stream.
	bool Read
	(
		/// The root tag to create children from.
		GXmlTag *Root,
		/// The stream to read from.
		GStream *File,
		/// [Optional] The factory to create GXmlTag type objects. If not specified
		/// vanilla GXmlTag objects will be created.
		GXmlFactory *Factory
	);
	/// Write an XML file from a DOM tree of GXmlTag objects into a stream.
	bool Write
	(
		/// The DOM tree.
		GXmlTag *Root,
		/// The output stream.
		GStream *File
	);
	/// Gets the last error message.
	char *GetErrorMsg();
};

#endif
