/**
	\file
	\author Matthew Allen
	\date 28/5/1998
	\brief Network sockets classes
	Copyright (C) 1998, <a href="mailto:fret@memecode.com">Matthew Allen</a>
*/

#ifndef __INET_H
#define __INET_H

#include "GMem.h"
#include "GContainers.h"
#include "LgiNetInc.h"
#include "GStream.h"

#if defined WIN32

#elif defined BEOS

	#include <socket.h>
	#include <netdb.h>

#elif defined ATHEOS

	#include <atheos/socket.h>

#else

typedef int SOCKET;

#endif

#define DEFAULT_TIMEOUT			30

// Standard ports
#define SMTP_PORT				25
#define SMTP_SSL_PORT			465
#define POP3_PORT				110
#define POP3_SSL_PORT			995
#define IMAP_PORT				143
#define IMAP_SSL_PORT			993

// Parameters for passing to GSocket::SetParameter

/// Turn on/off logging. Used with GSocketImpl::SetParameter.
#define GSocket_Log				1
/// Set the progress object. Used with GSocketImpl::SetParameter. Value = (GProgress*)Prog
#define GSocket_Progress		2
/// Set the size of the transfer. Used with GSocketImpl::SetParameter. Value = (int)Size
#define GSocket_TransferSize	3
/// Set the type of protocol. Used with GSocketImpl::SetParameter. Value = (char*)Protocol
#define GSocket_Protocol		4

#if defined BEOS

	#include "NetworkKit.h"
	#define ValidSocket(s)			((s)>=0)
	#define INVALID_SOCKET			-1
	typedef int OsSocket;

#elif defined ATHEOS

	#define ValidSocket(s)			((s)>=0)
	#define INVALID_SOCKET			-1
	typedef int OsSocket;

#elif defined WIN32

	#include <winsock2.h>
	#define ValidSocket(s)			((s) != INVALID_SOCKET)
	typedef SOCKET OsSocket;

#elif defined LINUX

	#define ValidSocket(s)			((s)>=0)
	#define INVALID_SOCKET			-1
	typedef int OsSocket;

#endif

// Logging types..
enum
{
	NET_LOG_NONE = 0,
	NET_LOG_HEX_DUMP = 1,
	NET_LOG_ALL_BYTES = 2
};

// Functions
LgiNetFunc bool HaveNetConnection();
LgiNetFunc bool WhatsMyIp(char *IpAddr);
LgiNetFunc void MDStringToDigest(unsigned char digest[16], char *Str);

// Classes
class GNetwork;
class GSocket;

/// Singleton class to initialize access to the network.
class LgiNetClass GNetwork
{
	bool SocketsOpen;

public:
	GNetwork();
	virtual ~GNetwork();

	/// Returns true if the network is ready for use
	bool IsValid() { return SocketsOpen; }

	/// Enumerates the current interfaces
	bool EnumInterfaces(List<char> &Lst);
};

/// Socket state type
enum GSocketProp
{
	/// The socket is readable
	GReadable,
	/// The socket is writable
	GWriteable,
	/// The socket can accept a connection
	GAcceptable,
	/// The socket is in async (non-blocking) mode
	GAsync,
	GException,
	GNoDelay,
};

/// Virtual base class for a socket. See the documentation for GSocketImpl for a more
/// through treatment of this object's API.
class LgiNetClass GSocket : public GStream
{
public:
	virtual ~GSocket() {}

	/// Returns the actual socket (as defined by the OS)
	virtual OsSocket Handle(OsSocket Set = INVALID_SOCKET) = 0;
	/// Returns the IP at this end of the socket
	virtual bool WhatsMyIp
	(
		/// Ptr to a buffer of at least 16 bytes
		char *IpAddr
	) = 0;
	/// Return the port at this end of the connection
	virtual int WhatsMyPort() = 0;
	/// Gets the remote IP
	virtual bool GetRemoteIp(char *IpAddr) = 0;
	/// True if there is data available to read.
	virtual bool IsReadable() = 0;
	/// True if the current action has been cancelled
	virtual bool Cancel() = 0;
	/// Cancel the current action (might not be implemented)
	virtual void Cancel(bool i) = 0;
	/// Returns some state information
	virtual int Is
	(
		/// The type of state to return.
		GSocketProp p
	) = 0;
	/// Sets some state in the socket, i.e. GAsync
	virtual void Is(GSocketProp p, int i) = 0;
	/// Output the data to a file
	virtual void SetLogFile(char *FileName = 0, int Type = NET_LOG_HEX_DUMP) = 0;

	/// Listens on a given port for an incomming connection.
	virtual bool Listen(int Port = 0) = 0;
	/// Accepts an incomming connection and connects the socket you pass in to the remote host.
	virtual bool Accept(GSocket *c) = 0;

	/// Process an error
	virtual int Error() = 0;

	/// Called when the connection is dropped
	virtual void OnDisconnect() = 0;
	/// Called when data is read
	virtual void OnRead(char *Data, int Len) = 0;
	/// Called when data is written
	virtual void OnWrite(char *Data, int Len) = 0;
	/// Called when an error occurs
	virtual void OnError(int ErrorCode, char *ErrorDescription) = 0;
	/// Called when some events happens
	virtual void OnInformation(char *Str) = 0;
	/// Call to change a parameter (e.g. #GSocket_Log)
	virtual int SetParameter(int Param, int Value) = 0;
};

/// Implementation of a network socket
class LgiNetClass GSocketImpl : public GSocket
{
	friend class GNetwork;

protected:
	class GSocketImplPrivate *d;

	// Methods
	void Log(char *Msg, int Ret, char *Buf, int Len);

public:
	int			BytesWritten;
	int			BytesRead;

	/// Creates the class
	GSocketImpl();
	
	/// Destroys the class
	~GSocketImpl();

	/// Returns the operating system handle to the socket.
	OsSocket Handle(OsSocket Set = INVALID_SOCKET);

	/// Returns true if the internal state of the class is ok
	bool IsOK();
	
	/// Sets logging on the socket
	void SetLogFile
	(
		/// The file to log to
		char *FileName = 0,
		/// The type of logging to do
		int Type = NET_LOG_HEX_DUMP
	);		

	/// Returns the IP address at this end of the socket.
	bool WhatsMyIp(char *IpAddr);
	
	/// Returns the port at this end of the socket.
	int WhatsMyPort();
	
	/// Gets the IP address at the remote end of the socket.
	bool GetRemoteIp(char *IpAddr);
	
	/// Returns whether there is data available for reading.
	bool IsReadable();

	/// Allows you to query the socket for various things.
	int Is(GSocketProp p);

	/// Sets an option on the socket.
	void Is(GSocketProp p, int i);

	/// Returns true if the operation was cancelled
	bool Cancel();
	
	/// Cancel the current operation
	void Cancel(bool i);

	/// Opens a connection.
	int Open
	(
		/// The name of the remote host.
		char *HostAddr,
		/// The port on the remote host.
		int Port
	);
	
	/// Returns true if the socket is connected.
	bool IsOpen();
	
	/// Closes the connection to the remote host.
	int Close();

	/// Listens on a given port for an incomming connection.
	bool Listen(int Port = 0);
	
	/// Accepts an incomming connection and connects the socket you pass in to the remote host.
	bool Accept
	(
		/// The socket to handle the connection.
		GSocket *c
	);

	/// \brief Sends data to the remote host.
	/// \return the number of bytes written or <= 0 on error.
	int Write
	(
		/// Pointer to the data to write
		void *Data,
		/// Numbers of bytes to write
		int Len,
		/// Flags to pass to send
		int Flags
	);
	
	/// \brief Reads data from the remote host.
	/// \return the number of bytes read or <= 0 on error.
	///
	/// Generally the number of bytes returned is less than the buffer size. Depending on how much data
	/// you are expecting you will need to keep reading until you get and end of field marker or the number
	/// of bytes your looking for.
	int Read
	(
		/// Pointer to the buffer to write output to
		void *Data,
		/// The length of the receive buffer.
		int Len,
		/// The flags to pass to recv
		int Flags
	);
	
	/// Returns the last error or 0.
	int Error();

	/// Not supported
	int64 GetSize() { return -1; }
	/// Not supported
	int64 SetSize(int64 Size) { return -1; }
	/// Not supported
	int64 GetPos() { return -1; }
	/// Not supported
	int64 SetPos(int64 Pos) { return -1; }

	/// Gets called when the connection is disconnected
	void OnDisconnect();
	
	/// Gets called when data is received.
	void OnRead(char *Data, int Len) {}

	/// Gets called when data is sent.
	void OnWrite(char *Data, int Len) {}
	
	/// Gets called when an error occurs.
	void OnError(int ErrorCode, char *ErrorDescription) {}
	
	/// Gets called when some information is available.
	void OnInformation(char *Str) {}
	
	/// Parameter change handler.
	int SetParameter
	(
		/// e.g. #GSocket_Log
		int Param,
		int Value
	) { return false; }
};

class LgiNetClass GSocks5Socket : public GSocketImpl
{
	char *Proxy;
	int Port;
	char *UserName;
	char *Password;

protected:
	bool Socks5Connected;

public:
	GSocks5Socket(char *proxy, int port, char *username, char *password);
	~GSocks5Socket();

	// Connection
	int Open(char *HostAddr, int port);

	// Server
	bool Listen(int Port) { return false; }
};

/// This is a handy helper tool for parsing text in the form server[:port] into a easy to use form.
class LgiNetClass GHost
{
	int DefaultPort;

public:
	/// The host part
	char *Server;
	/// The port
	int Port;

	/// Construct the class
	GHost
	(
		/// The default port to use if the input has no explicit port
		int DefaultPort,
		/// An optional initial string to parse
		char *Str = 0
	);
	
	~GHost();

	void Empty();
	GHost &operator =(char *s);
	operator int()
	{
		return Server != 0;
	}
};

#endif

