/**
	\file
	\author Matthew Allen
	\date 24/9/1999
	\brief Defines and types
	Copyright (C) 1999-2004, <a href="mailto:fret@memecode.com">Matthew Allen</a>
*/

#ifndef _LGIDEFS_H_
#define _LGIDEFS_H_

// Unsafe typedefs, for backward compatibility
typedef		unsigned char				uchar;
typedef		unsigned short				ushort;
typedef		unsigned int				uint;
typedef		unsigned long				ulong;

// Length safe typedesf, use these in new code

#ifndef BEOS
/// 8-bit signed int type (size safe, garenteed to be 8 bits)
typedef		char						int8;
/// 8-bit unsigned int type (size safe, garenteed to be 8 bits)
typedef		unsigned char				uint8;
#endif
/// 16-bit signed int type (size safe, garenteed to be 16 bits)
typedef		short						int16;
/// 16-bit unsigned int type (size safe, garenteed to be 16 bits)
typedef		unsigned short				uint16;
#ifndef BEOS
/// 32-bit signed int type (size safe, garenteed to be 32 bits)
typedef		int							int32;
/// 32-bit unsigned int type (size safe, garenteed to be 32 bits)
typedef		unsigned int				uint32;
#endif

#ifdef WIN32
/// 64-bit signed int type (size safe, garenteed to be 64 bits)
typedef		signed __int64				int64;
/// 64-bit unsigned int type (size safe, garenteed to be 64 bits)
typedef		unsigned __int64			uint64;
#else
/// 64-bit signed int type (size safe, garenteed to be 64 bits)
typedef		signed long long			int64;
/// 64-bit unsigned int type (size safe, garenteed to be 64 bits)
typedef		unsigned long long			uint64;
#endif

#ifndef WIN32
/// \brief Wide unicode char
///
/// This is 16 bits on Win32 and 32 bits on unix platforms. There are a number
/// of wide character string function available for manipulating wide char strings.
///
/// Firstly to convert to and from utf-8 there is:
/// <ul>
/// 	<li> LgiNewUtf8To16()
///		<li> LgiNewUtf16To8()
/// </ul>
///
/// Wide versions of standard library functions are available:
/// <ul>
///		<li> StrchrW()
///		<li> StrrchrW()
///		<li> StrnchrW()
///		<li> StrstrW()
///		<li> StristrW()
///		<li> StrnstrW()
///		<li> StrnistrW()
///		<li> StrcmpW()
///		<li> StricmpW()
///		<li> StrncmpW()
///		<li> StrnicmpW()
///		<li> StrcpyW()
///		<li> StrncpyW()
///		<li> StrlenW()
///		<li> StrcatW()
///		<li> HtoiW()
///		<li> NewStrW()
///		<li> TrimStrW()
///		<li> ValidStrW()
///		<li> MatchStrW()
/// </ul>
typedef		unsigned int				char16;
#else
typedef		unsigned short				char16;
#endif

// Basic macros
// #define abs(a)						(((a) > 0) ? (a) : -(a))
#define min(a,b)						(((a) < (b)) ? (a) : (b))
#define max(a,b)						(((a) > (b)) ? (a) : (b))
#define limit(i,l,u)					(((i)<(l)) ? (l) : (((i)>(u)) ? (u) : (i)))
#define makelong(a, b)					((a)<<16 | (b&0xFFFF))
#define loword(a)						(a&0xFFFF)
#define hiword(a)						(a>>16)
#define LgiSwap(a, b)					{ int n = a; a = b; b = n; }

// Operators
#define NOT								!
#define AND								&&
#define OR								||

// Good ol NULLy
#ifndef NULL
#define NULL							0
#endif

// Some objectish ones
#define ZeroObj(obj)					memset(&obj, 0, sizeof(obj))
#define CountOf(array)					(sizeof(array)/sizeof(array[0]))
#define StrEmpty(s)						(s)[0] = 0
#define IsStrEmpty(s)					((NOT s) OR ((s)[0] == 0))

#ifndef MEMORY_DEBUG

	#define NEW(obj)					new obj
	#define DeleteObj(obj)				if (obj) { delete obj; obj = 0; }
	#define DeleteArray(obj)			if (obj) { delete [] obj; obj = 0; }

#endif

// Asserts
#ifdef WIN32

	#include "assert.h"
	#define LgiAssert(a)				assert(a)

#else

	extern void							_lgi_assert(bool b, char *test, char *file, int line);
	#define LgiAssert(b)				_lgi_assert(b, #b, __FILE__, __LINE__)
	
#endif

// Flags
#define SetFlag(i, f)					(i) |= (f)
#define ClearFlag(i, f)					(i) &= ~(f)
#define TestFlag(i, f)					(((i) & (f)) != 0)

// Defines

/// \brief Unknown OS
/// \sa LgiGetOs
#define LGI_OS_UNKNOWN					0
/// \brief Windows 95, 98[se] or ME. (95 isnt't really supported but mostly works anyway)
/// \sa LgiGetOs
#define LGI_OS_WIN9X					1
/// \brief Windows NT, 2k, XP or later. (Supported)
/// \sa LgiGetOs
#define LGI_OS_WINNT					2
/// \brief BeOS r5. (Used to be supported)
/// \sa LgiGetOs
#define LGI_OS_BEOS						3
/// \brief Linux. (Kernels v2.4 and up supported)
/// \sa LgiGetOs
#define LGI_OS_LINUX					4
/// \brief Atheos. (Not supported)
/// \sa LgiGetOs
#define LGI_OS_ATHEOS					5
/// \brief Mac OS-9. (Not supported)
/// \sa LgiGetOs
#define LGI_OS_MAC_OS_9					6
/// \brief Mac OS-X. (Don't do it! Think of Be, Inc!)
/// \sa LgiGetOs
#define LGI_OS_MAC_OS_X					7
/// One higher than the maximum OS define
#define LGI_OS_MAX						8

// System Colours

/// Black
#define LC_BLACK						LgiColour(0)
/// Dark grey
#define LC_DKGREY						LgiColour(1)
/// Medium grey
#define LC_MIDGREY						LgiColour(2)
/// Light grey
#define LC_LTGREY						LgiColour(3)
/// White
#define LC_WHITE						LgiColour(4)

/// 3d dark shadow
#define LC_SHADOW						LgiColour(5)
/// 3d light shadow
#define LC_LOW							LgiColour(6)
/// Flat colour for dialogs, windows and buttons
#define LC_MED							LgiColour(7)
/// 3d dark hilight
#define LC_HIGH							LgiColour(8)
/// 3d light hilight
#define LC_LIGHT						LgiColour(9)

/// Dialog colour
#define LC_DIALOG						LgiColour(10)
/// Workspace area
#define LC_WORKSPACE					LgiColour(11)
/// Default text colour
#define LC_TEXT							LgiColour(12)
/// Selection colour
#define LC_SELECTION					LgiColour(13)
/// Selected text colour
#define LC_SEL_TEXT						LgiColour(14)

#define LC_ACTIVE_TITLE					LgiColour(15)
#define LC_ACTIVE_TITLE_TEXT			LgiColour(16)
#define LC_INACTIVE_TITLE				LgiColour(17)
#define LC_INACTIVE_TITLE_TEXT			LgiColour(18)

#define LC_MENU_BACKGROUND				LgiColour(19)
#define LC_MENU_TEXT					LgiColour(20)

#define LC_MAXIMUM						21

// Cursors

/// Normal arrow
#define LCUR_Normal						0
/// Upwards arrow
#define LCUR_UpArrow					1
/// Crosshair
#define LCUR_Cross						2
/// Hourglass/watch
#define LCUR_Wait						3
/// Ibeam/text entry
#define LCUR_Ibeam						4
/// Vertical resize (|)
#define LCUR_SizeVer					5
/// Horizontal resize (-)
#define LCUR_SizeHor					6
/// Diagonal resize (/)
#define LCUR_SizeBDiag					7
/// Diagonal resize (\)
#define LCUR_SizeFDiag					8
/// All directions resize
#define LCUR_SizeAll					9
/// Blank/invisible cursor (don't use!)
#define LCUR_Blank						10
/// Vertical splitting
#define LCUR_SplitV						11
/// Horziontal splitting
#define LCUR_SplitH						12
/// A pointing hand
#define LCUR_PointingHand				13
/// A slashed circle
#define LCUR_Forbidden					14
/// Copy Drop
#define LCUR_DropCopy					15
/// Copy Move
#define LCUR_DropMove					16

// General Event Flags
#define LGI_EF_LCTRL					0x00000001
#define LGI_EF_RCTRL					0x00000002
#define LGI_EF_CTRL						(LGI_EF_LCTRL | LGI_EF_RCTRL)

#define LGI_EF_LALT						0x00000004
#define LGI_EF_RALT						0x00000008
#define LGI_EF_ALT						(LGI_EF_LALT | LGI_EF_RALT)

#define LGI_EF_LSHIFT					0x00000010
#define LGI_EF_RSHIFT					0x00000020
#define LGI_EF_SHIFT					(LGI_EF_LSHIFT | LGI_EF_RSHIFT)

#define LGI_EF_DOWN						0x00000040
#define LGI_EF_DOUBLE					0x00000080
#define LGI_EF_CAPS_LOCK				0x00000100
#define LGI_EF_IS_CHAR					0x00000200
#define LGI_EF_IS_NOT_CHAR				0x00000400

// Mouse Event Flags
#define LGI_EF_LEFT						0x00001000
#define LGI_EF_MIDDLE					0x00002000
#define LGI_EF_RIGHT					0x00004000


// Os metrics

/// Get the standard window horizontal border size
/// \sa GApp::GetMetric()
#define LGI_MET_DECOR_X					1
/// Get the standard window vertical border size
/// \sa GApp::GetMetric()
#define LGI_MET_DECOR_Y					2
/// Get the height of a single line menu bar
/// \sa GApp::GetMetric()
#define LGI_MET_MENU					3

//
#ifdef _DEBUG
#define DeclDebugArgs				, char *_file, int _line
#define PassDebugArgs				, __FILE__, __LINE__
#else
#define DeclDebugArgs
#define PassDebugArgs
#endif

// Byte order
#ifdef MAC
#define LGI_LITTLE_ENDIAN				0
#define LGI_BIG_ENDIAN					1
#else
#define LGI_LITTLE_ENDIAN				1
#define LGI_BIG_ENDIAN					0
#endif

#endif
