#include <Path.h>
#include "Lgi.h"
#include "GSkinEngine.h"

class GAppPrivate
{
public:
	// Common
	OsAppArguments Args;
	Xml *Config;
	GFileSystem *FileSystem;
	GLibrary *SkinLib;
	GdcDevice *GdcSystem;

	GAppPrivate()
	{
		FileSystem = 0;
		GdcSystem = 0;
		Config = 0;
		SkinLib = 0;
	}

	~GAppPrivate()
	{
		DeleteObj(Config);
		DeleteObj(FileSystem);
		DeleteObj(GdcSystem);
		DeleteObj(SkinLib)
	}
};

/////////////////////////////////////////////////////////////////////////////
GApp *GApp::TheApp = 0;
class GSkinEngine *GApp::SkinEngine = 0;

GApp::GApp(char *Mime, OsAppArguments &AppArgs) : BApplication(Mime)
{
	// Sanity Checks
	LgiAssert(sizeof(int8) == 1);
	LgiAssert(sizeof(uint8) == 1);

	LgiAssert(sizeof(int16) == 2);
	LgiAssert(sizeof(uint16) == 2);

	LgiAssert(sizeof(int32) == 4);
	LgiAssert(sizeof(uint32) == 4);

	LgiAssert(sizeof(int64) == 8);
	LgiAssert(sizeof(uint64) == 8);
	
	LgiAssert(sizeof(char16) == 4);

	// Private data
	d = NEW(GAppPrivate);
	TheApp = this;
	d->Args = AppArgs;

	// Setup LGI Sub-systems
	GFontSystem::Inst();
	d->FileSystem = NEW(GFileSystem);
	d->GdcSystem = NEW(GdcDevice);

	srand(system_time());
	LgiInitColours();
	AppWnd = 0;
	SetAppArgs(AppArgs);

	// System font
	SystemNormal = 0;
	GFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemNormal = SysFontType.Create();
		if (SystemNormal)
		{
			SystemNormal->Transparent(true);
			
			SystemBold = NEW(GFont);
			if (SystemBold)
			{
				*SystemBold = *SystemNormal;
				SystemBold->Bold(true);
			}
		}
	}

	if (NOT SystemNormal)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}

	if (NOT GetOption("noskin"))
	{
		// Load library
		d->SkinLib = NEW(GLibrary("liblgiskin"));
		if (d->SkinLib)
		{
			if (d->SkinLib->IsLoaded())
			{
				Proc_CreateSkinEngine CreateSkinEngine =
					(Proc_CreateSkinEngine)d->SkinLib->GetAddress(LgiSkinEntryPoint);
				if (CreateSkinEngine)
				{
					SkinEngine = CreateSkinEngine(this);
					if (NOT SkinEngine)
					{
						printf("%s:%i - CreateSkinEngine returned NULL.\n", __FILE__, __LINE__);
					}
				}
				else
				{
					printf("%s:%i - No CreateSkinEngine entry point.\n", __FILE__, __LINE__);
				}
			}
			else
			{
				printf("%s:%i - Can't load skin library.\n", __FILE__, __LINE__);
				DeleteObj(d->SkinLib);
			}
		}
		else
		{
			printf("%s:%i - Can't find skin library.\n", __FILE__, __LINE__);
		}
	}
}

GApp::~GApp()
{
	TheApp = 0;
	DeleteObj(d);
}

bool GApp::IsOk()
{
	return true;
}

GView *GApp::GetFocus()
{
	for (int i=0; i<CountWindows(); i++)
	{
		BWindow *w = WindowAt(i);
		if (w AND w->IsActive())
		{
			BView *f = w->CurrentFocus();
			if (f)
			{
				BViewRedir *r = dynamic_cast<BViewRedir*>(f);
				if (r)
				{
					return r->WindowHandle();
				}
			}
		}
	}
	
	return 0;
}

int32 GApp::GetMetric(int Which)
{
	return 0;
}

bool GApp::Run(bool Loop)
{
	if (Loop)
	{
		OnCommandLine();
		BApplication::Run();
	}
	else
	{
		// ???
	}
}

void GApp::RefsReceived(BMessage *Msg)
{
	List<char> Files;
	int32 Count = 0;
	type_code Type = 0;
	
	if (Msg->GetInfo("refs", &Type, &Count) == B_OK)
	{
		for (int i=0; i<Count; i++)
		{
			entry_ref Ref;
			if (Msg->FindRef("refs", i, &Ref) == B_OK)
			{
				BPath Path;
				BEntry Entry(&Ref, true);
				Entry.GetPath(&Path);
				char *p = (char*) Path.Path();
				if (p)
				{
					Files.Insert(NewStr(p));
				}
			}
		}
	}
	
	if (AppWnd AND Files.GetItems() > 0)
	{
		AppWnd->OnReceiveFiles(Files);
	}
	
	for (char *f=Files.First(); f; f=Files.Next())
	{
		DeleteArray(f);
	}
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

void GApp::SetConfig(XmlTag *Tag)
{
	if (IsOk() AND Tag)
	{
		XmlTag *Old = GetConfig(Tag->Name);
		if (Old)
		{
			d->Config->Values.Delete(Old);
			DeleteObj(Old);
		}

		if (NOT d->Config)
		{
			d->Config = NEW(Xml);
		}
		if (d->Config)
		{
			d->Config->Values.Insert(Tag);
		}
	}
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT d->Config)
	{
		char Path[256];
		if (LgiGetExePath(Path, sizeof(Path)))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				d->Config = NEW(Xml);
				if (d->Config)
				{
					if (d->Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(d->Config);
					}
				}
			}
		}
	}

	if (Tag AND d->Config)
	{
		for (XmlTag *t=d->Config->Values.First(); t; t=d->Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Option, char *Dest, int DestLen)
{
	if (Option)
	{
		int OptLen = strlen(Option);
		
		for (int i=1; i<d->Args.Args; i++)
		{
			char *c = d->Args.Arg[i];
			if (*c == '/' OR *c == '\\' OR *c == '-')
			{
				c++;
				if (strncmp(c, Option, OptLen) == 0)
				{
					c += OptLen;
					
					if (NOT *c AND i < d->Args.Args - 1)
					{
						c = d->Args.Arg[++i];
					}
					
					if (c)
					{
						while (*c AND strchr(" \t\r\n", *c))
						{
							c++;
						}
		
						for (char *i = c; *i AND *i != ' ';)
						{
							*Dest++ = *i++;
						}
						*Dest++ = 0;
						return true;
					}
				}
			}
		}
	}

	return false;
}

void GApp::OnCommandLine()
{
	char WhiteSpace[] = " \r\n\t";

	List<char> Files;
	
	char *Delim = "\'\"";
	char *s;
	/*
	for (s = CmdLine; *s; )
	{
		// skip ws
		while (strchr(WhiteSpace, *s)) s++;

		// read to end of token
		char *e = s;
		if (strchr(Delim, *s))
		{
			for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
		}
		else
		{
			for (; *e AND NOT strchr(WhiteSpace, *e); e++);
		}

		char *Arg = NewStr(s, (int)e-(int)s);
		if (Arg)
		{
			if (FileExists(Arg))
			{
				Files.Insert(Arg);
			}
			else
			{
				DeleteArray(Arg);
			}
		}

		// next
		s = (*e) ? e + 1 : e;
	}
	*/

	// call app
	if (Files.GetItems() > 0)
	{
		OnReceiveFiles(Files);
	}

	// clear up
	Files.DeleteArrays();
}

OsAppArguments *GApp::GetAppArgs()
{
	return &d->Args;
}

void GApp::SetAppArgs(OsAppArguments &AppArgs)
{
	d->Args = AppArgs;
}

void GApp::Exit(int Code)
{
	if (Code)
	{
		// hard exit
		exit(-1);
	}
	else
	{
		// normal exit
		BMessenger m(0, be_app);
		m.SendMessage(B_QUIT_REQUESTED);
	}
}

