/**
 * \file
 * \author Matthew Allen
 * \date 27/7/1995
 * \brief Container classes.\n
	Copyright (C) 1995-2003, <a href="mailto:fret@memecode.com">Matthew Allen</a>
 */

#ifdef LINUX
#define _ISOC99_SOURCE		1
#include <wchar.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#include "LgiDefs.h"
#include "GMem.h"
#include "GContainers.h"
#include "GString.h"
#include "LgiCommon.h"

/////////////////////////////////////////////////////////////////////////////////////////////////

#if 0 && defined(_DEBUG)
#define AssertValid()			LgiAssert(IsValid())
#else
#define AssertValid()
#endif

#if 0
#define InvalidListError()		{ _asm { int 3 } return false; }
#else
#define InvalidListError()		{ LgiAssert(0); return false; }
#endif

#define ITEM_PTRS	64
class Item
{
public:
	DLinkList *List;
	Item *Next, *Prev;
	char Count;
	void *Ptr[ITEM_PTRS];

	Item(DLinkList *list)
	{
		LgiAssert(list);

		List = list;
		Next = Prev = 0;
		Count = 0;
		ZeroObj(Ptr);

		LgiAssert(List->FirstObj == 0);
		LgiAssert(List->LastObj == 0);
		List->FirstObj = List->LastObj = this;
	}

	Item(Item *item)
	{
		LgiAssert(item);

		Count = 0;
		ZeroObj(Ptr);
		List = item->List;
		Prev = item;
		if (Prev->Next)
		{
			Next = Prev->Next;
			Next->Prev = this;
		}
		else
		{
			Next = 0;
			if (List->LastObj == item)
			{
				List->LastObj = this;
			}
		}
		Prev->Next = this;
	}

	~Item()
	{
		if (Prev)
		{
			Prev->Next = Next;
		}
		else if (List)
		{
			LgiAssert(List->FirstObj == this);
			List->FirstObj = Next;
		}

		if (Next)
		{
			Next->Prev = Prev;
		}
		else if (List)
		{
			LgiAssert(List->LastObj == this);
			List->LastObj = Prev;
		}
	}

	void Delete(int &Index, Item *&Owner)
	{
		if (Index < Count-1)
		{
			memmove(Ptr+Index, Ptr+Index+1, (Count-Index-1) * sizeof(*Ptr));
		}

		if (--Count == 0)
		{
			// This Item is now empty, remove and reset current
			// into the next Item
			Owner = Next;
			Index = 0;
			delete this;
		}
		else if (Index >= Count)
		{
			// Carry current item over to next Item
			Owner = Next;
			Index = 0;
		}
	}

	bool Full()
	{
		return Count >= ITEM_PTRS;
	}

	void Insert(void *p, int Index = -1)
	{
		if (NOT Full())
		{
			if (Index < 0)
			{
				Index = Count;
			}
			else if (Index < Count)
			{
				memmove(Ptr+Index+1, Ptr+Index, (Count-Index) * sizeof(*Ptr));
			}
			Ptr[Index] = p;
			Count++;

			LgiAssert(Count <= ITEM_PTRS);
		}
		else
		{
			if (NOT Next)
			{
				NEW(Item(this));
			}

			if (Next)
			{
				if (Index < 0)
				{
					Next->Insert(p, Index);
				}
				else
				{
					// Push last pointer into Next
					if (Next->Full())
					{
						// Create an empty Next
						NEW(Item(this));
						if (Next)
						{
							Next->Insert(Ptr[ITEM_PTRS-1], 0);
						}
					}
					else
					{
						Next->Insert(Ptr[ITEM_PTRS-1], 0);
					}

					// Shulffle up points
					Count--;
					
					// Insert here
					Insert(p, Index);
				}
			}
		}
	}
};

class ItemIter
{
public:
	Item *i;
	int Cur;

	ItemIter()
	{
		i = 0;
		Cur = 0;
	}

	ItemIter(Item *item, int c)
	{
		i = item;
		Cur = c;
	}

	operator void*()
	{
		return (i AND Cur>=0 AND Cur<i->Count) ? i->Ptr[Cur] : 0;
	}
	
	ItemIter &operator =(Item *item)
	{
		i = item;
		return *this;
	}
	
	ItemIter &operator =(int c)
	{
		Cur = c;
		return *this;
	}

	ItemIter &operator =(ItemIter *iter)
	{
		i = iter->i;
		Cur = iter->Cur;
		return *this;
	}

	int GetIndex(int Base)
	{
		if (i)
		{
			return Base + Cur;
		}

		return -1;
	}

	bool Next()
	{
		if (i)
		{
			Cur++;
			if (Cur >= i->Count)
			{
				i = i->Next;
				if (i)
				{
					Cur = 0;
					return i->Count > 0;
				}
			}
			else return true;
		}

		return false;
	}

	bool Prev()
	{
		if (i)
		{
			Cur--;
			if (Cur < 0)
			{
				i = i->Prev;
				if (i AND i->Count > 0)
				{
					Cur = i->Count - 1;
					return true;
				}
			}
			else return true;
		}

		return false;
	}

	bool Delete()
	{
		if (i)
		{
			i->Delete(Cur, i);
			return true;
		}

		return false;
	}
};

DLinkList::DLinkList()
{
	Items = 0;
	FirstObj = LastObj = 0;
	Cur = 0;
}

DLinkList::~DLinkList()
{
	Empty();
	DeleteObj(Cur);
}

DLinkList &DLinkList::operator =(DLinkList &lst)
{
	Empty();

	for (Item *p = lst.FirstObj; p; p = p->Next)
	{
		for (int i=0; i<p->Count; i++)
		{
			Insert(p->Ptr[i]);
		}
	}
	
	return *this;
}

bool DLinkList::IsValid()
{
	static int Run = 0;

	// Check the cursor is not pointing at something weird
	if (Cur)
	{
		if (Cur->i)
		{
			if (Cur->i->Count > 0)
			{
				int asd=0;
			}
		}
	}

	// Run down the list and verify objects
	{
		int Count = 0;

		for (Item *i = FirstObj; i; i = i->Next)
		{
			if (i->List != this)
			{
				InvalidListError();
			}

			if (i->Next)
			{
				if (i->Next->Prev != i)
				{
					InvalidListError();
				}
			}
			else
			{
				if (i != LastObj)
				{
					InvalidListError();
				}
			}

			if (i->Count == 0)
			{
				InvalidListError();
			}

			for (int n=0; n<i->Count; n++)
			{
				if (i->Ptr[n] == 0)
				{
					InvalidListError();
				}

				#if 0
				if (NOT LgiCanReadMemory(i->Ptr[n]))
				{
					InvalidListError();
				}
				#endif
			}

			Count += i->Count;
		}
		
		if (Items != Count)
		{
			InvalidListError();
		}
	}

	return true;
}

void DLinkList::Empty()
{
	AssertValid();

	DeleteObj(Cur);
	while (FirstObj)
	{
		Item *n = FirstObj->Next;
		DeleteObj(FirstObj);
		FirstObj = n;
	}
	Items = 0;

	AssertValid();
}

ItemIter DLinkList::GetIndex(int Index)
{
	ItemIter Iter;

	int n = 0;
	for (Item *i = FirstObj; i; i = i->Next)
	{
		if (Index >= n AND Index < n + i->Count)
		{
			Iter = i;
			Iter = Index - n;
			break;
		}

		n += i->Count;
	}

	return Iter;
}

ItemIter DLinkList::GetPtr(void *Ptr, int &Base)
{
	ItemIter Iter;

	int n = 0;
	for (Item *i = FirstObj; i; i = i->Next)
	{
		for (int j=0; j<i->Count; j++)
		{
			if (i->Ptr[j] == Ptr)
			{
				Base = n;
				Iter = j;
				Iter = i;
				break;
			}
		}

		n += i->Count;
	}

	return Iter;
}

void *DLinkList::operator [](int Index)
{
	AssertValid();

	if (Cur)
	{
		return (*Cur = GetIndex(Index));
	}

	return 0;
}

int DLinkList::IndexOf(void *p)
{
	AssertValid();

	if (Cur)
	{
		int b;
		*Cur = GetPtr(p, b);
		return Cur->GetIndex(b);
	}

	return -1;
}

bool DLinkList::Delete()
{
	bool Status = false;

	AssertValid();

	if (Cur)
	{
		Cur->Delete();
		Items--;
		Status = true;

		if (Items == 0)
		{
			DeleteObj(Cur);
		}
	}

	AssertValid();

	return Status;
}

bool DLinkList::Delete(int Index)
{
	bool Status = false;

	AssertValid();

	if (Cur)
	{
		*Cur = GetIndex(Index);
		if (Status = Cur->Delete())
		{
			Items--;

			if (Items == 0)
			{
				DeleteObj(Cur);
			}
		}
	}

	AssertValid();

	return Status;
}

bool DLinkList::Delete(void *p)
{
	bool Status = false;

	AssertValid();

	if (Cur)
	{
		int b;
		*Cur = GetPtr(p, b);
		if (Status = Cur->Delete())
		{
			Items--;

			if (Items == 0)
			{
				DeleteObj(Cur);
			}
		}
	}

	AssertValid();

	return Status;
}

bool DLinkList::Insert(void *p, int Index)
{
	bool Status = false;

	#ifdef _DEBUG
	if ((uint)p < 100)
	{
		char *s = 0;
		*s = 0;
	}
	#endif

	AssertValid();

	LgiAssert(p);

	if (p)
	{
		if (Index >= 0 AND Index < Items)
		{
			ItemIter Iter = GetIndex(Index);
			if (Iter.i)
			{
				Iter.i->Insert(p, Iter.Cur);
				Items++;
				Status = true;
			}
		}
		else
		{
			if (LastObj)
			{
				LastObj->Insert(p);
				Items++;
				Status = true;
			}
			else
			{
				NEW(Item(this));
				if (FirstObj)
				{
					FirstObj->Insert(p);

					if (NOT Cur)
					{
						Cur = NEW(ItemIter);
					}

					Items++;
					Status = true;
				}
			}
		}
	}

	AssertValid();

	return Status;
}

bool DLinkList::HasItem(void *p)
{
	AssertValid();

	if (Cur)
	{
		int b;
		*Cur = GetPtr(p, b);
		if ((void*)*Cur)
		{
			return true;
		}
	}

	return false;
}

void *DLinkList::ItemAt(int Index)
{
	AssertValid();

	if (Cur)
	{
		*Cur = GetIndex(Index);
		return *Cur;
	}

	return 0;
}

void *DLinkList::Current()
{
	AssertValid();

	if (Cur)
	{
		return *Cur;
	}

	return 0;
}

void *DLinkList::First()
{
	AssertValid();

	if (Cur)
	{
		*Cur = FirstObj;
		*Cur = 0;
		return *Cur;
	}

	return 0;
}

void *DLinkList::Last()
{
	AssertValid();

	if (Cur)
	{
		*Cur = LastObj;
		if (LastObj) *Cur = LastObj->Count-1;
		return *Cur;
	}

	return 0;
}

void *DLinkList::Next()
{
	if (Cur)
	{
		Cur->Next();
		return *Cur;
	}
	return 0;
}

void *DLinkList::Prev()
{
	if (Cur)
	{
		Cur->Prev();
		return *Cur;
	}
	return 0;
}

typedef int (*BTreeCompare)(void *a, void *b, int Data);
typedef void **AddrVoidPtr;

class BTreeNode
{
public:
	void *Node;
	BTreeNode *Left;
	BTreeNode *Right;

	AddrVoidPtr *Index(AddrVoidPtr *Items)
	{
		if (Left)
		{
			Items = Left->Index(Items);
		}

		**Items = Node;
		Items++;

		if (Right)
		{
			Items = Right->Index(Items);
		}

		return Items;
	}	
};

class BTree
{
	int Items;
	int Used;
	BTreeNode *Node;
	BTreeNode *Min;
	BTreeNode *Max;

public:
	BTree(int i)
	{
		Used = 0;
		Min = Max = 0;
		Node = NEW(BTreeNode[Items = i]);
		if (Node)
		{
			memset(Node, 0, Items * sizeof(BTreeNode));
		}
	}

	~BTree()
	{
		DeleteArray(Node);
	}

	int GetItems() { return Used; }

	bool Add(void *Obj, BTreeCompare Proc, int Data)
	{
		if (Used)
		{
			BTreeNode *Cur = Node;
			
			if (Used < Items)
			{
				if (Proc(Obj, Max->Node, Data) >= 0)
				{
					Max->Right = Node + Used++;
					Max = Max->Right;
					Max->Node = Obj;
					return true;
				}

				if (Proc(Obj, Min->Node, Data) < 0)
				{
					Min->Left = Node + Used++;
					Min = Min->Left;
					Min->Node = Obj;
					return true;
				}

				while (true)
				{
					int c = Proc(Obj, Cur->Node, Data);
					BTreeNode **Ptr = (c < 0) ? &Cur->Left : &Cur->Right;
					if (*Ptr)
					{
						Cur = *Ptr;
					}
					else if (Used < Items)
					{
						*Ptr = &Node[Used++];
						(*Ptr)->Node = Obj;
						return true;
					}
					else return false;
				}
			}
			else
			{
				LgiAssert(0);
			}

			return false;
		}
		else
		{
			Min = Max = Node;
			Node[Used++].Node = Obj;
			return true;
		}
	}

	void Index(void ***Items)
	{
		if (Node)
		{
			Node->Index(Items);
		}
	}
};

void DLinkList::Sort(ListSortFunc Compare, int Data)
{
	if (Items > 1)
	{
		BTree Tree(Items);
		AddrVoidPtr *iLst = NEW(AddrVoidPtr[Items]);
		if (iLst)
		{
			int n=0;
			Item *i = FirstObj;
			while (i)
			{
				for (int k=0; k<i->Count; k++)
				{
					iLst[n++] = i->Ptr + k;
					Tree.Add(i->Ptr[k], Compare, Data);
				}
				i = i->Next;
			}

			Tree.Index(iLst);
			DeleteArray(iLst);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
DLinkIterator::DLinkIterator(DLinkList *list)
{
	List = list;
	Cur = NEW(ItemIter);
}

DLinkIterator::~DLinkIterator()
{
	List = 0;
	DeleteObj(Cur);
}

int DLinkIterator::GetItems()
{
	return List->GetItems();
}

void *DLinkIterator::operator [](int Index)
{
	if (Cur)
	{
		return (*Cur = List->GetIndex(Index));
	}

	return 0;
}

void *DLinkIterator::Current()
{
	if (Cur)
	{
		return *Cur;
	}

	return 0;
}

void *DLinkIterator::First()
{
	if (Cur)
	{
		*Cur = List->FirstObj;
		*Cur = 0;
		return *Cur;
	}

	return 0;
}

void *DLinkIterator::Last()
{
	if (Cur)
	{
		*Cur = List->LastObj;
		if (List->LastObj) *Cur = List->LastObj->Count-1;
		return *Cur;
	}

	return 0;
}

void *DLinkIterator::Next()
{
	if (Cur)
	{
		Cur->Next();
		return *Cur;
	}
	return 0;
}

void *DLinkIterator::Prev()
{
	if (Cur)
	{
		Cur->Prev();
		return *Cur;
	}
	return 0;
}

bool DLinkIterator::HasItem(void *p)
{
	if (Cur)
	{
		int b;
		*Cur = List->GetPtr(p, b);
		if ((void*)*Cur)
		{
			return true;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////////////
GBytePipe::GBytePipe(int prealloc)
{
	PreAlloc = prealloc;
}

GBytePipe::~GBytePipe()
{
	Empty();
}

void GBytePipe::Empty()
{
	for (Block *b = Mem.First(); b; b = Mem.First())
	{
		Mem.Delete(b);
		free(b);
	}
}

int64 GBytePipe::GetSize()
{
	int Size = 0;
	for (Block *b = Mem.First(); b; b = Mem.Next())
	{
		Size += b->Used - b->Next;
	}
	return Size;
}

int GBytePipe::Peek(uchar *Ptr, int Size)
{
	bool Status = false;

	if (Ptr AND Size <= GetSize())
	{
		Block *b = 0;
		for (b = Mem.First(); b AND Size > 0; b = Mem.Next())
		{
			int Copy = min(Size, b->Size - b->Next);
			if (Copy > 0)
			{
				memcpy(Ptr, b->Ptr() + b->Next, Copy);
				Ptr += Copy;
				Size -= Copy;
			}
		}

		Status = (Size == 0);
	}

	return Status;
}

void *GBytePipe::New(int AddBytes)
{
	int Len = GetSize();
	uchar *Data = Len > 0 ? NEW(uchar[Len+AddBytes]) : 0;
	if (Data)
	{
		Read(Data, Len);
		if (AddBytes)
		{
			memset(Data+Len, 0, AddBytes);
		}
	}
	return Data;
}

int GBytePipe::Read(void *Ptr, int Size, int Flags)
{
	int Status = 0;

	if (Ptr AND Size > 0)
	{
		Block *b = 0;
		for (b = Mem.First(); b AND Size > 0; b = Mem.Next())
		{
			int Copy = min(Size, b->Used - b->Next);
			if (Copy > 0)
			{
				memcpy(Ptr, b->Ptr() + b->Next, Copy);
				((uchar*&)Ptr) += Copy;
				Size -= Copy;
				b->Next += Copy;
				Status += Copy;
			}
		}

		for (b = Mem.First(); b AND b->Next >= b->Used; b = Mem.First())
		{
			Mem.Delete(b);
			free(b);
		}
	}

	return Status;
}

int GBytePipe::Pop(short &i)
{
	short n;
	if (Read((uchar*) &n, sizeof(n)))
	{
		i = n;
		return sizeof(n);
	}
	return 0;
}

int GBytePipe::Pop(int &i)
{
	int n;
	if (Read((uchar*) &n, sizeof(n)))
	{
		i = n;
		return sizeof(n);
	}
	return 0;
}

int GBytePipe::Pop(double &i)
{
	double n;
	if (Read((uchar*) &n, sizeof(n)))
	{
		i = n;
		return sizeof(n);
	}
	return 0;
}

int GBytePipe::Write(void *Ptr, int Size, int Flags)
{
	int Status = 0;

	if (Ptr AND Size > 0)
	{
		if (PreAlloc > 0)
		{
			Block *Last = Mem.Last();
			if (Last)
			{
				int Len = min(Size, Last->Size - Last->Used);
				if (Len > 0)
				{
					memcpy(Last->Ptr() + Last->Used, Ptr, Len);
					Last->Used += Len;
					Size -= Len;
					((uchar*&)Ptr) += Len;
					Status += Len;
				}
			}
		}

		if (Size > 0)
		{
			int Alloc = max(PreAlloc, Size);

			Block *b = (Block*) malloc(sizeof(Block) + Alloc);
			if (b)
			{
				void *p = b->Ptr();
				memcpy(p, Ptr, Size);
				b->Size = Alloc;
				b->Used = Size;
				b->Next = 0;
				Mem.Insert(b);
				Status += Size;
			}
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////
int GStringPipe::Pop(char *Str, int BufSize)
{
	if (Str)
	{
		char *Start = Str;
		char *End = Str + BufSize - 2;
		Block *m;
		bool HasLf = false;

		for (m = Mem.First(); m; m = Mem.Next())
		{
			char *MPtr = (char*)m->Ptr();
			if (strnchr(MPtr + m->Next, '\n', m->Used - m->Next))
			{
				HasLf = true;
				break;
			}
		}

		if (HasLf)
		{
			m = Mem.First();
			while (m)
			{
				char *MPtr = (char*)m->Ptr();
				while (Str < End AND m->Next < m->Used)
				{
					*Str = MPtr[m->Next++];
					if (*Str == '\n')
					{
						Str++;
						goto EndPop;
					}
					Str++;
				}

				if (m->Next >= m->Used)
				{
					Mem.Delete(m);
					free(m);
					m = Mem.First();
				}
				else if (Str >= End)
				{
					break;
				}
			}

			EndPop:
			*Str = 0;
			return (int)Str-(int)Start;
		}
	}

	return 0;
}

int GStringPipe::Push(char *Str, int Len)
{
	if (Str)
	{
		return Write((uchar*) Str, (Len>=0) ? Len : strlen(Str));
	}
	return 0;
}

int GStringPipe::Push(char16 *Str, int Len)
{
	if (Str)
	{
		return Write((uchar*) Str, ((Len>=0) ? Len : StrlenW(Str)) * sizeof(char16));
	}
	return 0;
}

bool GStringPipe::Printf(char *Str, ...)
{
	bool Status = false;

	if (Str)
	{
		va_list Arg;
		va_start(Arg, Str);
		#ifdef WIN32
		// Stupid windows can't tell me how many bytes I need to
		// allocate for the buffer, so I just keep increasing the
		// size until it works.
		for (int i=256; i <= (1 << 20); i <<= 2)
		{
			char *Buf = NEW(char[i]);
			if (Buf)
			{
				int Len = _vsnprintf(Buf, i, Str, Arg);
				if (Len >= 0)
				{
					Status = Push(Buf, Len);
					DeleteArray(Buf);
					break;
				}

				DeleteArray(Buf);
			}
		}
		#else
		int Len = vsnprintf(0, 0, Str, Arg);
		if (Len > 0)
		{
			char *Buf = NEW(char[Len+1]);
			if (Buf)
			{
				vsprintf(Buf, Str, Arg);
				Status = Push(Buf, Len);
				DeleteArray(Buf);
			}
		}
		else
		{
			LgiAssert(0);
		}
		#endif
		va_end(Arg);
	}

	return Status;
}

bool GStringPipe::Printf(char16 *Str, ...)
{
	bool Status = false;
	
	if (Str)
	{
		va_list Arg;
		va_start(Arg, Str);

		#if defined(LINUX) || defined(BEOS)

			char *Format = LgiNewUtf16To8(Str);
			if (Format)
			{
				int Len = vsnprintf(0, 0, Format, Arg);
				if (Len > 0)
				{
					char *Buf = NEW(char[Len+1]);
					if (Buf)
					{
						vsprintf(Buf, Format, Arg);
						Status = Push(Buf, Len);
						DeleteArray(Buf);
					}
				}
				else
				{
					LgiAssert(0);
				}
				
				DeleteArray(Format);
			}
		
		#elif defined(WIN32)
		
			// Stupid windows can't tell me how many bytes I need to
			// allocate for the buffer, so I just keep increasing the
			// size until it works.
			for (int i=256; i <= (1 << 20); i <<= 2)
			{
				char16 *Buf = NEW(char16[i]);
				if (Buf)
				{
					int Len = _vsnwprintf(Buf, i, Str, Arg);
					if (Len >= 0)
					{
						Status = Push(Buf, Len);
						DeleteArray(Buf);
						break;
					}

					DeleteArray(Buf);
				}
			}

		#else

			#error "Impl. Me"

		#endif

		va_end(Arg);
	}

	return Status;
}

char *GStringPipe::NewStr()
{
	return (char*)New(1);
}

char16 *GStringPipe::NewStrW()
{
	return (char16*)New(sizeof(char16));
}

////////////////////////////////////////////////////////////////////////
// Hashing
class GHashEntry
{
public:
	char *Key;
	void *Value;

	GHashEntry()
	{
		Key = 0;
		Value = 0;
	}

	~GHashEntry()
	{
		DeleteArray(Key);
	}
};

/** \brief Creates a 32-bit hash of the input string.
 */
uint LgiHash(uchar *v, int l, bool Case)
{
	uint32 h = 0;

	if (Case)
	{
		// case sensitive
		if (l > 0)
		{
			while (l--)
			{
				h = (h << 5) - h + *v++;
			}
		}
		else
		{
			for (; *v; v ++)
			{
				h = (h << 5) - h + *v;
			}
		}
	}
	else
	{
		// case insensitive
		char c;
		if (l > 0)
		{
			while (l--)
			{
				c = tolower(*v);
				v++;
				h = (h << 5) - h + c;
			}
		}
		else
		{
			for (; *v; v++)
			{
				c = tolower(*v);
				h = (h << 5) - h + c;
			}
		}
	}

	return h;
}
	
#define HASH_TABLE_SHRINK_THRESHOLD			15
#define HASH_TABLE_GROW_THRESHOLD			70

struct StringPool
{
	int Size;
	int Used;
	char *Mem;

	StringPool()
	{
		Size = 1024 << 10;
		Used = 0;
		Mem = NEW(char[Size]);
	}

	~StringPool()
	{
		DeleteArray(Mem);
	}

	char *NewStr(char *s)
	{
		int Len = strlen(s) + 1;
		if (Used < Size - Len)
		{
			char *p = Mem + Used;
			strcpy(p, s);
			Used += Len;
			return p;
		}
		return 0;
	}
};

class GHashTablePrivate
{
public:
	int Size;
	GHashEntry *Table;
	int Cur;
	int Used;
	bool Case;

	// String pooling: reduces memory allocs/frees but will waste memory
	// if you are deleting keys.
	bool Pool;
	GArray<StringPool*> Pools;

	GHashTablePrivate(int size)
	{
		Used = 0;
		Cur = -1;
		Case = true;
		Pool = false;
		Size = size ? max(size, 16) : 512;
		Table = NEW(GHashEntry[Size]);
	}

	~GHashTablePrivate()
	{
		Empty();
		DeleteArray(Table);
	}

	int Percent()
	{
		return Used * 100 / Size;
	}

	bool SetSize(int NewSize)
	{
		bool Status = false;
		
// LgiTrace("*******************\n** SET SIZE %i\n*******************\n", NewSize);

		NewSize = max(NewSize, Used + 16);
		if (NewSize != Size)
		{
			int OldSize = Size;
			GHashEntry *OldTable = Table;

			Used = 0;
			Size = NewSize;
			Table = NEW(GHashEntry[Size]);
			if (Table)
			{
				GArray<StringPool*> OldPools;
				OldPools = Pools;
				Pools.Length(0);

				for (int i=0; i<OldSize; i++)
				{
					if (OldTable[i].Key)
					{
						if (NOT Add(OldTable[i].Key, OldTable[i].Value))
						{
							LgiAssert(0);
						}
					}
				}

				OldPools.DeleteObjects();
				Status = true;
			}

			if (Pool AND OldTable)
			{
				memset(OldTable, 0, sizeof(*OldTable) * OldSize);
			}
			DeleteArray(OldTable);
		}

		return Status;
	}

	int Compare(char *a, char *b)
	{
		if (Case)
		{
			return strcmp(a, b);
		}
		else
		{
			return stricmp(a, b);
		}
	}

	bool GetEntry(char *Key, int &Index, bool Debug = false)
	{
		if (Key AND Table)
		{
			uint32 h = LgiHash((uchar*)Key, 0, Case);

			if (Debug)
			{
				printf("GetEntry h=%i Size=%i\n", h, Size);
			}
			
			for (int i=0; i<Size; i++)
			{
				Index = (h + i) % Size;
				LgiAssert(Index >= 0);

				if (Debug)
				{
					printf("Get[%i]: Key='%s'\n", Index, Table[Index].Key);
				}
				
				if (NOT Table[Index].Key)
				{
					return false;
				}
				
				if (Compare(Table[Index].Key, Key) == 0)
				{
					return true;
				}
			}
		}

		return false;
	}

	bool Add(char *Key, void *Value)
	{
		if (Key AND Value)
		{
			uint32 h = LgiHash((uchar*)Key, 0, Case);

			int Index = -1;
			if (Case)
			{
				for (int i=0; i<Size; i++)
				{
					int k = (h + i) % Size;
					if
					(
						NOT Table[k].Key
						OR
						(
							Table[k].Key[0] == Key[0]
							AND
							strcmp(Table[k].Key, Key) == 0
						)
					)
					{
						Index = k;
						break;
					}
				}
			}
			else
			{
				for (int i=0; i<Size; i++)
				{
					int k = (h + i) % Size;
					if
					(
						NOT Table[k].Key
						OR
						(
							tolower(Table[k].Key[0]) == tolower(Key[0])
							AND
							stricmp(Table[k].Key, Key) == 0
						)
					)
					{
						Index = k;
						break;
					}
				}
			}

			if (Index >= 0)
			{
				if (NOT Table[Index].Key)
				{
					if (Pool)
					{
						StringPool *p = 0;
						if (Pools.Length() == 0)
						{
							Pools[0] = p = NEW(StringPool);
						}
						else
						{
							p = Pools[Pools.Length()-1];
						}
						Table[Index].Key = p->NewStr(Key);
						if (!Table[Index].Key)
						{
							Pools.Add(p = NEW(StringPool));
							Table[Index].Key = p->NewStr(Key);
						}
					}
					else
					{
						Table[Index].Key = NewStr(Key);
					}
					Used++;
				}
				Table[Index].Value = Value;

				if (Percent() > HASH_TABLE_GROW_THRESHOLD)
				{
					SetSize(Size << 1);
				}
				return true;
			}
		}

		return false;
	}
	
	bool Between(int Value, int Min, int Max)
	{
		if (Min <= Max)
		{
			// Not wrapped
			return Value >= Min AND Value <= Max;
		}
		else
		{
			// Wrapped
			return Value <= Max OR Value >= Min;
		}
	}

	bool Delete(char *Key)
	{
		int Index = -1;
		if (GetEntry(Key, Index))
		{
			// Delete the entry
// LgiTrace("Hash::Delete[%i] Key='%s'\n", Index, Table[Index].Key);
			
			if (Pool)
			{
				// Memory is owned by the string pool, not us.
				Table[Index].Key = 0;
			}
			else
			{
				DeleteArray(Table[Index].Key);
			}
			Table[Index].Value = 0;
			Used--;
			
			// Bubble down any entries above the hole
			int Hole = Index;
			for (int i = (Index + 1) % Size; i != Index; i = (i + 1) % Size)
			{
				if (Table[i].Key)
				{
					uint32 Hash = LgiHash((uchar*)Table[i].Key, 0, Case);
					uint32 HashIndex = Hash % Size;
					
					if (HashIndex != i AND Between(Hole, HashIndex, i))
					{
						// Do bubble
// LgiTrace("**** BUBBLE NODE %i->%i HashIndex=%i key=%s\n", i, Hole, HashIndex, Table[i].Key);
						
						if (Table[Hole].Key)
						{
							printf("Error at %i, Key='%s'\n", Hole, Table[Hole].Key);
							LgiAssert(0);
						}
						memmove(Table + Hole, Table + i, sizeof(Table[i]));
						memset(Table + i, 0, sizeof(Table[i]));
						Hole = i;
					}
				}
				else
				{
					// Reached the end of entries that could have bubbled
					break;
				}
			}

			// Check for auto-shrink limit
			if (Percent() < HASH_TABLE_SHRINK_THRESHOLD)
			{
				SetSize(Size >> 1);
			}
			
			return true;
		}
		else
		{
			printf("%s:%i - Couldn't get entry to delete key='%s'.\n", __FILE__, __LINE__, Key);
			GetEntry(Key, Index, true);
		}

		return false;
	}

	void *First(char **Key)
	{
		Cur = 0;
		if (Table)
		{
			while (Cur < Size)
			{
				if (Table[Cur].Key)
				{
					if (Key)
					{
						*Key = Table[Cur].Key;
					}
					return Table[Cur].Value;
				}
				else Cur++;
			}
		}

		return 0;
	}

	void *Current(char **Key)
	{
		if (Cur >= 0 AND
			Cur < Size AND
			Table AND
			Table[Cur].Key)
		{
			if (Key)
			{
				*Key = Table[Cur].Key;
			}
			return Table[Cur].Value;
		}

		return 0;
	}

	void *Next(char **Key)
	{
		if (Table AND Cur >= 0)
		{
			while (++Cur < Size)
			{
				if (Table[Cur].Key)
				{
					if (Key)
					{
						*Key = Table[Cur].Key;
					}
					return Table[Cur].Value;
				}
			}
		}

		return 0;
	}
	
	void Empty()
	{
		if (Pool)
		{
			Pools.DeleteObjects();
			memset(Table, 0, sizeof(*Table) * Size);
		}
		else
		{
			for (int i=0; i<Size; i++)
			{
				DeleteArray(Table[i].Key);
				Table[i].Value = 0;
			}
		}
		Used = 0;
		SetSize(512);
	}
};

GHashTable::GHashTable(int size, bool Case)
{
	d = NEW(GHashTablePrivate(size));
	IsCase(Case);
}

GHashTable::GHashTable(List<char> &strs)
{
	d = NEW(GHashTablePrivate(strs.GetItems() * 2));
	Set(strs);
}

GHashTable::~GHashTable()
{
	DeleteObj(d);
}

void GHashTable::Set(List<char> &strs)
{
	for (char *s=strs.First(); s; s=strs.Next())
	{
		Add(s);
	}
}

int GHashTable::Length()
{
	return IsOk() ? d->Used : 0;
}

bool GHashTable::IsOk()
{
	bool Status =	this != 0 AND
					d != 0 AND
					d->Table != 0;
	LgiAssert(Status);
	return Status;
}

bool GHashTable::Add(char *Key, void *Value)
{
	return IsOk() AND d->Add(Key, Value);
}

bool GHashTable::Delete(char *Key)
{
	return IsOk() AND d->Delete(Key);
}

void *GHashTable::Find(char *Key)
{
	int Index = -1;
	if (IsOk() AND d->GetEntry(Key, Index))
	{
		return d->Table[Index].Value;
	}

	return 0;
}

void *GHashTable::First(char **Key)
{
	return IsOk() ? d->First(Key) : 0;
}

void *GHashTable::Current(char **Key)
{
	return IsOk() ? d->Current(Key) : 0;
}

void *GHashTable::Next(char **Key)
{
	return IsOk() ? d->Next(Key) : 0;
}

void GHashTable::Empty()
{
	if (IsOk())
	{
		d->Empty();
	}
}

bool GHashTable::IsCase()
{
	return IsOk() ? d->Case : false;
}

void GHashTable::IsCase(bool c)
{
	if (IsOk())
	{
		d->Case = c;
	}
}

int64 GHashTable::GetSize()
{
	return IsOk() ? d->Size : 0;
}

void GHashTable::SetSize(int64 s)
{
	if (IsOk())
	{
		s = max(s, d->Size * 10 / 7);
		d->SetSize(s);
	}
}

bool GHashTable::GetStringPool()
{
	return d->Pool;
}

void GHashTable::SetStringPool(bool b)
{
	d->Pool = b;
}

#ifdef _DEBUG
int64 GHashTable::Sizeof()
{
	int64 Size = sizeof(*this) + sizeof(*d);
	char s[64];

	Size += d->Size * sizeof(GHashEntry);

	LgiFormatSize(s, Size);
	LgiTrace("%s in %i hash entries (%i used)\n", s, d->Size, d->Used);

	int Keys = 0;
	int64 KeySize = 0;
	for (int i=0; i<d->Size; i++)
	{
		if (d->Table[i].Key)
		{
			Keys++;
			KeySize += strlen(d->Table[i].Key) + 1;
		}
	}

	LgiFormatSize(s, KeySize);
	LgiTrace("%s in %i keys\n", s, Keys);

	return Size + KeySize;
}
#endif