/*
**	FILE:		GDateTime.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		11/11/98
**	DESCRIPTION:	Scribe Date Time Object
**
**	Copyright (C) 1998, Matthew Allen
**		fret@memecode.com
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <sys/timeb.h>

#include "Lgi.h"
#include "GDateTime.h"
#include "GToken.h"

//////////////////////////////////////////////////////////////////////////////
uchar GDateTime::DefaultFormat = GDTF_DEFAULT;
uchar GDateTime::GetDefaultFormat()
{
	if (DefaultFormat == GDTF_DEFAULT)
	{
		#ifdef WIN32

		char s[2] = {'1', 0};
		GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDATE, s, sizeof(s));
		switch (atoi(s))
		{
			case 0:
				DefaultFormat = GDTF_MONTH_DAY_YEAR;
				break;
			default:
			case 1:
				DefaultFormat = GDTF_DAY_MONTH_YEAR;
				break;
			case 2:
				DefaultFormat = GDTF_YEAR_DAY_MONTH;
				break;
		}

		GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_ITIME, s, sizeof(s));
		if (atoi(s) == 1)
		{
			DefaultFormat |= GDTF_24HOUR;
		}
		else
		{
			DefaultFormat |= GDTF_12HOUR;
		}

		#else

		// get some user setting from the system??
		DefaultFormat = GDTF_DAY_MONTH_YEAR | GDTF_12HOUR;

		#endif
	}

	return DefaultFormat;
}

//////////////////////////////////////////////////////////////////////////////
GDateTime::GDateTime()
{
	_Day = 0;
	_Month = 0;
	_Year = 0;
	_Thousands = 0;
	_Seconds = 0;
	_Minutes = 0;
	_Hours = 0;

	_Format = GetDefaultFormat();
}

GDateTime::~GDateTime()
{
}

int GDateTime::TimeZone()
{
	timeb tbTime;
	ftime(&tbTime);
	return tbTime.timezone;
}

int GDateTime::TimeZoneOffset()
{
	int Status = 0;

	#ifdef WIN32
	/*
	int Mul = 1;

	#ifndef LGI_STATIC
	if (LgiGetOs() == LGI_OS_WIN9X)
	{
		// Win'9x seems to get stuck in "daylight savings" mode.
		RegKey k("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\TimeZoneInformation");
		if (k.GetInt("DisableAutoDaylightTimeSet"))
		{
			// Unstick it..
			Mul = 0;
		}

		// Maybe we should do this for other versions of windows.
		// I know XP works properly... I havn't tested ME/NT
	}
	#endif
	*/

	TIME_ZONE_INFORMATION Tzi;
	if (GetTimeZoneInformation(&Tzi) == TIME_ZONE_ID_DAYLIGHT)
	{
		Status = Tzi.DaylightBias;
	}
	#endif

	return Status;
}

int GDateTime::DayOfWeek()
{
	int Index = 0;
	int Day = IsLeapYear() ? 29 : 28;

	switch (_Year / 100)
	{
		case 19:
		{
			Index = 3;
			break;
		}
		case 20:
		{
			Index = 2;
			break;
		}
	}

	// get year right
	int y = _Year % 100;
	int r = y % 12;
	Index = (Index + (y / 12) + r + (r / 4)) % 7;
	
	// get month right
	if (_Month % 2 == 0)
	{
		// even month
		if (_Month > 2) Day = _Month;
	}
	else
	{
		// odd month
		switch (_Month)
		{
			case 1:
			{
				Day = 31;
				if (IsLeapYear())
				{
					Index = Index > 0 ? Index - 1 : Index + 6;
				}
				break;
			}
			case 11:
			case 3:
			{
				Day = 7;
				break;
			}
			case 5:
			{
				Day = 9;
				break;
			}
			case 7:
			{
				Day = 11;
				break;
			}
			case 9:
			{
				Day = 5;
				break;
			}
		}
	}

	// get day right
	int Diff = Index - (Day - _Day);
	while (Diff < 0) Diff += 7;

	return Diff % 7;
}

bool GDateTime::Set(int64 s)
{
	bool Status = false;

	#if defined WIN32
	FILETIME Utc, Local;
	SYSTEMTIME System;
	Utc.dwHighDateTime = s >> 32;
	Utc.dwLowDateTime = s & 0xffffffff;
	if (FileTimeToLocalFileTime(&Utc, &Local) AND
		FileTimeToSystemTime(&Local, &System))
	{
		_Year = System.wYear;
		_Month = System.wMonth;
		_Day = System.wDay;

		_Hours = System.wHour;
		_Minutes = System.wMinute;
		_Seconds = System.wSecond;
		_Thousands = System.wMilliseconds;

		Status = true;
	}
	#elif defined(LINUX) || defined(BEOS)
	struct tm *t = localtime((time_t*)&s);
	if (t)
	{
		/*
		printf("\tSet %i/%i/%i %i:%i:%i\n",
			t->tm_mday, t->tm_mon, t->tm_year,
			t->tm_hour, t->tm_min, t->tm_sec);
		*/
					
		_Year = t->tm_year + 1900;
		_Month = t->tm_mon + 1;
		_Day = t->tm_mday;

		_Hours = t->tm_hour;
		_Minutes = t->tm_min;
		_Seconds = t->tm_sec;
		_Thousands = 0;
	}
	#else
	#error Impl me.
	#endif

	return Status;
}

void GDateTime::SetNow()
{
	time_t now;
	time(&now);
	struct tm *time = localtime(&now);
	if (time)
	{
    	_Seconds = time->tm_sec;
    	_Minutes = time->tm_min;
    	_Hours = time->tm_hour;
    	_Day = time->tm_mday;
    	_Month = time->tm_mon + 1;
    	_Year = time->tm_year + 1900;
    }
	#ifndef LGI_STATIC
    else
    {
    	LgiTrace("%s:%i - Error: localtime failed, now=%u\n", __FILE__, __LINE__, now);
    }
	#endif
}

#define Convert24HrTo12Hr(h)			( (h) == 0 ? 12 : (h) > 12 ? (h) % 12 : (h) )
#define Convert24HrToAmPm(h)		( (h) >= 12 ? "p" : "a" )

void GDateTime::GetDate(char *Str)
{
	if (Str)
	{
		switch (_Format & GDTF_DATE_MASK)
		{
			case GDTF_MONTH_DAY_YEAR:
				sprintf(Str, "%i/%i/%i", _Month, _Day, _Year);
				break;
			default:
			case GDTF_DAY_MONTH_YEAR:
				sprintf(Str, "%i/%i/%i", _Day, _Month, _Year);
				break;
			case GDTF_YEAR_DAY_MONTH:
				sprintf(Str, "%i/%i/%i", _Year, _Month, _Day);
				break;
		}
	}
}

void GDateTime::GetTime(char *Str)
{
	if (Str)
	{
		switch (_Format & GDTF_TIME_MASK)
		{
			case GDTF_12HOUR:
			default:
			{
				sprintf(Str, "%i:%02.2i:%02.2i%s", Convert24HrTo12Hr(_Hours), _Minutes, _Seconds, Convert24HrToAmPm(_Hours));
				break;
			}
			case GDTF_24HOUR:
				sprintf(Str, "%i:%02.2i:%02.2i", _Hours, _Minutes, _Seconds);
				break;
		}
	}
}

bool GDateTime::Get(int64 &s)
{
	bool Status = false;

	#ifdef WIN32
	FILETIME Utc, Local;
	SYSTEMTIME System;

	System.wYear = _Year;
	System.wMonth = _Month;
	System.wDay = _Day;
	System.wHour = _Hours;
	System.wMinute = _Minutes;
	System.wSecond = _Seconds;
	System.wMilliseconds = _Thousands;

	if (SystemTimeToFileTime(&System, &Local) AND
		LocalFileTimeToFileTime(&Local, &Utc))
	{
		s = ((int64)Utc.dwHighDateTime << 32) | Utc.dwLowDateTime;
		Status = true;
	}		
	#elif defined(LINUX) || defined(BEOS)
	struct tm t;
	t.tm_year = _Year - 1900;
	t.tm_mon = _Month - 1;
	t.tm_mday = _Day;

	t.tm_hour = _Hours;
	t.tm_min = _Minutes;
	t.tm_sec = _Seconds;
	
	s = mktime(&t);
	#else
	#error Impl me.
	#endif

	return Status;
}

void GDateTime::Get(char *Str)
{
	if (Str)
	{
		GetDate(Str);
		strcat(Str, " ");
		GetTime(Str+strlen(Str));
	}
}

bool GDateTime::Set(char *Str)
{
	bool Status = false;

	if (Str)
	{
		char Local[256];
		strcpy(Local, Str);
		char *Sep = strchr(Local, ' ');
		if (Sep)
		{
			*Sep++ = 0;
			Status |= SetTime(Sep);
		}

		Status |= SetDate(Local);
	}

	return Status;
}

bool GDateTime::SetDate(char *Str)
{
	bool Status = false;

	if (Str)
	{
		GToken T(Str, "/-");
		if (T.Length() == 3)
		{
			switch (_Format & GDTF_DATE_MASK)
			{
				case GDTF_MONTH_DAY_YEAR:
				{
					_Month = atoi(T[0]);
					_Day = atoi(T[1]);
					_Year = atoi(T[2]);
					break;
				}
				default:
				case GDTF_DAY_MONTH_YEAR:
				{
					_Day = atoi(T[0]);
					_Month = atoi(T[1]);
					_Year = atoi(T[2]);
					break;
				}
				case GDTF_YEAR_DAY_MONTH:
				{
					_Year = atoi(T[0]);
					_Month = atoi(T[1]);
					_Day = atoi(T[2]);
					break;
				}
			}

			if (_Year < 100)
			{
				if (_Year >= 80)
				{
					_Year += 1900;
				}
				else
				{
					_Year += 2000;
				}
			}

			Status = true;
		}
	}

	return Status;
}

bool GDateTime::SetTime(char *Str)
{
	bool Status = false;

	if (Str)
	{
		GToken T(Str, ":");
		if (T.Length() >= 2 AND T.Length() <= 3)
		{
			_Hours = atoi(T[0]);
			_Minutes = atoi(T[1]);
			
			char *s = T[2];
			if (s) _Seconds = atoi(s);
			else _Seconds = 0;
			
			s = T[T.Length()-1];
			if (s AND
				(strchr(s, 'p') OR
				strchr(s, 'P')))
			{
				if (_Hours != 12)
				{
					_Hours += 12;
				}
			}
			else if (_Hours == 12)
			{
				_Hours -= 12;
			}

			Status = true;
		}
	}

	return Status;
}

int GDateTime::Sizeof()
{
	return sizeof(int) * 7;
}

bool GDateTime::Serialize(GFile &f, bool Write)
{
	int32 i;

	if (Write)
	{
		#define wf(fld) i = fld; f << i;
		wf(_Day);
		wf(_Month);
		wf(_Year);
		wf(_Thousands);
		wf(_Seconds);
		wf(_Minutes);
		wf(_Hours);
	}
	else
	{
		#define rf(fld) f >> i; fld = i;
		rf(_Day);
		rf(_Month);
		rf(_Year);
		rf(_Thousands);
		rf(_Seconds);
		rf(_Minutes);
		rf(_Hours);
	}

	return true;
}

bool GDateTime::Serialize(ObjProperties *Props, char *Name, bool Write)
{
	#ifndef LGI_STATIC
	if (Props AND Name)
	{
		struct _Date
		{
			uint8 Day;
			uint8 Month;
			int16 Year;
			uint8 Hour;
			uint8 Minute;
			uint16 ThouSec;
		};

		LgiAssert(sizeof(_Date) == 8);

		if (Write)
		{
			_Date d;

			d.Day = _Day;
			d.Month = _Month;
			d.Year = _Year;
			d.Hour = _Hours;
			d.Minute = _Minutes;
			d.ThouSec = (_Seconds * 1000) + _Thousands;

			return Props->Set(Name, &d, sizeof(d));
		}
		else // Read
		{
			void *Ptr;
			int Len;

			if (Props->Get(Name, Ptr, Len) AND
				sizeof(_Date) == Len)
			{
				_Date *d = (_Date*) Ptr;

				_Day = d->Day;
				_Month = d->Month;
				_Year = d->Year;
				_Hours = d->Hour;
				_Minutes = d->Minute;
				_Seconds = d->ThouSec / 1000;
				_Thousands = d->ThouSec % 1000;

				return true;
			}
		}
	}
	#endif

	return false;
}

int GDateTime::Compare(GDateTime *d)
{
	int c = 0;
	if (d)
	{
		c = _Year - d->_Year;
		if (NOT c)
		{
			c = _Month - d->_Month;
			if (NOT c)
			{
				c = _Day - d->_Day;
				if (NOT c)
				{
					c = _Hours - d->_Hours;
					if (NOT c)
					{
						c = _Minutes - d->_Minutes;
						if (NOT c)
						{
							c = _Seconds - d->_Seconds;
							if (NOT c)
							{
								c = _Thousands - d->_Thousands;
							}
						}
					}
				}
			}
		}
	}
	return c;
}

bool GDateTime::operator <(GDateTime &dt)
{
	if (_Year < dt._Year) return true;
	else if (_Year > dt._Year) return false;

	if (_Month < dt._Month) return true;
	else if (_Month > dt._Month) return false;

	if (_Day < dt._Day) return true;
	else if (_Day > dt._Day) return false;

	if (_Hours < dt._Hours) return true;
	else if (_Hours > dt._Hours) return false;

	if (_Minutes < dt._Minutes) return true;
	else if (_Minutes > dt._Minutes) return false;

	if (_Seconds < dt._Seconds) return true;
	else if (_Seconds > dt._Seconds) return false;

	if (_Thousands < dt._Thousands) return true;
	else if (_Thousands > dt._Thousands) return false;

	return false;
}

bool GDateTime::operator <=(GDateTime &dt)
{
	return NOT (*this > dt);
}

bool GDateTime::operator >(GDateTime &dt)
{
	if (_Year > dt._Year) return true;
	else if (_Year < dt._Year) return false;

	if (_Month > dt._Month) return true;
	else if (_Month < dt._Month) return false;

	if (_Day > dt._Day) return true;
	else if (_Day < dt._Day) return false;

	if (_Hours > dt._Hours) return true;
	else if (_Hours < dt._Hours) return false;

	if (_Minutes > dt._Minutes) return true;
	else if (_Minutes < dt._Minutes) return false;

	if (_Seconds > dt._Seconds) return true;
	else if (_Seconds < dt._Seconds) return false;

	if (_Thousands > dt._Thousands) return true;
	else if (_Thousands < dt._Thousands) return false;

	return false;
}

bool GDateTime::operator >=(GDateTime &dt)
{
	return NOT (*this < dt);
}

bool GDateTime::operator ==(GDateTime &dt)
{
	return	_Year == dt._Year AND
			_Month == dt._Month AND
			_Day == dt._Day AND
			_Hours == dt._Hours AND
			_Minutes == dt._Minutes AND
			_Seconds == dt._Seconds AND
			_Thousands == dt._Thousands;
}

bool GDateTime::operator !=(GDateTime &dt)
{
	return	_Year != dt._Year OR
			_Month != dt._Month OR
			_Day != dt._Day OR
			_Hours != dt._Hours OR
			_Minutes != dt._Minutes OR
			_Seconds != dt._Seconds OR
			_Thousands != dt._Thousands;
}

GDateTime GDateTime::DiffMonths(GDateTime &dt)
{
	GDateTime s;

	int Months = 0;
	s._Year = _Year;
	s._Month = _Month;
	s._Day = _Day;

	while (s._Year != dt._Year)
	{
		if (s._Year > dt._Year)
		{
			s._Year--;
			Months += 12;
		}
		else
		{
			Months -= 12;
			s._Year++;
		}
	}

	while (s._Month != dt._Month)
	{
		if
		(
			s._Month > dt._Month + 1
			OR
			(
				s._Month > dt._Month
				AND
				s._Day >= dt._Day
			)
		)
		{
			s._Month--;
			Months++;
		}
		else if
		(
			s._Month < dt._Month - 1
			OR
			(
				s._Month < dt._Month
				AND
				s._Day <= dt._Day
			)
		)
		{
			Months--;
			s._Month++;
		}
		else break;
	}

	s._Month = Months;
	s._Day = 0;
	s._Year = 0;

	return s;
}

GDateTime GDateTime::operator -(GDateTime &dt)
{
	GDateTime s;

	int Days = 0;
	s._Year = _Year;
	s._Month = _Month;
	s._Day = _Day;

	while (s._Year != dt._Year)
	{
		if (s._Year > dt._Year)
		{
			s._Year--;
			Days += s.IsLeapYear() ? 366 : 365;
		}
		else
		{
			Days -= s.IsLeapYear() ? 366 : 365;
			s._Year++;
		}
	}

	while (s._Month != dt._Month)
	{
		if (s._Month > dt._Month)
		{
			s._Month--;
			Days += s.DaysInMonth();
		}
		else
		{
			Days -= s.DaysInMonth();
			s._Month++;
		}
	}

	s._Month = 0;
	s._Day = Days + (_Day - dt._Day);

	s._Hours = _Hours - dt._Hours;
	s._Seconds = _Seconds - dt._Seconds;
	s._Thousands = _Thousands - dt._Thousands;

	return s;
}

GDateTime GDateTime::operator +(GDateTime &dt)
{
	GDateTime s = *this;

	s.AddMonths(dt.Month());
	s.AddDays(dt.Day());
	s.AddHours(dt.Hours());
	s.AddMinutes(dt.Minutes());
	// s.AddSeconds(dt.Seconds());

	return s;
}

bool GDateTime::IsSameDay(GDateTime &d)
{
	return	Day() == d.Day() AND
			Month() == d.Month() AND
			Year() == d.Year();
}

bool GDateTime::IsLeapYear(int Year)
{
	if (Year < 0) Year = _Year;

	if (Year % 4 != 0)
	{
		return false;
	}

	if (Year % 400 == 0)
	{
		return true;
	}

	if (Year % 100 == 0)
	{
		return false;
	}

	return true;
}

int GDateTime::DaysInMonth()
{
	if (_Month == 2 AND
		IsLeapYear())
	{
		return 29;
	}

	short DaysInMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	return _Month >= 1 AND _Month <= 12 ? DaysInMonth[_Month-1] : 0;
}

#define MinutesInDay (60*24)

void GDateTime::AddMinutes(int Minutes)
{
	int m = (int)_Minutes + Minutes;

	do
	{
		if (m < 0)
		{
			AddHours(-1);
			m += 60;
		}
		else if (m >= 60)
		{
			AddHours(1);
			m -= 60;
		}
		else
		{
			break;
		}
	}
	while (1);

	_Minutes = m;
}

void GDateTime::AddHours(int Hours)
{
	int h = _Hours + Hours;

	do
	{
		if (h < 0)
		{
			AddDays(-1);
			h += 24;
		}
		else if (h >= 24)
		{
			AddDays(1);
			h -= 24;
		}
		else
		{
			break;
		}
	}
	while (1);

	_Hours = h;
}

void GDateTime::AddDays(int Days)
{
	_Day += Days;

	do
	{
		if (_Day < 1)
		{
			AddMonths(-1);
			_Day += DaysInMonth();
		}
		else if (_Day > DaysInMonth())
		{
			_Day -= DaysInMonth();
			AddMonths(1);
		}
		else
		{
			break;
		}
	}
	while (1);
}

void GDateTime::AddMonths(int Months)
{
	int m = _Month + Months;

	do
	{
		if (m < 1)
		{
			_Year--;
			m += 12;
		}
		else if (m > 12)
		{
			_Year++;
			m -= 12;
		}
		else
		{
			break;
		}
	}
	while (1);

	_Month = m;
}


