#include <stdio.h>

#include "LgiNetInc.h"
#include "Lgi.h"
#include "GMime.h"
#include "GToken.h"
#include "Base64.h"

static char *MimeEol				= "\r\n";
static char *MimeWs					= " \t\r\n";
static char *MimeStr				= "\'\"";
static char *MimeQuotedPrintable	= "quoted-printable";
static char *MimeBase64				= "base64";

#define MimeMagic					( ('M'<<24) | ('I'<<24) | ('M'<<24) | ('E'<<24) )
#define SkipWs(s)					while (*s AND strchr(MimeWs, *s)) s++
#define SkipNonWs(s)				while (*s AND NOT strchr(MimeWs, *s)) s++

///////////////////////////////////////////////////////////////////////
enum MimeBoundary
{
	MimeData = 0,
	MimeNextSeg = 1,
	MimeEndSeg = 2
};

MimeBoundary IsMimeBoundary(char *Boundary, char *Line)
{
	if (Boundary)
	{
		int BoundaryLen = strlen(Boundary);
		if (Line AND
			*Line++ == '-' AND
			*Line++ == '-' AND
			strncmp(Line, Boundary, strlen(Boundary)) == 0)
		{
			// MIME segment boundary
			Line += BoundaryLen;
			if (Line[0] == '-' AND
				Line[1] == '-')
			{
				return MimeEndSeg;
			}
			else
			{
				return MimeNextSeg;
			}
		}
	}

	return MimeData;
}

void CreateMimeBoundary(char *Buf)
{
	if (Buf)
	{
		sprintf(Buf, "----=_NextPart_%08.8X.%08.8X", LgiCurrentTime(), LgiGetCurrentThread());
	}
}

///////////////////////////////////////////////////////////////////////
class GMimeStream : public GStream
{
protected:
	GStream *Out;

public:
	GMimeStream(GStream *o)
	{
		Out = o;
	}
};

class GMimeQuotedPrintableEncode : public GMimeStream
{
	char *d;
	char Buf[256];

public:
	GMimeQuotedPrintableEncode(GStream *o) : GMimeStream(o)
	{
		Buf[0] = 0;
		d = Buf;
	}

	int Write(void *p, int size, int f = 0)
	{
		char *s = (char*)p;
		char *e = s + size;
		while (s < e)
		{
			if (*s == '\n' OR NOT *s)
			{
				*d++ = '\r';
				*d++ = '\n';

				int Len = (int)d-(int)Buf;
				if (Out->Write(Buf, Len) < Len OR
					NOT *s)
				{
					break;
				}

				d = Buf;
				s++;
			}
			else if (*s & 0x80 OR
					 *s == '.' OR
					 *s == '=')
			{
				sprintf(d, "=%02.2X", (uchar)*s);
				d += 3;
				s++;
			}
			else if (*s != '\r')
			{
				*d++ = *s++;
			}
			else
			{
				s++;
			}

			if ((int)d-(int)Buf > 73)
			{
				// time for a new line.
				strcpy(d, "=\r\n");
				d += 3;

				int Len = (int)d-(int)Buf;
				if (Out->Write(Buf, Len) < Len)
				{
					break;
				}

				d = Buf;
			}
		}

		return (int)s-(int)p;
	}
};

class GMimeQuotedPrintableDecode : public GMimeStream
{
	GStringPipe Buf;

	char ConvHexToBin(char c)
	{
		if (c >= '0' AND c <= '9')
			return c - '0';
		if (c >= 'a' AND c <= 'f')
			return c + 10 - 'a';
		if (c >= 'A' AND c <= 'F')
			return c + 10 - 'A';
		return 0;
	}

public:
	GMimeQuotedPrintableDecode(GStream *o) : GMimeStream(o) {}

	int Write(void *p, int size, int f = 0)
	{
		Buf.Push((char*)p, size);

		char In[256];
		char Line[256];
		char *o = Line;
		while (Buf.Pop(In, sizeof(In)))
		{
			for (char *s=In; *s; )
			{
				if (*s == '=')
				{
					s++; // skip '='
					if (*s == '\r' OR *s == '\n')
					{
						if (*s == '\r')
							s++;
						if (*s == '\n')
							s++;
					}
					else
					{
						char c = ConvHexToBin(*s++);
						*o++ = (c << 4) | (ConvHexToBin(*s++) & 0xF);
					}
				}
				else
				{
					*o++ = *s++;
				}
			}

			int Len = (int)o-(int)Line;
			if (Out->Write(Line, Len) < Len)
			{
				// Error
				return 0;
			}
			o = Line;
		}

		return size;
	}
};

class GMimeBase64Encode : public GMimeStream
{
	GBytePipe Buf;

public:
	GMimeBase64Encode(GStream *o) : GMimeStream(o) {}
	
	~GMimeBase64Encode()
	{
		uchar b[100];

		int Len = Buf.GetSize();
		LgiAssert(Len < sizeof(b));
		int r = Buf.Read(b, Len);
		if (r)
		{
			char t[256];
			int w = ConvertBinaryToBase64(t, sizeof(t), b, r);
			Out->Write(t, w);
			Out->Write(MimeEol, 2);
		}
	}

	int Write(void *p, int size, int f = 0)
	{
		Buf.Write((uchar*)p, size);

		while (Buf.GetSize() >= 72)
		{
			uchar b[100];
			int r = Buf.Read(b, 72);
			if (r)
			{
				char t[256];
				int w = ConvertBinaryToBase64(t, sizeof(t), b, r);
				Out->Write(t, w);
				Out->Write(MimeEol, 2);
			}
			else return 0;
		}

		return size;
	}
};

class GMimeBase64Decode : public GMimeStream
{
	GStringPipe Buf;

public:
	GMimeBase64Decode(GStream *o) : GMimeStream(o) {}

	int Write(void *p, int size, int f = 0)
	{
		int Status = 0;

		// Push non whitespace into the memory buf
		char *s = (char*)p;
		char *e = s + size;
		while (s < e)
		{
			while (*s AND strchr(MimeWs, *s) AND s < e) s++;
			char *Start = s;
			while (*s AND NOT strchr(MimeWs, *s) AND s < e) s++;
			Buf.Push(Start, (int)s-(int)Start);
		}

		// While there is at least one run of base64 (4 bytes) convert it to text
		// and write it to the output stream
		int Size;
		while ((Size = Buf.GetSize()) > 3)
		{
			Size &= ~3;

			char t[256];
			int r = min(sizeof(t), Size);
			if ((r = Buf.GBytePipe::Read((uchar*)t, r)) > 0)
			{
				uchar b[256];
				int w = ConvertBase64ToBinary(b, sizeof(b), t, r);
				Out->Write(b, w);
				Status += w;
			}
		}

		return Status;
	}
};

///////////////////////////////////////////////////////////////////////
class GMimeBuf : public GStringPipe
{
	int Total;
	GStream *Src;
	GStreamEnd *End;

public:
	GMimeBuf(GStream *src, GStreamEnd *end)
	{
		Total = 0;
		Src = src;
		End = end;
	}

	int Pop(char *Str, int BufSize)
	{
		int Ret = 0;

		while (NOT (Ret = GStringPipe::Pop(Str, BufSize)))
		{
			if (Src)
			{
				char Buf[1024];
				int r = Src ? Src->Read(Buf, sizeof(Buf)) : 0;
				if (r)
				{
					if (End)
					{
						int e = End->IsEnd(Buf, r);
						if (e >= 0)
						{
							// End of stream
							int s = e - Total;
							Push(Buf, s);
							Total += s;
							Src = 0; // no more data anyway
						}
						else
						{
							// Not the end
							Push(Buf, r);
							Total += r;
						}
					}
					else
					{
						Push(Buf, r);
						Total += r;
					}
				}
				else break;
			}
			else break;
		}

		return Ret;
	}
};

/// Opens a temporary file, which is deleted after it's closed.
class GTempFile : public GFile
{
	char *File;

public:
	GTempFile()
	{
		char Path[256];
		LgiGetTempPath(Path, sizeof(Path));

		char FileName[32];
		sprintf(FileName, "~mime-%X.tmp", LgiRand());
		LgiMakePath(Path, sizeof(Path), Path, FileName);

		File = NewStr(Path);
		Open(Path, O_READWRITE);
	}

	~GTempFile()
	{
		Close();

		if (File)
		{
			FileDev->DeleteFile(File, false);
			DeleteArray(File);
		}
	}
};

///////////////////////////////////////////////////////////////////////
// Mime Object
GMime::GMime()
{
	Headers = 0;
	DataPos = 0;
	DataSize = 0;
	DataLock = 0;
	DataStore = 0;
	OwnDataStore = 0;

	Text.Decode.Mime = this;
	Text.Encode.Mime = this;
	Binary.Read.Mime = this;
	Binary.Write.Mime = this;
}

GMime::~GMime()
{
	Empty();
}

bool GMime::Lock()
{
	bool Lock = true;
	if (DataLock)
	{
		Lock = DataLock->Lock();
	}
	return Lock;
}

void GMime::Unlock()
{
	if (DataLock)
	{
		DataLock->Unlock();
	}
}

bool GMime::CreateTempData()
{
	bool Status = false;

	char Path[256];
	if (LgiGetTempPath(Path, sizeof(Path)))
	{
		DataPos = 0;
		DataSize = 0;
		OwnDataStore = true;
		DataStore = NEW(GTempFile);
		if (DataStore)
		{
			Status = DataStore->IsOpen();
		}
	}

	return Status;
}

GStream *GMime::GetData()
{
	if (DataStore)
	{
		DataStore->SetPos(DataPos);
	}

	return DataStore;
}

bool GMime::SetData(GStream *d, int Pos, int Size, GSemaphore *l)
{
	if (DataStore AND Lock())
	{
		DeleteObj(DataStore);
		Unlock();
	}

	if (d)
	{
		DataPos = Pos;
		DataSize = Size >= 0 ? Size : d->GetSize();
		DataLock = l;
		DataStore = d;
	}

	return true;
}

bool GMime::SetData(char *Str, int Len)
{
	if (DataStore AND Lock())
	{
		DeleteObj(DataStore);
		Unlock();
	}

	if (Str)
	{
		if (Len < 0)
		{
			Len = strlen(Str);
		}

		DataLock = 0;
		DataPos = 0;
		DataSize = Len;
		DataStore = NEW(GTempFile);
		if (DataStore)
		{
			DataStore->Write(Str, Len);
		}
	}

	return true;
}

void GMime::Empty()
{
	if (OwnDataStore)
	{
		if (Lock())
		{
			DeleteObj(DataStore);
			Unlock();
		}
	}

	Children.DeleteObjects();
	DeleteArray(Headers);

	DataPos = 0;
	DataSize = 0;
	DataLock = 0;
	DataStore = 0;
	OwnDataStore = 0;
}

char *GMime::NewValue(char *&s, bool Alloc)
{
	char *Status = 0;

	int Inc = 0;
	char *End;
	if (strchr(MimeStr, *s))
	{
		// Delimited string
		char Delim = *s++;
		End = strchr(s, Delim);
		Inc = 1;
	}
	else
	{
		// Raw string
		End = s;
		while (*End AND *End != ';' AND *End != '\n' AND *End != '\r') End++;
		while (strchr(MimeWs, End[-1])) End--;
	}

	if (End)
	{
		if (Alloc)
		{
			Status = NewStr(s, (int)End-(int)s);
		}
		s = End + Inc;
	}

	SkipWs(s);

	return Status;
}

char *GMime::StartOfField(char *s, char *Field)
{
	char *Status = 0;

	if (s AND Field)
	{
		int FieldLen = strlen(Field);
		while (s AND *s)
		{
			if (strchr(MimeWs, *s))
			{
				s = strchr(s, '\n');
				if (s) s++;
			}
			else
			{
				char *f = s;
				while (*s AND *s != ':' AND NOT strchr(MimeWs, *s)) s++;
				int fLen = (int)s-(int)f;
				if (*s++ == ':' AND
					fLen == FieldLen AND
					strnicmp(f, Field, FieldLen) == 0)
				{
					return f;
					break;
				}
				else
				{
					s = strchr(s, '\n');
					if (s) s++;
				}
			}
		}
	}

	return 0;
}

char *GMime::NextField(char *s)
{
	while (s)
	{
		while (*s AND *s != '\n') s++;
		if (*s == '\n')
		{
			s++;
			if (NOT strchr(MimeWs, *s))
			{
				break;
			}
		}
		else
		{
			break;
		}
	}

	return s;
}

char *GMime::Get(char *Name, bool Short, char *Default)
{
	char *Status = 0;

	if (Name AND Headers)
	{
		char *s = StartOfField(Headers, Name);
		if (s)
		{
			s = strchr(s, ':');
			if (s)
			{
				s++;
				SkipWs(s);
				if (Short)
				{
					Status = NewValue(s);
				}
				else
				{
					char *e = NextField(s);
					while (strchr(MimeWs, e[-1])) e--;
					Status = NewStr(s, (int)e-(int)s);
				}
			}
		}

		if (NOT Status AND Default)
		{
			Status = NewStr(Default);
		}
	}

	return Status;
}

bool GMime::Set(char *Name, char *Value)
{
	bool Status = false;
	
	if (Name)
	{
		GStringPipe p;

		char *h = Headers;
		if (h)
		{
			char *f = StartOfField(h, Name);
			if (f)
			{
				// 'Name' exists, push out pre 'Name' header text
				p.Push(h, (int)f-(int)Headers);
				h = NextField(f);
			}
			else
			{
				// 'Name' doesn't exist, push out all the headers
				p.Push(Headers);
				h = 0;
			}
		}

		if (Value)
		{
			// Push new field
			int Vlen = strlen(Value);
			while (Vlen > 0 AND strchr(MimeWs, Value[Vlen-1])) Vlen--;

			p.Push(Name);
			p.Push(": ");
			p.Push(Value, Vlen);
			p.Push(MimeEol);
		}
		// else we're deleting the feild

		if (h)
		{
			// Push out any header text post the 'Name' field.
			p.Push(h);
		}

		DeleteArray(Headers);
		Headers = (char*)p.New(sizeof(char));
	}

	return Status;
}

char *GMime::GetSub(char *Field, char *Sub)
{
	char *Status = 0;

	if (Field AND Sub)
	{
		int SubLen = strlen(Sub);
		char *v = Get(Field, false);
		if (v)
		{
			// Move past the field value into the sub fields
			char *s = v;
			SkipWs(s);
			while (*s AND *s != ';' AND NOT strchr(MimeWs, *s)) s++;

			while (s AND *s++ == ';')
			{
				// Parse each name=value pair
				SkipWs(s);
				char *Name = s;
				while (*s AND *s != '=' AND NOT strchr(MimeWs, *s)) s++;
				int NameLen = (int)s-(int)Name;
				SkipWs(s);
				if (*s++ == '=')
				{
					bool Found = SubLen == NameLen AND strnicmp(Name, Sub, NameLen) == 0;
					Status = NewValue(s, Found);
					if (Found) break;
				}
				else break;
			}

			DeleteArray(v);
		}
	}

	return Status;
}

bool GMime::SetSub(char *Field, char *Sub, char *Value, char *DefaultValue)
{
	if (Field AND Sub)
	{
		char Buf[256];

		char *s = StartOfField(Headers, Field);
		if (s)
		{
			// Header already exists
			s = strchr(s, ':');
			if (s++)
			{
				SkipWs(s);

				GStringPipe p;

				// Push the field data
				char *e = s;
				while (*e AND NOT strchr("; \t\r\n", *e)) e++;
				p.Push(s, (int)e-(int)s);
				SkipWs(e);

				// Loop through the subfields and push all those that are not 'Sub'
				s = e;
				while (*s++ == ';')
				{
					SkipWs(s);
					char *e = s;
					while (*e AND *e != '=' AND NOT strchr(MimeWs, *e)) e++;
					char *Name = NewStr(s, (int)e-(int)s);
					if (Name)
					{
						s = e;
						SkipWs(s);
						if (*s++ == '=')
						{
							char *v = NewValue(s);
							if (stricmp(Name, Sub) != 0)
							{
								sprintf(Buf, ";\r\n\t%s=\"%s\"", Name, v);
								p.Push(Buf);
							}
							DeleteArray(v);
						}
						else break;
					}
					else break;
				}

				if (Value)
				{
					// Push the new sub field
					sprintf(Buf, ";\r\n\t%s=\"%s\"", Sub, Value);
					p.Push(Buf);
				}

				char *Data = p.NewStr();
				if (Data)
				{
					Set(Field, Data);
					DeleteArray(Data);
				}
			}
		}
		else if (DefaultValue)
		{
			// Header doesn't exist at all
			if (Value)
			{
				// Set
				sprintf(Buf, "%s;\r\n\t%s=\"%s\"", DefaultValue, Sub, Value);
				return Set(Field, Buf);
			}
			else
			{
				// Remove
				return Set(Field, DefaultValue);
			}
		}
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////
// Mime Text Conversion

// Rfc822 Text -> Object
int GMime::GMimeText::GMimeDecode::Pull(GStream *Source, GStreamEnd *End)
{
	GMimeBuf Buf(Source, End); // Stream -> Lines
	return Parse(&Buf);
}

class ParentState
{
public:
	char *Boundary;
	MimeBoundary Type;

	ParentState()
	{
		Boundary = 0;
		Type = MimeData;
	}
};

int GMime::GMimeText::GMimeDecode::Parse(GStringPipe *Source, ParentState *State)
{
	int Status = 0;

	if (Mime AND Source)
	{
		Mime->Empty();

		if (Mime->CreateTempData())
		{
			// Read the headers..
			char Buf[1024];
			GStringPipe HeaderBuf;
			int r;
			while (r = Source->Pop(Buf, sizeof(Buf)))
			{
				if (NOT strchr(MimeEol, Buf[0]))
				{
					// Store part of the headers
					HeaderBuf.Push(Buf, r);
				}
				else break;
			}

			// Not an error
			Mime->Headers = HeaderBuf.NewStr();
			if (Mime->Headers)
			{
				// Get various bits out of the header
				char *Encoding = Mime->GetEncoding();
				char *Boundary = Mime->GetBoundary();
				int BoundaryLen = Boundary ? strlen(Boundary) : 0;
				GStream *Decoder = 0;
				if (Encoding)
				{
					if (stricmp(Encoding, MimeQuotedPrintable) == 0)
					{
						Decoder = NEW(GMimeQuotedPrintableDecode(Mime->DataStore));
					}
					else if (stricmp(Encoding, MimeBase64) == 0)
					{
						Decoder = NEW(GMimeBase64Decode(Mime->DataStore));
					}
				}

				// Read in the rest of the MIME segment
				bool Done = false;
				int StartPos = Mime->DataStore->GetPos();
				while (NOT Done)
				{
					// Process existing lines
					int Len;
					char Buf[1024];
					int Written = 0;
					while (Len = Source->Pop(Buf, sizeof(Buf)))
					{
						// Check for boundary
						MimeBoundary Type = MimeData;
						if (Boundary)
						{
							 Type = IsMimeBoundary(Boundary, Buf);
						}
						else if (State)
						{
							State->Type = IsMimeBoundary(State->Boundary, Buf);
							if (State->Type)
							{
								Status = Done = true;
								break;
							}
						}

						DoSegment:
						if (Type == MimeNextSeg)
						{
							ParentState MyState;
							MyState.Boundary = Boundary;

							GMime *Seg = NEW(GMime);
							if (Seg AND
								Seg->Text.Decode.Parse(Source, &MyState))
							{
								Mime->Children.Insert(Seg);

								if (MyState.Type)
								{
									Type = MyState.Type;
									goto DoSegment;
								}
							}
							else break;
						}
						else if (Type == MimeEndSeg)
						{
							Status = Done = true;
							break;
						}
						else
						{
							// Process data
							if (Decoder)
							{
								Written += Decoder->Write(Buf, Len);
							}
							else
							{
								int w = Mime->DataStore->Write(Buf, Len);
								if (w > 0)
								{
									Written += w;
								}
								else
								{
									Done = true;
									break;
								}
							}
							Status = true;
						}
					}

					if (Len == 0)
					{
						Done = true;
					}
				}

				DeleteObj(Decoder);

				Mime->DataSize = Mime->DataStore->GetPos() - StartPos;
			}
		}
	}

	return Status;
}

void GMime::GMimeText::GMimeDecode::Empty()
{
}

// Object -> Rfc822 Text
int GMime::GMimeText::GMimeEncode::Push(GStream *Dest, GStreamEnd *End)
{
	int Status = 0;

	if (Mime)
	{
		char Buf[1024];

		// Check boundary
		char *Boundary = Mime->GetBoundary();
		if (Mime->Children.First())
		{
			// Boundary required
			if (NOT Boundary)
			{
				// Create one
				char b[256];
				CreateMimeBoundary(b);
				Mime->SetBoundary(b);
				Boundary = Mime->GetBoundary();
			}
		}
		else if (Boundary)
		{
			// Remove boundary
			Mime->SetBoundary(0);
			DeleteArray(Boundary);
		}

		// Check encoding
		char *Encoding = Mime->GetEncoding();
		if (NOT Encoding)
		{
			// Detect an appropriate encoding
			int MaxLine = 0;
			bool Has8Bit = false;
			bool HasBin = false;
			if (Mime->DataStore AND
				Mime->Lock())
			{
				Mime->DataStore->SetPos(Mime->DataPos);
				int x = 0;
				for (int i=0; i<Mime->DataSize; )
				{
					int m = min(Mime->DataSize - i, sizeof(Buf));
					int r = Mime->DataStore->Read(Buf, m);
					if (r > 0)
					{
						for (int n=0; n<r; n++)
						{
							if (Buf[n] == '\n')
							{
								MaxLine = max(x, MaxLine);
								x = 0;
							}
							else
							{
								x++;
							}

							if (Buf[n] & 0x80)
							{
								Has8Bit = true;
							}
							if (Buf[n] < ' ' AND
								Buf[n] != '\t' AND
								Buf[n] != '\r' AND
								Buf[n] != '\n')
							{
								HasBin = true;
							}
						}

						i += r;
					}
					else break;
				}
				
				Mime->Unlock();			
			}

			if (HasBin)
			{
				Encoding = NewStr(MimeBase64);
			}
			else if (Has8Bit OR MaxLine > 70)
			{
				Encoding = NewStr(MimeQuotedPrintable);
			}

			if (Encoding)
			{
				Mime->SetEncoding(Encoding);
			}
		}

		// Write the headers
		GToken h(Mime->Headers, MimeEol);
		for (int i=0; i<h.Length(); i++)
		{
			Dest->Write(h[i], strlen(h[i]));
			Dest->Write(MimeEol, 2);
		}
		Dest->Write(MimeEol, 2);

		// Write data
		GStream *Encoder = 0;
		if (Encoding)
		{
			if (stricmp(Encoding, MimeQuotedPrintable) == 0)
			{
				Encoder = NEW(GMimeQuotedPrintableEncode(Dest));
			}
			else if (stricmp(Encoding, MimeBase64) == 0)
			{
				Encoder = NEW(GMimeBase64Encode(Dest));
			}
		}

		if (Mime->DataStore)
		{
			if (Mime->Lock())
			{
				Mime->DataStore->SetPos(Mime->DataPos);
				Status = Mime->DataSize == 0; // Nothing is a valid segment??
				for (int i=0; i<Mime->DataSize; )
				{
					int m = min(Mime->DataSize-i, sizeof(Buf));
					int r = Mime->DataStore->Read(Buf, m);
					if (r > 0)
					{
						if (Encoder)
						{
							Encoder->Write(Buf, r);
						}
						else
						{
							Dest->Write(Buf, r);
						}
						Status = true;
					}
					else break;
				}
				Mime->Unlock();
			}
		}
		else
		{
			Status = true;
		}

		DeleteObj(Encoder);

		// Write children
		if (Mime->Children.First() AND Boundary)
		{
			for (GMime *c = Mime->Children.First(); c; c = Mime->Children.Next())
			{
				sprintf(Buf, "\r\n\r\n--%s\r\n", Boundary);
				Dest->Write(Buf, strlen(Buf));

				if (NOT c->Text.Encode.Push(Dest, End))
				{
					break;
				}

				Status = 1;
			}

			sprintf(Buf, "\r\n\r\n--%s--\r\n", Boundary);
			Dest->Write(Buf, strlen(Buf));
		}

		// Clean up
		DeleteArray(Encoding);
		DeleteArray(Boundary);
	}

	return Status;
}

void GMime::GMimeText::GMimeEncode::Empty()
{
}

/////////////////////////////////////////////////////////////////////////
// Mime Binary Serialization

// Source -> Object
int GMime::GMimeBinary::GMimeRead::Pull(GStream *Source, GStreamEnd *End)
{
	if (Source)
	{
		int32 Header[4];
		Mime->Empty();

		// Read header block (Magic, HeaderSize, DataSize, # of Children)
		// and check magic
		if (Source->Read(Header, sizeof(Header)) == sizeof(Header) AND
			Header[0] == MimeMagic)
		{
			// Read header data
			Mime->Headers = NEW(char[Header[1]+1]);
			if (Mime->Headers AND
				Source->Read(Mime->Headers, Header[1]) == Header[1])
			{
				// NUL terminate
				Mime->Headers[Header[1]] = 0;

				// Skip body data
				if (Source->SetPos(Source->GetPos() + Header[2]) > 0)
				{
					// Read the children in
					for (int i=0; i<Header[3]; i++)
					{
						GMime *c = NEW(GMime);
						if (c AND
							c->Binary.Read.Pull(Source, End))
						{
							Mime->Children.Insert(c);
						}
						else break;
					}
				}
			}

			return 1; // success
		}
	}

	return 0; // failure
}

void GMime::GMimeBinary::GMimeRead::Empty()
{
}

// Object -> Dest
int64 GMime::GMimeBinary::GMimeWrite::GetSize()
{
	int Size = 0;

	if (Mime)
	{
		Size =	(sizeof(int32) * 4) + // Header magic + block sizes
				(Mime->Headers ? strlen(Mime->Headers) : 0) + // Headers
				(Mime->DataStore ? Mime->DataSize : 0); // Data

		// Children
		Iterator<GMime> It(&Mime->Children);
		for (GMime *c=It.First(); c; c=It.Next())
		{
			Size += c->Binary.Write.GetSize();
		}
	}

	return Size;
}

int GMime::GMimeBinary::GMimeWrite::Push(GStream *Dest, GStreamEnd *End)
{
	if (Dest AND Mime)
	{
		int32 Header[4] =
		{
			MimeMagic,
			Mime->Headers ? strlen(Mime->Headers) : 0,
			Mime->DataStore ? Mime->DataSize : 0,
			Mime->Children.GetItems()
		};

		if (Dest->Write(Header, sizeof(Header)) == sizeof(Header))
		{
			if (Mime->Headers)
			{
				Dest->Write(Mime->Headers, Header[1]);
			}

			if (Mime->DataStore)
			{
				char Buf[1024];
				int Written = 0;
				int Read = 0;
				int r;
				while ((r = Mime->DataStore->Read(Buf, min(sizeof(Buf), Header[2]-Read) )) > 0)
				{
					int w;
					if ((w = Dest->Write(Buf, r)) <= 0)
					{
						// Write error
						break;
					}

					Written += w;
					Read += r;
				}

				// Check we've written out all the data
				LgiAssert(Written < Header[2]);
				if (Written < Header[2])
				{
					return 0;
				}
			}

			Iterator<GMime> It(&Mime->Children);
			for (GMime *c=It.First(); c; c=It.Next())
			{
				c->Binary.Write.Push(Dest, End);
			}

			return 1;
		}
	}

	return 0;
}

void GMime::GMimeBinary::GMimeWrite::Empty()
{
}

