#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include "Gdc2.h"
#include "IFtp.h"
#include "GToken.h"
#include "GString.h"

///////////////////////////////////////////////////////////////////
enum FtpTransferMode
{
	FT_Unknown,
	FT_Ascii,
	FT_Binary
};

class IFtpPrivate
{
public:
	char OutBuf[256];
	char InBuf[256];
	List<char> In;
	FtpTransferMode CurMode;
	GSocket *Listen;	// listen for data
	GSocket *Data;		// get data
	GFile *F;

	IFtpPrivate()
	{
		InBuf[0] = 0;
		CurMode = FT_Unknown;
		Data = Listen = 0;
		F = 0;
	}
};

///////////////////////////////////////////////////////////////////
int LookupMonth(char *m)
{
	char *Months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	for (int i=0; i<12; i++)
	{
		if (m AND stricmp(m, Months[i]) == 0)
		{
			return i + 1;
		}
	}
	return 1;
}

///////////////////////////////////////////////////////////////////
class ILogProxy : public GSocketImpl
{
	GSocket *Dest;
	IFtpDataPipe *Pipe;
	Progress *Meter;

public:
	ILogProxy(GSocket *dest, IFtpDataPipe *pipe, Progress *meter)
	{
		Dest = dest;
		Pipe = pipe;
		Meter = meter;
	}

	int Write(char *Data, int Len, int Flags)
	{
		if (Pipe)
		{
			return Pipe->WritePipe(this, Data, Len, Flags);
		}

		return GSocketImpl::Write(Data, Len, Flags);
	}

	int Read(char *Data, int Len, int Flags)
	{
		if (Pipe)
		{
			return Pipe->ReadPipe(this, Data, Len, Flags);
		}

		return GSocketImpl::Read(Data, Len, Flags);
	}

	void OnError(int ErrorCode, char *ErrorDescription)
	{
		if (Dest AND
			ErrorCode > 0 AND
			ErrorDescription)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", ErrorDescription);
			Dest->OnInformation(Str);
		}
	}

	void OnInformation(char *s)
	{
		if (Dest AND s)
		{
			char Str[256];
			sprintf(Str, "[Data] %s", s);
			Dest->OnInformation(Str);
		}
	}

	void OnDisconnect()
	{
		if (Dest)
		{
			Dest->OnInformation("[Data] Disconnect");
		}
	}

	void OnRead(char *Data, int Len)
	{
		if (Meter)
		{
			Meter->Value(Meter->Value() + Len);
		}
	}

	void OnWrite(char *Data, int Len)
	{
		if (Meter)
		{
			Meter->Value(Meter->Value() + Len);
		}
	}
};

///////////////////////////////////////////////////////////////////
IFtpEntry::IFtpEntry()
{
	Attributes = 0;
	Size = 0;
	Name = 0;
	Path = 0;
	User = Group = 0;
}

IFtpEntry::IFtpEntry(IFtpEntry *Entry)
{
	Attributes = Entry->Attributes;
	Size = Entry->Size;
	Name = NewStr(Entry->Name);
	Path = NewStr(Entry->Path);
	Date = Entry->Date;
	User = NewStr(Entry->User);
	Group = NewStr(Entry->Group);
}

IFtpEntry::IFtpEntry(char *Entry)
{
	Attributes = 0;
	Size = 0;
	Name = 0;
	Path = 0;
	User = 0;
	Group = 0;
	if (Entry)
	{
		char *Ws = " \t";
		GToken T(Entry, Ws);
		if (T.Length() > 0)
		{
			char *_Size = 0, *_Perm = 0;
			int _Name = 0;

			if (isdigit(*T[0]))
			{
				// M$ format
				_Size = T[2];
				if (stricmp(_Size, "<dir>") == 0)
				{
					Attributes |= IFTP_DIR;
				}

				_Name = 3;
				Date.SetDate(T[0]);
				Date.SetTime(T[1]);
			}
			else if (T.Length() > 7)
			{
				// Unix format
				int SizeElement = (isdigit(*T[4])) ? 4 : 3;
				_Size = T[SizeElement];
				_Name = SizeElement + 4;
				_Perm = T[0];
				User = NewStr(T[SizeElement-2]);
				Group = NewStr(T[SizeElement-1]);

				char *MonthStr = T[SizeElement+1];
				char *YearOrTime = T[SizeElement+3];
				if (MonthStr AND YearOrTime)
				{
					int Day = atoi(T[SizeElement+2]);
					int Month = LookupMonth(MonthStr);
					int Year;

					if (strchr(YearOrTime, ':'))
					{
						// It's time
						Date.SetTime(YearOrTime);

						// Default the year
						time_t Time = time(NULL);
						tm *ptm = localtime(&Time);
						if (ptm)
						{
							Year = 1900 + ptm->tm_year;
						}
					}
					else 
					{
						// It's a year
						Year = atoi(YearOrTime);
					}

					char DateStr[64];
					sprintf(DateStr, "%i/%i/%i", Day, Month, Year);
					Date.SetDate(DateStr);
				}
			}

			if (_Name)
			{
				char *n = Entry;
				for (int i=0; i<_Name; i++)
				{
					for (; *n AND NOT strchr(Ws, *n); n++);
					for (; *n AND strchr(Ws, *n); n++);
				}
				Name = NewStr(n);
				if (Name)
				{
					if (Name[0] == '.')
					{
						Attributes |= IFTP_HIDDEN;
					}
				}
			}

			if (_Perm AND strlen(_Perm) == 10)
			{
				if (_Perm[0] == 'd' OR
					_Perm[0] == 'l') Attributes |= IFTP_DIR;
				if (_Perm[1] != '-') Attributes |= IFTP_READ;
				if (_Perm[2] != '-') Attributes |= IFTP_WRITE;
				if (_Perm[3] != '-') Attributes |= IFTP_EXECUTE;

				if (_Perm[4] != '-') Attributes |= IFTP_GRP_READ;
				if (_Perm[5] != '-') Attributes |= IFTP_GRP_WRITE;
				if (_Perm[6] != '-') Attributes |= IFTP_GRP_EXECUTE;

				if (_Perm[7] != '-') Attributes |= IFTP_GLOB_READ;
				if (_Perm[8] != '-') Attributes |= IFTP_GLOB_WRITE;
				if (_Perm[9] != '-') Attributes |= IFTP_GLOB_EXECUTE;

				if (_Perm[0] == 'l')
				{
					// link...
					Attributes |= IFTP_SYM_LINK;
					
					char *Arrow = Name ? strstr(Name, "->") : 0;
					if (Arrow)
					{
						Arrow[-1] = 0;
					}
				}
			}

			if (_Size)
			{
				#ifdef WIN32
				Size = _atoi64(_Size);
				#else
				Size = atoll(_Size);
				#endif
			}
		}
	}
}

IFtpEntry::~IFtpEntry()
{
	DeleteArray(Name);
	DeleteArray(Path);
	DeleteArray(User);
	DeleteArray(Group);
}

///////////////////////////////////////////////////////////////////

// #define VERIFY(i, ret) if (i != ret) { Close(); return 0; }
// #define VERIFY_RANGE(i, range) if (((i)/100) != range) { Close(); return 0; }

#define Verify(i, ret) { int Code = i; if (Code != ret) throw Code; }
#define VerifyRange(i, range) { int Code = i; if ((Code/100) != range) throw Code; }

IFtp::IFtp(IFtpDataPipe *pipe)
{
	d = NEW(IFtpPrivate);
	Pipe = pipe;
	Authenticated = false;
	Meter = 0;
	Socket = 0;

	Ip[0] = 0;
	Port = 0;
	PassiveMode = false;
	ForceActive = false;
	LongList = true;
	ShowHidden = false;

	RestorePos = 0;
	AbortTransfer = false;
}

IFtp::~IFtp()
{
	DeleteObj(d);
}

int IFtp::WriteLine(char *Msg)
{
	int Status = 0;

	if (IsOpen())
	{
		char *b = Msg ? Msg : d->OutBuf;
		int l = strlen(b);
		Status = Socket->Write(b, l, 0);
	}

	return Status;	
}

int IFtp::ReadLine(char *Msg, bool Debug)
{
	int Status = 0;
	while (Socket)
	{
		// look through the input list for result codes
		char *s;
		while (s = d->In.First())
		{
			int i = 0;
			if (isdigit(*s))
			{
				char *e = s;
				while (isdigit(*e)) e++;

				if (*e != '-')
				{
					i = atoi(s);

					if (Msg)
					{
						char *Sp = strchr(s, ' ');
						if (Sp)
						{
							strcpy(Msg, Sp + 1);
						}
					}
				}
			}

			d->In.Delete(s);
			DeleteArray(s);
			
			if (i)
			{
				return i;
			}
		}

		// Ok, no result code in the input list so
		// Read data off socket
		int Len = strlen(d->InBuf);
		int Read = 0;

		if (IsOpen())
		{
			Read = Socket->Read(d->InBuf+Len, sizeof(d->InBuf)-Len-1, 0);
		}

		if (Read > 0)
		{
			// process input into lines
			char *Eol = 0;
			d->InBuf[Len+Read] = 0;
			do
			{
				Eol = strstr(d->InBuf, "\r\n");
				if (Eol)
				{
					// copy line into input list
					char *New = 0;
					d->In.Insert(New = NewStr(d->InBuf, (int)Eol-(int)d->InBuf));
					Eol += 2;

					// move data after line up to the start of the buffer
					int EolLen = strlen(Eol);
					if (EolLen > 0)
					{
						memmove(d->InBuf, Eol, EolLen+1);
					}
					else
					{
						d->InBuf[0] = 0;
						Eol = 0;
					}
				}
			}
			while (Eol);
		}
		else break; // the read failed....

	} // loop around to check results

	return Status;
}

bool IFtp::IsOpen()
{
	return Socket ? Socket->IsOpen() : false;
}

FtpOpenStatus IFtp::Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password)
{
	FtpOpenStatus Status = FO_ConnectFailed;

	try
	{
		Socket = S;
		Authenticated = false;	

		if (NOT Port) Port = 21;

		if (Socket AND
			Socket->Open(RemoteHost, Port))
		{
			Verify(ReadLine(), 220);

			if (User AND Password)
			{
				// User/pass
				sprintf(d->OutBuf, "USER %s\r\n", User);
				if (WriteLine())
				{
					Verify(ReadLine(), 331);

					sprintf(d->OutBuf, "PASS %s\r\n", Password);
					if (WriteLine())
					{
						Verify(ReadLine(), 230);
						Authenticated = true;
					}
				}
			}
			else
			{
				// Anonymous
				sprintf(d->OutBuf, "USER anonymous\r\n");
				if (WriteLine())
				{
					Verify(ReadLine(), 331);

					sprintf(d->OutBuf, "PASS me@privacy.net\r\n"); // garrenteed to never
															// be allocated to an IP
					if (WriteLine())
					{
						Verify(ReadLine(), 230);
						Authenticated = true;
					}
				}
			}

			if (Authenticated)
			{
				Status = FO_Connected;
			}
			else
			{
				Status = FO_LoginFailed;
			}
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		Status = FO_Error;
	}

	return Status;
}

bool IFtp::Close()
{
	if (d->F)
	{
		d->F->Close();
	}

	if (Socket)
	{
		Socket->Close();
		Socket = 0;
	}

	if (d->Data)
	{
		d->Data->Close();
		d->Data = 0;
	}

	if (d->Listen)
	{
		d->Listen->Close();
		d->Listen = 0;
	}

	return true;
}

void IFtp::Noop()
{
	if (IsOpen())
	{
		sprintf(d->OutBuf, "NOOP\r\n");
		if (WriteLine())
		{
			ReadLine();
		}
	}
}

bool IFtp::GetDir(char *Dir)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND
			Dir)
		{
			char Temp[256];

			sprintf(d->OutBuf, "PWD\r\n");
			WriteLine();
			Verify(ReadLine(Temp), 257);

			char *Start = strchr(Temp, '\"');
			char *End = (Start) ? strchr(Start+1, '\"') : 0;
			if (End)
			{
				*End = 0;
				strcpy(Dir, Start+1);
			}
			
			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::SetDir(char *Dir)
{
	bool Status = false;

	try
	{
		if (Dir AND IsOpen())
		{
			sprintf(d->OutBuf, "CWD %s\r\n", Dir);
			WriteLine();
			Verify(ReadLine(), 250);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
	}

	return Status;
}

bool IFtp::CreateDir(char *Dir)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND Dir)
		{
			sprintf(d->OutBuf, "MKD %s\r\n", Dir);
			WriteLine();
			VerifyRange(ReadLine(), 2);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::DeleteDir(char *Dir)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND Dir)
		{
			sprintf(d->OutBuf, "RMD %s\r\n", Dir);
			WriteLine();
			VerifyRange(ReadLine(), 2);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::UpDir()
{
	bool Status = false;

	try
	{
		if (IsOpen())
		{
			sprintf(d->OutBuf, "CDUP\r\n");
			WriteLine();
			VerifyRange(ReadLine(0, true), 2);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

// #include "LgiCommon.h"
// #include <malloc.h>
#ifdef _INC_MALLOC
#define IsHeapOk() (_heapchk() != _HEAPOK)
#else
#define IsHeapOk() (-1)
#endif

bool IFtp::ListDir(List<IFtpEntry> *Dir)
{
	bool Status = false;

	try
	{
		if (Dir AND
			IsOpen() AND
			SetupData(true))
		{
			GBytePipe Buf;

			// List command
			strcpy(d->OutBuf, "LIST");
			if (LongList OR ShowHidden) strcat(d->OutBuf, " -");
			if (LongList) strcat(d->OutBuf, "l");
			if (ShowHidden) strcat(d->OutBuf, "a");
			strcat(d->OutBuf, "\r\n");
			WriteLine();
				
			// Accept the data connection
			if (ConnectData())
			{
				VerifyRange(ReadLine(), 1);

				if (Meter)
				{
					Meter->SetLimits(0, 0);
					Meter->Value(0);
				}

				// Read the data
				uchar Buffer[1024];
				int Len;
				
				while ((Len = d->Data->Read((char*) Buffer, sizeof(Buffer), 0)) > 0)
				{
					if (Meter)
					{
						Meter->Value(Buf.GetSize());
					}

					Buf.Write(Buffer, Len);
				}

				if (Meter)
				{
					Meter->Value(0);
				}
				
				Status = true;
			}
			
			DeleteObj(d->Data);
			DeleteObj(d->Listen);
			
			// Parse the results
			int Len = Buf.GetSize();
			if (Status AND Len)
			{
				char *Text = NEW(char[Len+1]);
				if (Text AND Buf.Read((uchar*)Text, Len))
				{
					Text[Len] = 0;

					#ifdef LOG_LISTINGS
					GFile F;
					if (F.Open("c:\\temp\\ftplog.txt", O_WRITE))
					{
						F.Write(Text, Len);
						F.Close();
					}
					#endif

					GToken Lines(Text, "\r\n");
					for (int i=0; i<Lines.Length(); i++)
					{
						IFtpEntry *e = NEW(IFtpEntry(Lines[i]));
						if (e AND e->Name)
						{
							if (strcmp(e->Name, ".") != 0 AND
								strcmp(e->Name, "..") != 0)
							{
								Dir->Insert(e);
								e = 0;
							}
						}
						
						DeleteObj(e);
					}
					
					DeleteArray(Text);
					VerifyRange(ReadLine(), 2);
					Status = true;
				}
			}
			else
			{
				VerifyRange(ReadLine(), 2);
				Status = true;
			}
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		DeleteObj(d->Data);
		DeleteObj(d->Listen);
	}

	return Status;
}

bool IFtp::RenameFile(char *From, char *To)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND From AND To)
		{
			// From
			sprintf(d->OutBuf, "RNFR %s\r\n", From);
			WriteLine();
			VerifyRange(ReadLine(), 3);

			// To
			sprintf(d->OutBuf, "RNTO %s\r\n", To);
			WriteLine();
			VerifyRange(ReadLine(), 2);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::DeleteFile(char *Remote)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND Remote)
		{
			sprintf(d->OutBuf, "DELE %s\r\n", Remote);
			WriteLine();
			VerifyRange(ReadLine(), 2);
			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::DownloadFile(char *Local, IFtpEntry *Remote, bool Binary)
{
	if (Remote AND Remote->Name)
	{
		return TransferFile(Local, Remote->Name, Remote->Size, false, Binary);
	}
	return false;
}

bool IFtp::UploadFile(char *Local, char *Remote, bool Binary)
{
	return TransferFile(Local, Remote, FileSize(Local), true, Binary);
}

bool IFtp::ResumeAt(int Pos)
{
	sprintf(d->OutBuf, "REST %i\r\n", Pos);
	WriteLine();
	int FtpStatus = ReadLine();
	bool Status = (FtpStatus/100) < 4;
	RestorePos = (Status) ? Pos : 0;
	return Status;
}

bool IFtp::TransferFile(char *Local, char *Remote, int64 Size, bool Upload, bool Binary)
{
	bool Status = false;
	bool Aborted = false;

	d->F = NEW(GFile);
	
	try
	{
		if (d->F AND
			Socket AND
			Local AND
			Remote)
		{
			if (d->F->Open(Local, (Upload)?O_READ:O_WRITE))
			{
				if (SetupData(Binary))
				{
					// Data command
					sprintf(d->OutBuf, "%s %s\r\n", (Upload)?"STOR":"RETR", Remote);
					WriteLine();

					// Build data connection
					if (ConnectData())
					{
						VerifyRange(ReadLine(), 1);

						// do the transfer
						int TempLen = 64 << 10;
						uchar *Temp = NEW(uchar[TempLen]);
						if (Temp)
						{
							int64 Processed = 0;
							int64 Len = 0;

							if (Meter)
							{
								Meter->SetLimits(0, Size);
							}
							
							if (NOT Upload)
							{
								Size -= RestorePos;
							}
							AbortTransfer = false;

							bool Error = false;
							while (	Socket AND
									d->Data AND
									Processed < Size AND
									NOT Error AND
									NOT AbortTransfer AND
									(NOT Meter OR NOT Meter->Cancel()))
							{
								if (Meter)
								{
									Meter->Value(0);
								}

								if (Upload)
								{
									// upload loop
									do
									{
										Len = d->F->Read(Temp, TempLen);
										if (Len > 0)
										{
											int WriteLen = 0;
											
											if (d->Data)
											{
												if (Pipe)
												{
													WriteLen = Pipe->WritePipe(d->Data, (char*) Temp, Len, 0);
												}
												else
												{
													WriteLen = d->Data->Write((char*) Temp, Len, 0);
												}
											}

											if (WriteLen == Len)
											{
												Processed += Len;
											}
											else
											{
												printf("%s:%i - Data->Write failed, %i of %i bytes written.\n",
													__FILE__, __LINE__, WriteLen, Len);
												Error = true;
											}
										}

										if (Meter AND Meter->Cancel())
										{
											printf("%s:%i - Upload canceled.\n", __FILE__, __LINE__);
											break;
										}
									}
									while (	Len > 0 AND
											NOT AbortTransfer);
											
									printf("Upload loop done.\n");
								}
								else
								{
									// restore support
									d->F->Seek(RestorePos, SEEK_SET);
									if (Meter AND RestorePos > 0)
									{
										Meter->SetParameter(10, RestorePos);
									}
									RestorePos = 0;

									// download loop
									while (d->Data)
									{
										if (d->Data)
										{
											if (Pipe)
											{
												Len = Pipe->ReadPipe(d->Data, (char*)Temp, TempLen, 0);
											}
											else
											{
												Len = d->Data->Read((char*) Temp, TempLen, 0);
											}
										}

										if (Len <= 0 OR
											AbortTransfer)
										{
											break;
										}

										d->F->Write(Temp, Len);
										Processed += Len;
										if (Meter)
										{
											if (Meter->Cancel())
											{
												break;
											}
										}
									}

									if (Len <= 0)
									{
										Error = true;
									}
								}
							}
							
							if (AbortTransfer OR
								(Meter AND Meter->Cancel()))
							{
								// send abort command
								sprintf(d->OutBuf, "ABOR\r\n");
								WriteLine();
								if ((ReadLine()/100) == 4) // temp msg
								{
									// response... we don't care just read it
									ReadLine();
								}

								Aborted = true;
							}

							if (d->Data)
							{
								d->Data->Close();
							}

							Status = Size == Processed;
							if (Meter)
							{
								Meter->Value(0);
							}

							DeleteArray(Temp);
						}

						// read eof response
						if (NOT Aborted)
						{
							VerifyRange(ReadLine(), 2);
						}
					}

					DeleteObj(d->Data);
					DeleteObj(d->Listen);
				}
				else
				{
					// data setup failed
				}
			}
			else
			{
				// couldn't open file...
				char Str[256];
				sprintf(Str, "Error: Couldn't open '%s' for %s", Local, (Upload)?(char*)"reading":(char*)"writing");
				Socket->OnInformation(Str);
			}
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	DeleteObj(d->F);

	return Status;
}

bool IFtp::SetupData(bool Binary)
{
	bool Status = false;
	try
	{
		DeleteObj(d->Data);

		if (IsOpen())
		{
			d->Data = NEW(ILogProxy(Socket, Pipe, Meter));
		}

		if (d->Data)
		{
			PassiveMode = false;

			if ((Binary AND d->CurMode != FT_Binary) OR
				(NOT Binary AND d->CurMode != FT_Ascii))
			{
				// Type command
				sprintf(d->OutBuf, "TYPE %c\r\n", Binary ? 'I' : 'A');
				if (WriteLine())
				{
					VerifyRange(ReadLine(), 2);
					d->CurMode = Binary ? FT_Binary : FT_Ascii;
				}
			}

			if (NOT ForceActive)
			{
				// Try the PASV command first... it's better for
				// getting out of a firewall to the server
				sprintf(d->OutBuf, "PASV\r\n");
				WriteLine();


				char Temp[256];
				if (ReadLine(Temp)/100 == 2)
				{


					// Ok we have PASV mode
					// grab the IP and Port
					char *Start = strchr(Temp, '(');
					if (Start)
					{
						Start++;
						char *End = strchr(Start, ')');
						if (End)
						{
							*End = 0;
							GToken T(Start, ",");
							if (T.Length() == 6)
							{
								sprintf(Ip, "%s.%s.%s.%s", T[0], T[1], T[2], T[3]);
								Port = (atoi(T[4])<<8) | atoi(T[5]);
								PassiveMode = true;
								return true;
							}
						}
					}
				}
				else
				{
					printf("\tread failed.\n");
					return false;
				}
			}

			// Ok, no PASV so we use the PORT command

			// Port command
			// static int p = (4<<8) | 27;

			bool ListenStatus = false;
			d->Listen = NEW(GSocketImpl);			// listen
			if (d->Listen)
			{
				if (d->Listen->Listen())
				{
					Socket->WhatsMyIp(Ip);				// get current IP
					Port = d->Listen->WhatsMyPort();		// get current port

					ListenStatus = true;
				}
			}

			if (ListenStatus)
			{
				// convert IP to PORT's comma delimited format
				for (char *p = Ip; *p; p++)
				{
					if (*p == '.')
					{
						*p = ',';
					}
				}

				sprintf(d->OutBuf, "PORT %s,%i,%i\r\n", Ip, Port>>8, Port&0xFF);
				WriteLine();
				VerifyRange(ReadLine(), 2);
				Status = true;
			}
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}

bool IFtp::ConnectData()
{
	if (PassiveMode)
	{
		return d->Data->Open(Ip, Port);
	}
	else
	{
		d->Data = NEW(ILogProxy(Socket, Pipe, Meter));
		if (d->Data)
		{
			return d->Listen->Accept(d->Data);
		}
	}

	return false;
}

bool IFtp::SetPerms(char *File, int Perms)
{
	bool Status = false;

	try
	{
		if (IsOpen() AND File)
		{
			sprintf(d->OutBuf, "SITE CHMOD %03.3X %s\r\n", Perms, File);
			WriteLine();
			VerifyRange(ReadLine(), 2);

			Status = true;
		}
	}
	catch (int Error)
	{
		printf("%s:%i - error: %i\n", __FILE__, __LINE__, Error);
		if (IsOpen())
		{
			LgiAssert(0);
		}
	}

	return Status;
}
