#include <stdio.h>
#include <errno.h>

#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
char *SemPrint(OsSemaphore *s)
{
	static char Buf[256];
	
	char *c = Buf;
	for (int i=0; i<sizeof(OsSemaphore); i++)
	{
		sprintf(c, "%02.2X,", ((uchar*)s)[i]);
		c += 3;
	}
	
	return Buf;
}

GSemaphore::GSemaphore(char *name)
{
	_Thread = 0;
	_Count = 0;
	_Name = NewStr(name);

	#if defined WIN32
	
	_Sem = CreateMutex(NULL, FALSE, NULL);
	
	#elif defined BEOS
	
	_Sem = create_sem(1, "LGI.GSemaphore");

	#elif defined LINUX

	// memset(&_Sem, 0xff, sizeof(_Sem));
	ZeroObj(_Sem);

	/*	
	pthread_mutexattr_t Attr;
	pthread_mutexattr_init(&Attr);
	pthread_mutexattr_settype(&Attr, PTHREAD_MUTEX_ERRORCHECK);
	*/
	
	if (pthread_mutex_init(&_Sem, 0))
	{
		LgiTrace("%s:%i - Couldn't create mutex for GSemaphore\n", __FILE__, __LINE__);
	}
	
	// pthread_mutexattr_destroy(&Attr);

	#endif
}

GSemaphore::~GSemaphore()
{
	#if defined WIN32

	CloseHandle(_Sem);
	_Sem = 0;

	#elif defined BEOS

	if (_Sem > 0)
	{
		delete_sem(_Sem);
		_Sem = 0;
	}

	#elif defined LINUX

	pthread_mutex_destroy(&_Sem);

	#endif
	
	DeleteArray(_Name);
}

char *GSemaphore::GetName()
{
	return _Name;
}

void GSemaphore::SetName(char *s)
{
	DeleteArray(_Name);
	_Name = NewStr(s);
}

bool GSemaphore::_Lock()
{
	#if defined WIN32

	LgiAssert(_Sem != 0);
	return WaitForSingleObject(_Sem, INFINITE) == WAIT_OBJECT_0;

	#elif defined BEOS

	return acquire_sem(_Sem) == B_OK;

	#else

	if (pthread_mutex_trylock(&_Sem))
	{
		// printf("\t%s:%i - pthread_mutex_trylock errored: %s\n\t_Sem=%s\n", __FILE__, __LINE__, GetErrorName(errno), SemPrint(&_Sem));
		return false;
	}
	
	return true;

	#endif
}

void GSemaphore::_Unlock()
{
	#if defined WIN32

	ReleaseMutex(_Sem);

	#elif defined BEOS

	release_sem(_Sem);

	#else

	if (pthread_mutex_unlock(&_Sem))
	{
		printf("\t%s:%i - pthread_mutex_unlock errored\n", __FILE__, __LINE__);
	}

	#endif
}

bool GSemaphore::Lock()
{
	int Start = LgiCurrentTime();
	bool Status = false;
	OsThreadId CurrentThread = LgiGetCurrentThread();
	bool Warn = true;
	
	while (NOT Status)
	{
		if (_Lock())
		{
			if (NOT _Thread OR
				_Thread == CurrentThread)
			{
				_Thread = CurrentThread;
				_Count++;
				Status = true;
			}
			_Unlock();
		}
		else
		{
			// printf("%s:%i - Didn't lock...\n", __FILE__, __LINE__);
		}

		if (NOT Status)
		{
			LgiSleep(5);
		}

		#if 1 // _DEBUG
		int Now = LgiCurrentTime();
		if (Warn AND Now > Start + 1000)
		{
			LgiTrace("GSemaphore=%p(%s): Can't lock... LockingThread=%i ThisThread=%i Count=%i.\n",
					this,
					_Name,
					_Thread,
					CurrentThread,
					_Count);
			Start = Now;

			// Warn = false;
		}
		
		if (LgiCurrentTime() > Start + (2 * 60 * 1000))
		{
			// Obviously we've locked up and to un-deadlock things we'll fail the lock
			LgiTrace("::Lock timeout ask_thread=%i hold_thread=%i sem=%i count=%i\n", _Thread, CurrentThread, _Sem, _Count);
			break;
		}
		#endif
	}

	return Status;
}

bool GSemaphore::LockWithTimeout(int Timeout)
{
	int Start = LgiCurrentTime();
	bool Status = false;

	while (NOT Status AND
			LgiCurrentTime() < Start + Timeout)
	{
		if (_Lock())
		{
			OsThreadId CurrentThread = LgiGetCurrentThread();
			if (NOT _Thread OR
				_Thread == CurrentThread)
			{
				_Thread = CurrentThread;
				_Count++;
				Status = true;
			}
			_Unlock();
		}

		if (NOT Status)
		{
			LgiSleep(5);
		}
	}

	return Status;
}

void GSemaphore::Unlock()
{
	if (_Lock())
	{
		if (_Count < 1)
		{
			#ifdef _DEBUG
			// I want a stack trace the next time this happens
			int *i = 0;
			*i = 0;
			#endif
			
			// if this assert fails then you tryed to unlock an object that
			// wasn't locked in the first place
			LgiAssert(0);
		}

		// decrement the lock counter
		if (_Count > 0)
		{
			_Count--;
		}
		if (_Count < 1)
		{
			_Thread = 0;
		}

		_Unlock();
	}
}

