/** /file
	/author Matthew Allen
	/brief A filesystem in a file.
	
	This library takes some things from various
	popular filesystems such as BeOS's BFS and 
	Linux's Reiser4 and combines them at a simple 
	level to create a lightweight file system with 
	transactions, arbitary indexed metadata, live 
	queries and small code size. What it doesn't
	attempt to do is be blindingly fast or useful 
	to everyone. Neither does it try and be a real 
	filesystem in a block device. It stores data 
	in a file, like a database would.
 */

#ifndef _MFS_H_
#define _MFS_H_

#include "LgiDefs.h"
#include "GMem.h"
#include "GFile.h"
#include "GContainers.h"
#include "GVariant.h"
#include "GSemaphore.h"
#include "GString.h"
#include "GArray.h"

/// Fwd decl
class MFileSystem;
class MFileQuery;
class MMetaData;
class MFile;

/// The class that manages one filesystem (file).
class MFileSystem
{
	friend class MFileQuery;
	friend class MMetaData;
	friend class MFile;
	friend class MTreeNode;

	class MFileSystemPrivate *d;
	
public:
	/// Constructor
	MFileSystem(char *File = 0);
	/// Destructor
	virtual ~MFileSystem();
	
	// Global ops
	
	/// Create a new file system in a file
	bool CreateNewSystem(char *File);
	/// Open an existing file system
	bool OpenSystem(char *File);
	/// Close the current file system
	bool CloseSystem();
	/// Returns the current volume file name
	char *GetCurrentFile();
	/// Commits all outstanding changes to metadata
	bool Commit();
	
	// Folder ops
	
	/// Creates a new folder or ignores the existing one
	bool CreateFolder(char *Path);
	/// Deletes an existing folder
	bool DeleteFolder(char *Path);
	
	// File ops
	
	/// Deletes the given file in the file system
	bool DeleteFile(char *File);
	/// Moves a file to a new location
	bool MoveFile(char *To, char *From, int64 *Progress = 0);
	/// Copies a file to a new location
	bool CopyFile(char *To, char *From, int64 *Progress = 0);
};

/// A live query
class MFileQuery
{
	class MFileQueryPrivate *d;

public:
	MFileQuery(MFileSystem *sys, char *query = 0);
	virtual ~MFileQuery();
	
	/// Sets the query string
	bool SetQuery(char *query);
	/// Gets the query string
	char *GetQuery();
	
	/// Called when the results are available or they change. Override this
	/// to receive events relating to the result set.
	virtual void OnResults
	(
		/// Paths that are being added to the result set
		List<char> &NewPaths,
		/// Paths that are being removed from the result set
		List<char> &DelPaths
	) {}
};

/// A metadata value.
class MMetaData : public GVariant
{
	MFile *File;
	char *Name;
	bool Ro;

public:
	MMetaData(MFile *file, char *name);
	~MMetaData();

	/// Returns the name of the metadata field
	char *GetName();
	/// True if the metadata is read only
	bool IsReadOnly();	
	/// Commits any change in the value back to disk. These changes
	/// are not garrenteed to be on disk until the file system has been
	/// commited as well.
	bool Commit();
};

/// A Mfs File.
class MFile : public GFile
{
	class MFilePrivate *d;

public:
	MFile(MFileSystem *sys, char *Path = 0, int Mode = O_READ);
	~MFile();
	
	// Normal File IO Functions
	
	int Open(char *Name, int Attrib);
	bool IsOpen();
	int Close();
	int GetOpenMode();
	int64 GetPos();
	int64 SetPos(int64 Pos);
	int64 GetSize();
	int64 SetSize(int64 Size);
	int Read(void *Buffer, int Size, int Flags = 0);
	int Write(void *Buffer, int Size, int Flags = 0);
	char *GetName();
	int64 Seek(int64 To, int Whence);
	bool Eof();
	void SetStatus(bool s = false);
	bool GetStatus();
	void SetSwap(bool s);
	bool GetSwap();

	/// Meta data access

	/// Gets the first metadata value
	MMetaData *FirstMd();
	/// Gets the next metadata value
	MMetaData *NextMd();
	/// Adds a new peice of metadata
	MMetaData *AddMd();
	/// Deletes a peice of metadata
	bool DelMd(char *Name);
};

#endif