//
// This should not be used by anything other than the MFS library
//
#ifndef _MFS_PRIV_H_
#define _MFS_PRIV_H_

#define MFS_SUPER_BLOCK			'bsFM'	// MFsb
#define MFS_INDEX_NODE			'niFM'	// MFin
#define MFS_BITMAP_BLOCK		'bbFM'	// MFbb
#define MFS_PTR_NODE			'npFM'	// MFpn
#define MFS_META_NODE			'nmFM'	// MFmn

#define MfsSystemId				" MFS v1.0"

class MTreeNode;
typedef bool (*NodeWalker)(MTreeNode *Node, int Data);

// A block address type: limits the file system to 4096 GB
// Which seems reasonable considering this was originally
// written as a backend for an email client ;)
typedef uint32 BlockAddr;

// A block run is a set of contiguous blocks
class BlockRun
{
public:
	BlockAddr	Start;
	uint32		Length;
};

// The core file system object.
class MFileSystemPrivate
{
public:
	// The global file object and it's lock.
	GFile File;
	GSemaphore FileLock;
	bool ReadOnly;

	// The various trees of BTree nodes
	MTreeNode *Uid;
	GArray<MTreeNode*> Indexes;

	// Blocks
	int BlockSize;
	int BlockShift;
	MTreeNode *BlockBmp;
	GArray<MTreeNode*> Dirty;

	bool AllocateBlocks(GArray<uint32> &Blocks, int Num = 1);
	
	// File access (these take a int64 Pos and return the new File 
	// pointer though the same variable).
	int ReadAt(int64 &Pos, void *Buf, int Len, int Flags = 0);
	int WriteAt(int64 &Pos, void *Buf, int Len, int Flags = 0);

	// Object
	MFileSystemPrivate();
	~MFileSystemPrivate();
};

class Inode
{
public:
	uint32 Uid;			// The UID for the inode.
	MTreeNode *Md;		// The current meta data node.
	int MdOffset;		// The offset into the meta data node where
						// this inode starts. (In bytes).
};

class MFileQueryPrivate
{
public:
	char *Query;
};

class MFilePrivate
{
public:
	MFileSystem *Sys;	// Filesystem.
	char *Path;			// Full path of the file.
	uint32 Uid;			// Uid of the Inode
	
	MFilePrivate();
	~MFilePrivate();
};

class MTreeNode
{
public:
	MFileSystem *Sys;	// The filesystem
	MTreeNode *Left;	// The left node in the binary tree
	MTreeNode *Right;	// The right node in the binary tree
	MTreeNode *Parent;	// The parent node
	uint32 Block;		// The current address of the data block on disk.
						// If the disk is an array of blocks, each 'BlockSize'
						// bytes long then this is the array index used to
						// access the data.
	char *Data;			// The data block (always BlockSize bytes long)
	bool Dirty;			// True if the data block is dirty
	int Depth;			// Depth of the node in the tree
	
	uint32 GetMagic()
	{
		return (Data) ? *((uint32*)Data) : 0;
	}

	// Object
	MTreeNode(MFileSystem *sys, uint32 block);
	~MTreeNode();

	// Methods
	void SetDirty();
	bool SetClean();
	bool Load(uint32 Block = 0);
	bool Save();
	bool Walk(NodeWalker Callback, int Data = 0);
	bool OnChildMoved(MTreeNode *Child, uint32 NewPos);
};

class MSuperNode
{
public:
	uint32 Magic;			// MFS_SUPER_BLOCK
	char Id[32];			// Always set to MfsSystemId
	uint32 BlockSize;		// >= 1k AND <= 8k
	BlockAddr Root;
	BlockAddr Index;
	BlockRun Bmp;
	
	uint8 Reserved[972];	// Unused
};

class MString
{
public:
	uint8 Len;
	char Str[1];
	
	MString *Next()
	{
		return (MString*) (Str + Len);
	}
};

class MKey
{
public:
	uint8 Type;				// GVariant's type: GV_???
	union
	{
		uint32 Int;			// GV_INT32
		MString Str;		// GV_STRING
	};

	MKey *Next()
	{
		switch (Type)
		{
			case GV_INT32:
			{
				return (MKey*) (((char*)this) + sizeof(Int));
			}
			case GV_STRING:
			{
				return (MKey*) Str.Next();
			}
		}
		
		LgiAssert(0);
		return 0;
	}
};

class MValue
{
public:
	uint16 MetaName;		// Key into meta data name hash table
	MKey Value;				// Type and value

	MValue *Next()
	{
		return (MValue*) Value.Next();
	}
};

#define MINODE_RUN_SIZE			(8<<10) // 8kb

class MInode
{
public:
	// Basic information (36 bytes)
	BlockAddr	Inode;			// The address on disk
	uint32		Mode;			// protection
    uint16		UserId;			// user ID of owner
    uint16		GrpId;			// group ID of owner
	uint64		Size;			// total size, in bytes
	uint64		ModTime;		// time of last modification
	uint64		CreateTime;		// time of creation

	// Meta stuff (8 bytes)
	BlockRun	Attributes;		// Points to the blocks that store attributes

	// The file's data itself
	BlockAddr	Direct[16];		// First 16 blocks directly address (64 bytes)
	BlockRun	Runs[64];		// Some runs of MINODE_RUN_SIZE bytes (512 bytes)
	BlockAddr	Indirect[64];	// Indirect blocks storing runs pointing to MINODE_RUN_SIZE bytes each (256 bytes)
};

#endif