#include <math.h>

#include "Lgi.h"
#include "GTree.h"
#include "GScrollBar.h"

#define TREE_BLOCK			16
#define DRAG_THRESHOLD		4

#ifdef LINUX
#define TreeUpdateNow		false
#else
#define TreeUpdateNow		true
#endif

//////////////////////////////////////////////////////////////////////////////
// Private class definitions for binary compatibility
class GTreePrivate
{
public:
	// Private data
	int				LineFlags[4];
	bool			LayoutDirty;
	GdcPt2			Limit;
	GdcPt2			LastClick;
	GTreeItem		*LastHit;
	List<GTreeItem>	Selection;
	GTreeItem		*DropTarget;
	GMemDC			*IconCache;

	GTreePrivate()
	{
		LastHit = 0;
		DropTarget = 0;
		IconCache = 0;
		LayoutDirty = true;
	}
	
	~GTreePrivate()
	{
		DeleteObj(IconCache);
	}
};

class GTreeItemPrivate
{
public:
	GRect _Pos;
	GRect _Thumb;
	GRect _Text;
	GRect _Icon;
	bool _Open;
	bool _Selected;
	bool _Visible;
	bool _Last;
	int _Depth;
	
	GTreeItemPrivate()
	{
		_Pos.ZOff(-1, -1);
		_Open = false;
		_Selected = false;
		_Visible = false;
		_Last = false;
		_Depth = 0;
	}
};

//////////////////////////////////////////////////////////////////////////////
GTreeItem::GTreeItem()
{
	d = NEW(GTreeItemPrivate);
	Tree = 0;
	Parent = 0;
	Str = 0;
	Sys_Image = -1;
}

GTreeItem::~GTreeItem()
{
	if (Tree)
	{
		if (Tree->d->LastHit == this)
		{
			Tree->d->LastHit = 0;
		}
		Tree->Capture(false);
	}

	int y;
	GTree *t = 0;
	if (Parent AND Parent != (GTreeItem*)Tree)
	{
		t = Tree;
		y = Parent->d->_Pos.y1;
	}
	else if (this != (GTreeItem*)Tree)
	{
		t = Tree;
		GTreeItem *p = GetPrev();
		if (p)
		{
			y = p->d->_Pos.y1;
		}
		else
		{
			y = d->_Pos.y1;
		}
	}

	_Remove();
	DeleteArray(Str);
	DeleteObj(d);

	if (t)
	{
		t->_UpdateBelow(y);
	}
}

GdcPt2 GTreeItem::_ScrollPos()
{
	GdcPt2 p;
	if (Tree) p = Tree->_ScrollPos();
	return p;
}

GRect *GTreeItem::_GetRect(GTreeItemRect Which)
{
	switch (Which)
	{
		case TreeItemPos: return &d->_Pos;
		case TreeItemThumb: return &d->_Thumb;
		case TreeItemText: return &d->_Text;
		case TreeItemIcon: return &d->_Icon;
	}
	
	return 0;
}

void GTreeItem::_RePour()
{
	if (Tree)
	{
		Tree->_Pour();
	}
}

void GTreeItem::ScrollTo()
{
	if (Tree AND Tree->VScroll)
	{
		GRect c = Tree->GetClient();
		GRect p = d->_Pos;
		int y = d->_Pos.Y() ? d->_Pos.Y() : 16;
		p.Offset(0, -Tree->VScroll->Value() * y);

		if (p.y1 < c.y1)
		{
			int Lines = (c.y1 - p.y1 + y - 1) / y;
			Tree->VScroll->Value(Tree->VScroll->Value() - Lines);
		}
		else if (p.y2 > c.y2)
		{
			int Lines = (p.y2 - c.y2 + y - 1) / y;
			Tree->VScroll->Value(Tree->VScroll->Value() + Lines);
		}
	}
}

void GTreeItem::_SetTreePtr(GTree *t)
{
	Tree = t;
	for (GTreeItem *i=Items.First(); i; i=Items.Next())
	{
		i->_SetTreePtr(t);
	}
}

bool GTreeItem::IsRoot()
{
	return Parent AND Parent == (GTreeItem*)Tree;
}

void GTreeItem::_Remove()
{
	for (GTreeItem *i=Items.First(); i; i=Items.First())
	{
		i->_Remove();
	}

	if (this != (GTreeItem*)Tree)
	{
		if (Parent)
		{
			LgiAssert(Parent->Items.HasItem(this));
			Parent->Items.Delete(this);
		}

		if (d->_Selected)
		{
			if (Tree)
			{
				Tree->d->Selection.Delete(this);
			}
			d->_Selected = false;
		}

		if (Tree)
		{
			LgiAssert(Tree->d);

			if (Tree->d->LastHit == this)
			{
				Tree->d->LastHit = 0;
			}
			
			Tree->d->LayoutDirty = true;
			Tree->Capture(false);
		}
	}

	Parent = 0;
	Tree = 0;
}

void GTreeItem::_PourText(GdcPt2 &Size)
{
	char *t = GetText();
	Size.x = (t ? SysFont->X(t) : 0) + 4;
	Size.y = 0;
}

void GTreeItem::_PaintText(GSurface *pDC, COLOUR Fore, COLOUR Back)
{
	char *Text = GetText();
	if (Text)
	{
		int Tab = SysFont->TabSize();
		SysFont->TabSize(0);
		SysFont->Transparent(false);
		SysFont->Colour(Fore, Back);
		SysFont->Text(pDC, d->_Text.x1 + 2, d->_Text.y1 + 1, Text, -1, &d->_Text);
		SysFont->TabSize(Tab);
	}
	else
	{
		pDC->Colour(Back, 24);
		pDC->Rectangle(&d->_Text);
	}
}

void GTreeItem::_Pour(GdcPt2 *Limit, int Depth, bool Visible)
{
	d->_Visible = Visible;
	d->_Depth = Depth;

	if (d->_Visible)
	{
		GdcPt2 TextSize;
		_PourText(TextSize);
		int IconX = (GetImage() >= 0) ? TREE_BLOCK : 0;
		int Height = TextSize.y ? TextSize.y : TREE_BLOCK;

		d->_Pos.ZOff(	(TREE_BLOCK*d->_Depth) +	// trunk
						TREE_BLOCK +				// node
						IconX +						// icon if present
						TextSize.x,					// text if present
						max(Height, SysFont->Y("A"))-1);
		d->_Pos.Offset(0, Limit->y);

		Limit->x = max(Limit->x, d->_Pos.x2 + 1);
		Limit->y = max(Limit->y, d->_Pos.y2 + 1);
	}
	else
	{
		d->_Pos.ZOff(-1, -1);
	}

	GTreeItem *n;
	for (GTreeItem *i=Items.First(); i; i=n)
	{
		n=Items.Next();
		i->d->_Last = n == 0;
		i->_Pour(Limit, Depth+1, d->_Open AND d->_Visible);
	}
}

char *GTreeItem::GetText(int i)
{
	return Str;
}

int GTreeItem::IndexOf()
{
	if (Parent)
	{
		return Parent->Items.IndexOf(this);
	}
	else if (Tree)
	{
		return Tree->Items.IndexOf(this);
	}

	return -1;
}

GTreeItem *GTreeItem::GetChild()
{
	return Items.First();
}

GTreeItem *GTreeItem::GetPrev()
{
	List<GTreeItem> *l = (Parent) ? &Parent->Items : (Tree) ? &Tree->Items : 0;
	if (l)
	{
		int Index = l->IndexOf(this);
		if (Index >= 0)
		{
			return l->ItemAt(Index-1);
		}
	}

	return 0;
}

GTreeItem *GTreeItem::GetNext()
{
	List<GTreeItem> *l = (Parent) ? &Parent->Items : (Tree) ? &Tree->Items : 0;
	if (l)
	{
		int Index = l->IndexOf(this);
		if (Index >= 0)
		{
			return l->ItemAt(Index+1);
		}
	}

	return 0;
}

bool GTreeItem::SetText(char *s, int i)
{
	char *New = NewStr(s);
	DeleteArray(Str);
	Str = New;

	if (Tree)
	{
		Update();
	}

	return true;
}

int GTreeItem::GetImage(int Flags)
{
	return Sys_Image;
}

void GTreeItem::SetImage(int i)
{
	Sys_Image = i;
}

GTreeItem *GTreeItem::Insert(GTreeItem *Obj, int Pos)
{
	if (Obj AND Obj->Tree)
	{
		Obj->Remove();
	}
	
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj)
	{
		NewObj->Parent = this;
		NewObj->_SetTreePtr(Tree);

		Items.Delete(NewObj);
		Items.Insert(NewObj, Pos);

		if (Tree)
		{
			Tree->d->LayoutDirty = true;
			if (d->_Pos.Y() > 0)
			{
				Tree->_UpdateBelow(d->_Pos.y1);
			}
			else
			{
				Tree->Invalidate();
			}
		}
	}

	return NewObj;
}

void GTreeItem::Detach()
{
	if (Parent)
	{
		LgiAssert(Parent->Items.HasItem(this));
		Parent->Items.Delete();
		Parent = 0;
	}
	if (Tree)
	{
		Tree->d->LayoutDirty = true;
		Tree->Invalidate();
	}
	_SetTreePtr(0);
}

void GTreeItem::Remove()
{
	int y = 0;
	if (Parent)
	{
		if (Parent == Tree)
		{
			GTreeItem *Prev = GetPrev();
			if (Prev)
			{
				y = Prev->d->_Pos.y1;
			}
			else
			{
				y = d->_Pos.y1;
			}
		}
		else
		{
			y = Parent->d->_Pos.y1;
		}
	}

	GTree *t = Tree;

	_Remove();

	if (t)
	{
		t->_UpdateBelow(y);
	}
}


void GTreeItem::Update()
{
	if (Tree)
	{
		GRect p = d->_Pos;
		p.x2 = 1000;
		Tree->_Update(&p, TreeUpdateNow);
	}
}

bool GTreeItem::Select()
{
	return d->_Selected;
}

void GTreeItem::Select(bool b)
{
	if (d->_Selected != b)
	{
		d->_Selected = b;
		Update();

		if (b AND Tree)
		{
			Tree->_OnSelect(this);
			Tree->OnItemSelect(this);
		}
	}
}

bool GTreeItem::Expanded()
{
	return d->_Open;
}

void GTreeItem::Expanded(bool b)
{
	if (d->_Open != b)
	{
		d->_Open = b;

		if (Items.GetItems() > 0)
		{
			if (Tree)
			{
				Tree->d->LayoutDirty = true;
				Tree->_UpdateBelow(d->_Pos.y1);
			}
			OnExpand(b);
		}
	}
}

GTreeItem *GTreeItem::_HitTest(int x, int y)
{
	GTreeItem *Status = 0;

	if (d->_Pos.Overlap(x, y) AND
		x > (d->_Depth*TREE_BLOCK))
	{
		Status = this;
	}

	if (d->_Open)
	{
		for (GTreeItem *i=Items.First(); i AND NOT Status; i=Items.Next())
		{
			Status = i->_HitTest(x, y);
		}
	}

	return Status;
}

void GTreeItem::_MouseClick(GMouse &m)
{
	if (m.Down())
	{
		if ((Items.GetItems() > 0 AND
			d->_Thumb.Overlap(m.x, m.y)) OR
			m.Double())
		{
			Expanded(NOT Expanded());
		}

		if (d->_Text.Overlap(m.x, m.y) OR
			d->_Icon.Overlap(m.x, m.y))
		{
			Select(true);

			if (Tree)
			{
				Tree->OnItemClick(this, m);
			}
		}
	}
}

void GTreeItem::OnPaint(GSurface *pDC, GRect &r, COLOUR Back)
{
	LgiAssert(Tree);

	bool IsSelected = (Tree->d->DropTarget == this) OR (Tree->d->DropTarget == 0 AND d->_Selected);

	// background up to text
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(0, d->_Pos.y1, (d->_Depth*TREE_BLOCK)+TREE_BLOCK, d->_Pos.y2);

	// draw trunk
	int x = 0;
	pDC->Colour(LC_MED, 24);
	for (int i=0; i<d->_Depth; i++)
	{
		if (Tree->d->LineFlags[0] & (1 << i))
		{
			pDC->Line(x + 8, d->_Pos.y1, x + 8, d->_Pos.y2);
		}
		x += TREE_BLOCK;
	}

	// draw node
	int cy = d->_Pos.y1 + (d->_Pos.Y() >> 1);
	if (Items.GetItems() > 0)
	{
		// plus/minus symbol
		d->_Thumb.ZOff(8, 8);
		d->_Thumb.Offset(x + 4, cy - 4);

		pDC->Colour(LC_LOW, 24);
		pDC->Box(&d->_Thumb);
		pDC->Colour(LC_WHITE, 24);
		pDC->Rectangle(d->_Thumb.x1+1, d->_Thumb.y1+1, d->_Thumb.x2-1, d->_Thumb.y2-1);
		pDC->Colour(LC_SHADOW, 24);
		pDC->Line(	d->_Thumb.x1+2,
					d->_Thumb.y1+4,
					d->_Thumb.x1+6,
					d->_Thumb.y1+4);

		pDC->Colour(LC_MED, 24);

		if (Parent OR IndexOf() > 0)
		{
			// draw line to item above
			pDC->Line(x + 8, d->_Pos.y1, x + 8, d->_Thumb.y1-1);
		}

		// draw line to leaf beside
		pDC->Line(d->_Thumb.x2+1, cy, x + (TREE_BLOCK-1), cy);

		if (NOT d->_Last)
		{
			// draw line to item below
			pDC->Line(x + 8, d->_Thumb.y2+1, x + 8, d->_Pos.y2);
		}

		if (NOT d->_Open)
		{
			// not open, so draw the cross bar making the '-' into a '+'
			pDC->Colour(LC_SHADOW, 24);
			pDC->Line(	d->_Thumb.x1+4,
						d->_Thumb.y1+2,
						d->_Thumb.x1+4,
						d->_Thumb.y1+6);
		}
	}
	else
	{
		// leaf node
		pDC->Colour(LC_MED, 24);
		if (d->_Last)
		{
			pDC->Rectangle(x + 8, d->_Pos.y1, x + 8, cy);
		}
		else
		{
			pDC->Rectangle(x + 8, d->_Pos.y1, x + 8, d->_Pos.y2);
		}

		pDC->Rectangle(x + 8, cy, x + (TREE_BLOCK-1), cy);
	}
	x += TREE_BLOCK;

	// draw icon
	int Image = GetImage(Select());
	if (Image >= 0 AND
		Tree->GetImageList())
	{
		d->_Icon.ZOff(TREE_BLOCK-1, TREE_BLOCK-1);
		d->_Icon.Offset(x, d->_Pos.y1);

		if (Tree->d->IconCache)
		{
			// no flicker
			GRect From;
			From.ZOff(TREE_BLOCK-1, TREE_BLOCK-1);
			From.Offset(Tree->GetImageList()->TileX()*Image, 0);
			pDC->Blt(d->_Icon.x1, d->_Icon.y1, Tree->d->IconCache, &From);
		}
		else
		{
			// flickers...
			Tree->GetImageList()->Draw(pDC, x, d->_Pos.y1, Image);
		}

		if (d->_Icon.y2 < d->_Pos.y2)
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(x, d->_Icon.y2, d->_Icon.x2, d->_Pos.y2);
		}

		x += TREE_BLOCK;
	}

	// text
	char *Text = GetText();
	if (Text)
	{
		GdcPt2 TextSize;
		_PourText(TextSize);

		d->_Text.ZOff(TextSize.x, d->_Pos.Y()-1);
		d->_Text.Offset(x, d->_Pos.y1);
		_PaintText(	pDC,
					(IsSelected) ? LC_SEL_TEXT : LC_TEXT,
					(IsSelected) ? Back : LC_WORKSPACE);
		d->_Pos.x2 = d->_Text.x2;
	}
	else
	{
		d->_Text.ZOff(0, d->_Pos.Y()-1);
		d->_Text.Offset(x, d->_Pos.y1);
	}

	// background after text
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(d->_Text.x2, d->_Pos.y1, max(Tree->X(), Tree->d->Limit.x), d->_Pos.y2);

	// children
	if (d->_Open)
	{
		if (NOT d->_Last)
		{
			Tree->d->LineFlags[0] |= 1 << d->_Depth;
		}

		for (GTreeItem *i=Items.First(); i; i=Items.Next())
		{
			i->OnPaint(pDC, r, Back);
		}

		Tree->d->LineFlags[0] &= ~(1 << d->_Depth);
	}
}

//////////////////////////////////////////////////////////////////////////////
GTree::GTree(int id, int x, int y, int cx, int cy, char *name) :
	ResObject(Res_TreeView)
{
	d = NEW(GTreePrivate);
	SetId(id);
	GRect e(x, y, x+cx, y+cy);
	SetPos(e);
	if (name) Name(name);
	else Name("LGI_GTree");
	Sunken(true);

	Tree = this;
	Lines = true;
	Buttons = true;
	LinesAtRoot = true;
	EditLabels = false;
	Tree = this;
	MultipleSelect = false;

	#if defined WIN32
	SetStyle(GetStyle() | WS_CHILD | WS_VISIBLE | WS_TABSTOP);
	#elif defined LINUX
	Handle()->isTabStop(true);
	#endif
}

GTree::~GTree()
{
	Tree = 0;
	Empty();
	DeleteObj(d);
}

// Internal tree methods
void GTree::_Update(GRect *r, bool Now)
{
	if (r)
	{
		GRect u = *r;
		GdcPt2 s = _ScrollPos();
		GRect c = GetClient();
		u.Offset(c.x1-s.x, c.y1-s.y);
		Invalidate(&u, Now);
	}
	else
	{
		Invalidate((GRect*)0, Now);
	}
}

void GTree::_UpdateBelow(int y, bool Now)
{
	GdcPt2 s = _ScrollPos();
	GRect c = GetClient();
	GRect u(c.x1, y - s.y + c.y1, X()-1, Y()-1);
	Invalidate(&u, Now);
}

GdcPt2 GTree::_ScrollPos()
{
	/*
	int All = _Limit.y / TREE_BLOCK - 1;
	int Visible = Y() / TREE_BLOCK;
	*/
	GdcPt2 Status;

	Status.x = (HScroll) ? HScroll->Value() : 0;
	Status.y = (VScroll) ? VScroll->Value() * TREE_BLOCK : 0;

	return Status;
}

void GTree::_UpdateScrollBars()
{
	static bool Processing = false;
	if (NOT Processing)
	{
		Processing = true;

		GdcPt2 Old = _ScrollPos();
		
		GRect Client = GetClient();
		bool x = d->Limit.x > Client.X();
		bool y = d->Limit.y > Client.Y();
		SetScrollBars(x, y);
		Client = GetClient();

		// x scroll... in pixels
		if (HScroll)
		{
			HScroll->SetLimits(0, d->Limit.x-1);
			HScroll->SetPage(Client.X());

			int Max = d->Limit.x - Client.X();
			if (HScroll->Value() > Max)
			{
				HScroll->Value(Max+1);
			}
		}

		// y scroll... in items
		if (VScroll)
		{
			int All = d->Limit.y / TREE_BLOCK;
			int Visible = Client.Y() / TREE_BLOCK;

			VScroll->SetLimits(0, All - 1);
			VScroll->SetPage(Visible);

			/*
			int Max = All - Visible + 1;
			if (VScroll->Value() > Max)
			{
				VScroll->Value(Max);
			}
			*/
		}

		GdcPt2 New = _ScrollPos();
		if (Old.x != New.x OR
			Old.y != New.y)
		{
			Invalidate();
		}

		Processing = false;
	}
}

void GTree::_OnSelect(GTreeItem *Item)
{
	if (NOT MultipleSelect)
	{
		for (GTreeItem *i=d->Selection.First(); i; i=d->Selection.Next())
		{
			if (i != Item)
			{
				i->Select(false);
			}
		}

		d->Selection.Empty();
	}
	else
	{
		d->Selection.Delete(Item);
	}

	d->Selection.Insert(Item);
}

void GTree::_Pour()
{
	d->Limit.x = d->Limit.y = 0;

	GTreeItem *n;
	for (GTreeItem *i=Items.First(); i; i=n)
	{
		n=Items.Next();
		i->d->_Last = n == 0;
		i->_Pour(&d->Limit, 0, true);
	}

	_UpdateScrollBars();
	d->LayoutDirty = false;
}

// External methods and events
void GTree::OnItemSelect(GTreeItem *Item)
{
	if (Item)
	{
		Item->OnSelect();

		GView *n=GetNotify()?GetNotify():GLayout::GetParent();
		if (n)
		{
			n->OnNotify(this, GLIST_NOTIFY_SELECT);
		}
	}
}

void GTree::OnItemExpand(GTreeItem *Item, bool Expand)
{
	if (Item)
	{
		Item->OnExpand(Expand);
	}
}

GTreeItem *GTree::GetAdjacent(GTreeItem *i, bool Down)
{
	if (i)
	{
		if (Down)
		{
			GTreeItem *n = i->GetChild();
			
			if (NOT n OR
				NOT n->d->_Visible)
			{
				for (n = i; n; )
				{
					GTreeItem *p = n->GetParent();
					if (p)
					{
						int Index = n->IndexOf();
						if (Index < p->Items.GetItems()-1)
						{
							n = n->GetNext();
							break;
						}
						else
						{
							n = p;
						}
					}
					else
					{
						n = n->GetNext();
						break;
					}
				}
			}

			return n;
		}
		else
		{
			GTreeItem *p = i->GetParent() ? i->GetParent() : this;
			if (p)
			{
				int Index = i->IndexOf();
				GTreeItem *n = p;
				if (Index > 0)
				{
					n = i->GetPrev();
					while (	n->GetChild() AND
							n->GetChild()->d->_Visible)
					{
						n = n->Items.ItemAt(n->Items.GetItems()-1);
					}
				}

				if (n AND n != this)
				{
					return n;
				}
			}
		}
	}
	
	return 0;
}

bool GTree::OnKey(GKey &k)
{
	bool Status = false;
	
	GTreeItem *i = d->Selection.First();
	if (NOT i)
	{
		i = Items.First();
		if (i)
		{
			i->Select();
		}
	}

	if (k.Down())
	{
		switch (k.c)
		{
			case VK_PRIOR:
			case VK_NEXT:
			{
				if (i AND i->d->_Pos.Y() > 0)
				{
					int Page = GetClient().Y() / i->d->_Pos.Y();
					for (int j=0; j<Page; j++)
					{
						GTreeItem *n = GetAdjacent(i, k.c == VK_NEXT);
						if (n)
						{
							i = n;
						}
						else break;
					}
					if (i)
					{
						i->Select(true);
						i->ScrollTo();
					}
				}
				Status = true;
				break;
			}
			case VK_HOME:
			{
				GTreeItem *i;
				if (i = Tree->Items.First())
				{
					i->Select(true);
					i->ScrollTo();
				}
				Status = true;
				break;
			}
			case VK_END:
			{
				GTreeItem *n = i, *p = 0;
				while (n = GetAdjacent(n, true))
				{
					p = n;
				}
				if (p)
				{
					p->Select(true);
					p->ScrollTo();
				}
				Status = true;
				break;
			}
			case VK_LEFT:
			{
				if (i)
				{
					if (i->Items.First() AND i->Expanded())
					{
						i->Expanded(false);
						break;
					}
					else
					{
						GTreeItem *p = i->GetParent();
						if (p)
						{
							p->Select(true);
							p->Expanded(false);
							if (d->LayoutDirty)
							{
								_Pour();
								break;
							}
						}
					}
				}
				// fall thru
			}
			case VK_UP:
			{
				GTreeItem *n = GetAdjacent(i, false);
				if (n)
				{
					n->Select(true);
					n->ScrollTo();
				}
				Status = true;
				break;
			}
			case VK_RIGHT:
			{
				if (i)
				{
					i->Expanded(true);
					if (d->LayoutDirty)
					{
						_Pour();
						break;
					}
				}
				// fall thru
			}
			case VK_DOWN:
			{
				GTreeItem *n = GetAdjacent(i, true);
				if (n)
				{
					n->Select(true);
					n->ScrollTo();
				}
				Status = true;
				break;
			}
		}
	}

	if (i AND i != (GTreeItem*)this)
	{
		i->OnKey(k);
	}

	return Status;
}

GTreeItem *GTree::ItemAtPoint(int x, int y)
{
	GdcPt2 s = _ScrollPos();

	for (GTreeItem *i = Items.First(); i; i=Items.Next())
	{
		GTreeItem *Hit = i->_HitTest(s.x + x, s.y + y);
		if (Hit)
		{
			return Hit;
		}
	}

	return 0;
}

void GTree::OnMouseWheel(double Lines)
{
	if (VScroll)
	{
		VScroll->Value(VScroll->Value() + Lines);
	}
}

void GTree::OnMouseClick(GMouse &m)
{
	GRect c = GetClient();

	m.x -= c.x1;
	m.y -= c.y1;

	if (m.Down())
	{
		Focus(true);
	}
	if (m.Down())
	{
		Capture(true);
		d->LastClick.x = m.x;
		d->LastClick.y = m.y;
	}
	else if (IsCapturing())
	{
		Capture(false);
	}

	d->LastHit = ItemAtPoint(m.x, m.y);
	if (d->LastHit)
	{
		GdcPt2 c = _ScrollPos();
		m.x += c.x;
		m.y += c.y;
		d->LastHit->_MouseClick(m);
	}
}

void GTree::OnMouseMove(GMouse &m)
{
	if (IsCapturing())
	{
		if (abs(d->LastClick.x - m.x) > DRAG_THRESHOLD OR
			abs(d->LastClick.y - m.y) > DRAG_THRESHOLD)
		{
			OnItemBeginDrag(d->LastHit, m.Flags);
			Capture(false);
		}
	}
}

void GTree::OnPosChange()
{
	GLayout::OnPosChange();
	_UpdateScrollBars();
}

void GTree::OnPaint(GSurface *pDC)
{
	GRect r = GetClient();

	// icon cache
	if (GetImageList() AND
		NOT d->IconCache)
	{
		int Bits = GdcD->GetBits();
		d->IconCache = NEW(GMemDC);
		if (d->IconCache AND
			d->IconCache->Create(GetImageList()->X(), GetImageList()->Y(), Bits))
		{
			if (Bits <= 8)
			{
				d->IconCache->Palette(NEW(GPalette(GdcD->GetGlobalColour()->GetPalette())));
			}

			d->IconCache->Colour(LC_WORKSPACE, 24);
			d->IconCache->Rectangle();

			GetImageList()->Lock();
			for (int i=0; i<GetImageList()->GetItems(); i++)
			{
				GetImageList()->Draw(d->IconCache, i * GetImageList()->TileX(), 0, i);
			}
			GetImageList()->Unlock();
			d->IconCache->Unlock();
		}
	}

	// scroll
	GdcPt2 s = _ScrollPos();
	pDC->SetOrigin(s.x - r.x1, s.y - r.y1);

	// selection colour
	COLOUR SelectedCol = Focus() ? LC_SELECTION : GdcMixColour(LC_SELECTION, LC_WORKSPACE);

	// layout items
	if (d->LayoutDirty)
	{
		_Pour();
	}

	// paint items
	ZeroObj(d->LineFlags);
	for (GTreeItem *i = Items.First(); i; i=Items.Next())
	{
		GRect Item;
		i->OnPaint(pDC, Item, SelectedCol);
	}

	pDC->SetOrigin(0, 0);
	if (d->Limit.y-s.y < r.Y())
	{
		// paint after items
		pDC->Colour(LC_WORKSPACE, 24);
		pDC->Rectangle(r.x1, r.y1 + d->Limit.y - s.y, r.x2, r.y2);
	}
}

int GTree::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_HSCROLL:
		case IDC_VSCROLL:
		{
			Invalidate();
			break;
		}
	}

	return GLayout::OnNotify(Ctrl, Flags);
}

int GTree::OnEvent(GMessage *Msg)
{
	return GLayout::OnEvent(Msg);
}

GTreeItem *GTree::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = GTreeItem::Insert(Obj, Pos);
	if (NewObj)
	{
		NewObj->Parent = this;
		NewObj->_SetTreePtr(this);
	}

	return NewObj;
}

bool GTree::Remove(GTreeItem *Obj)
{
	if (Obj AND
		Obj->Tree == this)
	{
		Obj->Remove();
		return true;
	}
	return false;
}

void GTree::RemoveAll()
{
	for (GTreeItem *i=Items.First(); i; i=Items.First())
	{
		i->_Remove();
	}

	Invalidate();
}

void GTree::Empty()
{
	GTreeItem *i;
	while (i=Items.First())
	{
		Delete(i);
	}
}

bool GTree::Delete(GTreeItem *Obj)
{
	bool Status = false;
	
	if (Obj)
	{
		GTreeItem *i;
		while (i = Obj->Items.First())
		{
			Delete(i);
		}
		
		Obj->Remove();
		DeleteObj(Obj);
		Status = true;
	}
	
	return Status;
}

void GTree::SelectDropTarget(GTreeItem *Item)
{
	bool Update = (d->DropTarget != 0) ^ (Item != 0);
	GTreeItem *Old = d->DropTarget;

	d->DropTarget = Item;
	if (Old)
	{
		Old->Update();
	}
	if (d->DropTarget)
	{
		d->DropTarget->Update();
	}
	if (Update)
	{
		OnFocus(true);
	}
}

bool GTree::Select(GTreeItem *Obj)
{
	bool Status = false;
	if (Obj AND IsAttached())
	{
		Obj->Select(true);
		Status = true;
	}
	else if (d->Selection.GetItems())
	{		
		d->Selection.Empty();
		OnItemSelect(0);
		Status = true;
	}
	return Status;
}

GTreeItem *GTree::Selection()
{
	return d->Selection.First();
}

void GTree::OnItemClick(GTreeItem *Item, GMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GTree::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	if (Item)
	{
		GMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

void GTree::OnFocus(bool b)
{
	// errors during deletion of the control can cause 
	// this to be called after the destructor
	if (d)
	{
		for (GTreeItem *i=d->Selection.First(); i; i=d->Selection.Next())
		{
			i->Update();
		}
	}
}
