// Linux Implementation of General LGI functions
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>

#define _POSIX_TIMERS
#include <time.h>

#include "Lgi.h"
#include "GProcess.h"
#include "GTextLabel.h"
#include "GButton.h"

////////////////////////////////////////////////////////////////
// Local helper functions
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

void _lgi_sleep(int i)
{
	struct timespec request, remain;

	ZeroObj(request);
	ZeroObj(remain);
	
	request.tv_sec = i / 1000;
	request.tv_nsec = (i % 1000) * 1000000;

	//printf("%i LgiSleep(%i)\n", LgiGetCurrentThread(), i);
	while (nanosleep(&request, &remain) == -1)
	{
	    request = remain;
		//printf("\t%i Resleeping=%i\n", LgiGetCurrentThread(), request.tv_sec*1000 + request.tv_nsec/1000);
	}
}

void _lgi_assert(bool b, char *test, char *file, int line)
{
	static bool Asserting = false;

	if (NOT b AND NOT Asserting)
	{
		Asserting = true;

		printf("%s:%i - Assert failed:\n%s\n", file, line, test);

		LgiMsg(0,
				"Assert failed, file: %s, line: %i\n"
				"%s\n",
				"Debug",
				MB_OK,
				file, line, test);

		exit(-1);
	}
}

////////////////////////////////////////////////////////////////////////
// Implementations
GMessage CreateMsg(int m, int a, int b)
{
	static GMessage Msg;
	Msg.m() = m;
	Msg.a() = a;
	Msg.b() = b;
	return Msg;
}

bool LgiGetFileMimeType(char *File, char *Mime, int BufLen)
{
	return LgiApp->GetFileMimeType(File, Mime, BufLen);
}

#include "GToken.h"
bool _GetIniField(char *Grp, char *Field, char *In, char *Out, int OutSize)
{
	if (ValidStr(In))
	{
		bool InGroup = false;
		GToken t(In, "\r\n");
		
		for (int i=0; i<t.Length(); i++)
		{
			char *Line = t[i];
			if (*Line == '[')
			{
				// Reset inside group at the start of each group
				InGroup = false;

				char *e = strchr(++Line, ']');
				if (e)
				{
					*e = 0;
					InGroup = stricmp(Grp, Line) == 0;
				}
			}
			else if (InGroup)
			{
				char *e = strchr(Line, '=');
				if (e)
				{
					// Point v to the data after the equals
					// Leave e pointing to the byte before
					char *v = e-- + 1;

					// Seek e through any whitespace between the equals and the
					// field name
					while (e > Line AND strchr(" \t", *e)) e--;
					
					// Seek v through any whitespace after the equals
					while (*v AND strchr(" \t", *v)) v++;
					
					// Calculate the length of the field
					int flen = (int)e-(int)Line;

					// Check the current field against the input field
					if (strnicmp(Field, Line, flen) == 0)
					{
						while (	*v AND
								strchr("\r\n", *v) == 0 AND
								OutSize > 1) // leave space for NULL
						{
							*Out++ = *v++;
							OutSize--;
						}

						// Add the NULL
						*Out++ = 0;
						
						return true;
					}
				}
			}
		}
	}
	
	return false;
}

bool _GetKdePaths(GToken &t, char *Type)
{
	bool Status = false;

	char Args[256];
	sprintf(Args, "--path %s", Type);

	GStringPipe Out;
	GProcess p;
	if (p.Run("kde-config", Args, 0, true, 0, &Out))
	{
		char *Buf = Out.NewStr();
		if (Buf)
		{
			t.Parse(Buf, ":\r\n");
			DeleteArray(Buf);
			Status = true;
		}
		else
		{
			printf("_GetKdePaths: Didn't get string back from kde-config.\n");
		}
	}
	else
	{
		printf("_GetKdePaths: Couldn't run kde-config.\n");
	}

	return Status;
}

bool _GetKdeFont(char *FontType, char *Font, int FontBufSize, int &PointSize)
{
	bool Status = false;
	if (Font AND FontType)
	{
		GToken ConfPaths;

		if (_GetKdePaths(ConfPaths, "config"))
		{
			for (int i=0; i<ConfPaths.Length() AND NOT Status; i++)
			{
				char f[256];
				LgiMakePath(f, sizeof(f), ConfPaths[i], "kdeglobals");
				if (FileExists(f))
				{
					char *s = ReadTextFile(f);
					if (s)
					{
						char Fnt[256];

						if (_GetIniField("General", FontType, s, Fnt, sizeof(Fnt)))
						{
							GToken p(Fnt, ",\r\n");
							if (p.Length() > 1)
							{
								strsafecpy(Font, p[0], FontBufSize);
								PointSize = atoi(p[1]);
								Status = true;
							}
						}

						DeleteArray(s);
					}
				}
			}
		}
	}

	return Status;
}

bool LgiGetAppsForMimeType(char *Mime, GArray<GAppInfo*> &Apps, int Limit)
{
	bool Status = false;

	if (Mime)
	{
		char Home[256];
		if (LgiGetSystemPath(LSP_HOME, Home, sizeof(Home)))
		{
			if (stricmp(Mime, "application/email") == 0)
			{
				char EmailDefs[300];
				LgiMakePath(EmailDefs, sizeof(EmailDefs), Home, ".kde/share/config/emaildefaults");
				if (FileExists(EmailDefs))
				{
					char *e = ReadTextFile(EmailDefs);
					if (e)
					{
						char App[256];
						if (_GetIniField("PROFILE_Default", "EmailClient", e, App, sizeof(App)))
						{
							GAppInfo *a = NEW(GAppInfo);
							if (a)
							{
								Apps[Apps.Length()] = a;
								a->Path = NewStr(App);
								Status = true;
							}
						}
						else
						{
							printf("%s:%i - Error, no email defs field.\n", __FILE__, __LINE__);
						}
						
						DeleteArray(e);
					}
					else
					{
						printf("%s:%i - Error, couldn't read email defs.\n", __FILE__, __LINE__);
					}
				}
				else
				{
					printf("%s:%i - Error, no email defs.\n", __FILE__, __LINE__);
				}
			}
			else
			{
				// Find non "X-" version of the MIME type.
				char NonXMime[256];
				strcpy(NonXMime, Mime);
				char *NonX = strchr(NonXMime, '/');
				if (NonX)
				{
					NonX++;
					
					int Len = strlen(NonX) + 1;
					if (strnicmp(NonX, "x-", 2) == 0)
					{
						memmove(NonX, NonX+2, Len - 2);
					}
					else
					{
						memmove(NonX+2, NonX, Len);
						NonX[0] = 'x';
						NonX[1] = '-';
					}
					NonX = NonXMime;
				}

				// Look in the profile for a default
				GToken Paths;
				_GetKdePaths(Paths, "config");
				for (int i=0; i<Paths.Length() AND NOT Status; i++)
				{
					char Buf[256];

					LgiMakePath(Buf, sizeof(Buf), Paths[i], "profilerc");
					if (FileExists(Buf))
					{
						char *Text = ReadTextFile(Buf);
						if (Text)
						{
							char Grp[128];
							sprintf(Grp, "%s - 1", Mime);

							char App[128];
							bool HasApp = _GetIniField(Grp, "Application", Text, App, sizeof(App));
							if (NOT HasApp AND NonX)
							{
								sprintf(Grp, "%s - 1", NonX);
								HasApp = _GetIniField(Grp, "Application", Text, App, sizeof(App));
							}
							if (HasApp)
							{
								GToken AppsPath;
								if (_GetKdePaths(AppsPath, "apps"))
								{
									for (int a=0; a<AppsPath.Length() AND NOT Status; a++)
									{
										LgiMakePath(Buf, sizeof(Buf), AppsPath[a], App);
										if (FileExists(Buf))
										{
											char *d = ReadTextFile(Buf);
											if (d)
											{
												char AppPath[256];
												Status = _GetIniField("Desktop Entry", "Exec", d, AppPath, 256);
												if (Status)
												{
													GAppInfo *a = NEW(GAppInfo);
													if (a)
													{
														Apps[Apps.Length()] = a;
														a->Path = NewStr(AppPath);

														if (_GetIniField("Desktop Entry", "Name", d, AppPath, 256))
														{
															a->Name = NewStr(AppPath);
														}
														if (_GetIniField("Desktop Entry", "Icon", d, AppPath, 256))
														{
															a->Icon = NewStr(AppPath);
														}
													}
												}
												DeleteArray(d);
											}
										}
									}
								}
							}
							else
							{
								// printf("LgiGetAppForMimeType no profile mime='%s'\n", Mime);
							}

							DeleteArray(Text);
						}
					}
					else
					{
						printf("LgiGetAppForMimeType: '%s' doesn't exist.\n", Buf);
					}
				}

				if (Limit < 0 OR Apps.Length() < Limit)
				{
					GArray<GAppInfo*> Lst;
					if (LgiApp->GetAppsForMimeType(Mime, Lst))
					{
						for (int i=0; i<Lst.Length() AND (Limit < 0 OR Apps.Length() < Limit); i++)
						{
							Apps[Apps.Length()] = Lst[i];
							Status = true;
						}						
					}					
				}
			}
		}
		else
		{
			printf("LgiGetAppForMimeType couldn't get home dir.\n");
		}
	}

	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath, int BufSize)
{
	bool Status = false;
	if (AppPath)
	{
		GArray<GAppInfo*> Apps;
		Status = LgiGetAppsForMimeType(Mime, Apps, 1);
		if (Status)
		{
			strsafecpy(AppPath, Apps[0]->Path, BufSize);
		}
	}
	return Status;
}

int LgiRand(int Limit)
{
	return rand() % Limit;
}

bool LgiPlaySound(char *FileName, int ASync)
{
	return LgiExecute(FileName);
}

bool LgiExecute(char *File, char *Args, char *Dir)
{
	if (File)
	{
		bool IsUrl = false;

		char App[400] = "";
		if (strnicmp(File, "http://", 7) == 0 OR
			strnicmp(File, "https://", 8) == 0)
		{
			IsUrl = true;
			LgiGetAppForMimeType("text/html", App, sizeof(App));
		}
		else
		{
			struct stat f;
			char Path[MAX_PATH];
			
			// see if the file is executable
			bool InPath = false;
			bool Ok = stat(File, &f) == 0;
			if (Ok)
			{
				strcpy(Path, File);
			}
			else
			{
				// look in the path
				InPath = true;
				GToken p(getenv("PATH"), LGI_PATH_SEPARATOR);
				for (int i=0; i<p.Length() AND NOT Ok; i++)
				{
					LgiMakePath(Path, sizeof(Path), p[i], File);
					Ok = stat(Path, &f) == 0;
				}
			}
			
			if (Ok)			
			{
				if (S_ISDIR(f.st_mode))
				{
					// open the directory in the current browser
					LgiGetAppForMimeType("inode/directory", App, sizeof(App));
				}
				/*
				else if (f.st_mode & (S_IXUSR | S_IXGRP | 1) )
				{
					// execute the file...
					char f[512];
					sprintf(f, "\"%s\" %s &", File, Args ? Args : (char*)"");
					return system(f) == 0;
				}
				*/
				else // if (NOT InPath)
				{
					// look up the type...
					char Mime[256];
					if (LgiGetFileMimeType(File, Mime, sizeof(Mime)))
					{
						printf("LgiGetFileMimeType(%s)=%s\n", File, Mime);
						
						if (stricmp(Mime, "application/x-executable") == 0 OR
							stricmp(Mime, "application/x-shellscript") == 0)
						{
							TreatAsExe:
							char f[512];
							sprintf(f, "\"%s\" %s &", File, Args ? Args : (char*)"");
							return system(f) == 0;
						}
						else
						{
							// got the mime type
							if (NOT LgiGetAppForMimeType(Mime, App, sizeof(App)))
							{
								printf("%s:%i: LgiExecute - LgiGetAppForMimeType failed to return the app for '%s'\n", __FILE__, __LINE__, File);
								goto TreatAsExe;
							}
						}
					}
					else
					{
						// printf("LgiExecute: LgiGetFileMimeType failed to return a mime type for '%s'\n", File);
						
						// If a file can't be typed it's most likely an executable.
						goto TreatAsExe;
					}
				}
			}
			else
			{
				printf("LgiExecute: '%s' doesn't exist.\n", File);
			}
		}

		if (App[0])
		{
			bool FileAdded = false;
			char EscFile[512], *o = EscFile;
			for (char i=0; File[i]; i++)
			{
				if (File[i] == ' ' OR File[i] == '&')
				{
					*o++ = '\\';
				}
				
				*o++ = File[i];
			}
			*o++ = 0;

			for (char *s=App; *s; )
			{
				if (*s == '%')
				{
					switch (s[1])
					{
						case 'f':
						case 'F':
						{
							int Len = strlen(EscFile);
							memmove(s + 2, s + Len - 2, strlen(s + 2) + 1);
							memcpy(s, EscFile, Len);
							FileAdded = true;
							break;
						}
						case 'u':
						case 'U':
						{
							char Url[256];
							sprintf(Url, IsUrl ? "%s" : "file:%s", EscFile);
							int Len = strlen(Url);

							memmove(s + 2, s + Len - 2, strlen(s + 2) + 1);
							memcpy(s, Url, Len);
							FileAdded = true;
							break;
						}
						default:
						{
							// we don't understand this command
							memmove(s, s + 2, strlen(s + 2) + 1);
							break;
						}
					}
				}
				else s++;
			}

			if (NOT FileAdded)
			{
				strcat(App, " ");
				strcat(App, EscFile);
			}

			strcat(App, " > /dev/null 2> /dev/null &");

			printf("LgiExecute: document, system('%s')\n", App);
			int e;
			if (Dir) chdir(Dir);
			if (e = system(App))
			{
				printf("\tsystem(...) failed with error: %i\n", e);
				return false;
			}

			return true;
		}
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////
WindowManager LgiGetWindowManager()
{
	static WindowManager Status = WM_Unknown;

	if (Status == WM_Unknown)
	{
		#if 0

		// Running processes seems to be slower sometimes???
		GStringPipe Out;
		GProcess p;
		if (p.Run("ps", "-A", 0, true, 0, &Out))
		{
			char *b = Out.NewStr();
			if (b)
			{
				if (stristr(b, "gnome-session"))
				{
					Status = WM_Gnome;
				}
				else if (stristr(b, "startkde"))
				{
					Status = WM_Kde;
				}
				DeleteArray(b);
			}
		}
		
		#else
		
		GDirectory *d = FileDev->GetDir();
		if (d)
		{
			for (bool b=d->First("/proc"); b AND Status == WM_Unknown; b=d->Next())
			{
				if (d->IsDir() AND isdigit(d->GetName()[0]))
				{
					char Path[256];
					d->Path(Path, sizeof(Path));
					LgiMakePath(Path, sizeof(Path), Path, "status");
					
					GFile s;
					if (s.Open(Path, O_READ))
					{
						char Buf[256];
						Buf[sizeof(Buf)-1] = 0;
						s.Read(Buf, sizeof(Buf)-1);
						
						char *n = strchr(Buf, '\n');
						if (n)
						{
							*n = 0;
							if (stristr(Buf, "gnome-session"))
							{
								Status = WM_Gnome;
							}
							else if (stristr(Buf, "startkde"))
							{
								Status = WM_Kde;
							}
						}
					}
				}
			}
			
			DeleteObj(d);
		}
		
		#endif
	}
		
	return Status;
}

