#include <stdio.h>
#include <math.h>

#include "Lgi.h"
#include "GEdit.h"
#include "GPopup.h"

///////////////////////////////////////////////////////////////////////////////////////////////
class HookInfo
{
public:
	int Flags;
	GView *Target;
};

class GWindowPrivate
{
public:
	GArray<HookInfo> Hooks;
	bool SnapToEdge;

	GWindowPrivate()
	{
		SnapToEdge = false;
	}
	
	int GetHookIndex(GView *Target, bool Create = false)
	{
		for (int i=0; i<Hooks.Length(); i++)
		{
			if (Hooks[i].Target == Target) return i;
		}
		
		if (Create)
		{
			HookInfo *n = &Hooks[Hooks.Length()];
			if (n)
			{
				n->Target = Target;
				n->Flags = 0;
				return Hooks.Length() - 1;
			}
		}
		
		return -1;
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow() : GView(0)
{
	d = NEW(GWindowPrivate);
	Menu = 0;
	SetStyle(GetStyle() | WS_TILEDWINDOW | WS_CLIPCHILDREN);
	SetStyle(GetStyle() & ~WS_CHILD);
	SetExStyle(GetExStyle() | WS_EX_CONTROLPARENT);
	
	SetClass("Lgi::GWindow");
	GWin32Class *c = GWin32Class::Create(GetClass());
	if (c)
	{
		c->Register();
	}	
	
	LastFocus = 0;

	Visible(false);

	_Default = 0;
	_Window = this;
	_Lock = NEW(GSemaphore);
	_QuitOnClose = false;
}

GWindow::~GWindow()
{
	if (Menu)
	{
		Menu->Detach();
		DeleteObj(Menu);
	}

	DeleteObj(_Lock);
	DeleteObj(d);
}

bool GWindow::GetSnapToEdge()
{
	return d->SnapToEdge;
}

void GWindow::SetSnapToEdge(bool b)
{
	d->SnapToEdge = b;
}

void GWindow::Raise()
{
	SetForegroundWindow(Handle());
}

GWindowZoom GWindow::GetZoom()
{
	if (IsZoomed(Handle()))
	{
		return GZoomMax;
	}
	else if (IsIconic(Handle()))
	{
		return GZoomMin;
	}

	return GZoomNormal;
}

void GWindow::SetZoom(GWindowZoom i)
{
	switch (i)
	{
		case GZoomMax:
		{
			ShowWindow(Handle(), SW_MAXIMIZE);
			break;
		}
		case GZoomMin:
		{
			ShowWindow(Handle(), SW_MINIMIZE);
			break;
		}
		case GZoomNormal:
		{
			if (NOT Visible())
			{
				Visible(true);
			}

			if (IsIconic(Handle()) OR IsZoomed(Handle()))
			{
				ShowWindow(Handle(), SW_RESTORE);
			}
			else
			{
				SetForegroundWindow(Handle());
			}

			LgiYield();

			RECT r;
			GetWindowRect(Handle(), &r);

			if (r.left != Pos.x1 OR
				r.top != Pos.y1)
			{
				SetWindowPos(Handle(), 0, Pos.x1, Pos.y1, Pos.X(), Pos.Y(), SWP_NOZORDER);
			}
			// Raise();
			break;
		}
	}
}

bool GWindow::OnRequestClose(bool OsShuttingDown)
{
	if (GetQuitOnClose())
	{
		LgiCloseApp();
	}

	return true;
}

bool GWindow::OnViewMouse(GView *v, GMouse &m)
{
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GMouseEvents)
		{
			if (NOT d->Hooks[i].Target->OnViewMouse(v, m))
			{
				return false;
			}
		}
	}
	
	return true;
}

bool GWindow::OnViewKey(GView *v, GKey &k)
{
	// Any window in a popup always gets the key...
	GView *p;
	for (p = v->GetParent(); p; p = p->GetParent())
	{
		if (dynamic_cast<GPopup*>(p))
		{
			return v->OnKey(k);
		}
	}

	// Allow any hooks to see the key...
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GKeyEvents)
		{
			if (d->Hooks[i].Target->OnViewKey(v, k))
			{
				return true;
			}
		}
	}

	// Give the key to the window...
	if (v->OnKey(k))
	{
		return true;
	}

	// Check default controls
	p = 0;
	if (k.c16 == VK_RETURN)
	{
		p = _Default;
	}
	else if (k.c16 == VK_ESCAPE)
	{
		p = FindControl(IDCANCEL);
	}
	if (p AND p->OnKey(k))
	{
		return true;
	}

	// Menu shortcut?
	if (Menu AND Menu->OnKey(v, k))
	{
		return true;
	}

	return false;
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

bool GWindow::Visible()
{
	return GView::Visible();
}

void GWindow::Visible(bool v)
{
	if (v) Pour();
	GView::Visible(v);
}

void GWindow::Pour()
{
	GRegion Client(GetClient());
	GRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRect OldPos = v->Pos;
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(true);
			}

			if (OldPos != v->Pos)
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->Pos);
			Update.Subtract(&v->Pos);
		}
		else
		{
			// make the view not visible
			// v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.Length(); i++)
	{
		Invalidate(Update[i]);
	}
}

int GWindow::OnEvent(GMessage *Msg)
{
	int Status = 0;

	switch (MsgCode(Msg))
	{
		case WM_SYSCOLORCHANGE:
		{
			LgiInitColours();
			break;
		}
		case WM_WINDOWPOSCHANGING:
		{
			bool Icon = IsIconic(Handle());
			bool Zoom = IsZoomed(Handle());

			if (NOT Icon AND NOT Zoom)
			{
				WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
				if (Info)
				{
					if (GetMinimumSize().x AND
						GetMinimumSize().x > Info->cx)
					{
						Info->cx = GetMinimumSize().x;
					}

					if (GetMinimumSize().y AND
						GetMinimumSize().y > Info->cy)
					{
						Info->cy = GetMinimumSize().y;
					}

					RECT Rc;
					if (d->SnapToEdge AND
						SystemParametersInfo(SPI_GETWORKAREA, 0, &Rc, SPIF_SENDCHANGE))
					{
						GRect r = Rc;
						GRect p(Info->x, Info->y, Info->x + Info->cx - 1, Info->y + Info->cy - 1);
						if (r.Valid() AND p.Valid())
						{
							int Snap = 12;

							if (abs(p.x1 - r.x1) <= Snap)
							{
								// Snap left edge
								Info->x = r.x1;
							}
							else if (abs(p.x2 - r.x2) <= Snap)
							{
								// Snap right edge
								Info->x = r.x2 - Info->cx + 1;
							}

							if (abs(p.y1 - r.y1) <= Snap)
							{
								// Snap top edge
								Info->y = r.y1;
							}
							else if (abs(p.y2 - r.y2) <= Snap)
							{
								// Snap bottom edge
								Info->y = r.y2 - Info->cy + 1;
							}
						}
					}
				}
			}
			break;
		}
		case WM_SIZE:
		{
			if (Msg->a == SIZE_MINIMIZED)
			{
				OnZoom(GZoomMin);
			}
			Status = GView::OnEvent(Msg);
			break;
		}
		case WM_ACTIVATE:
		{
			// This is a hack to make Windows set the focus of a child
			// control when you Alt-Tab in and out of a top level window
			if (LOWORD(Msg->a) != WA_INACTIVE)
			{
				// gaining focus
				if (LastFocus)
				{
					SetFocus(LastFocus);
					LastFocus = 0;
				}
			}
			else
			{
				// losing focus
				LastFocus = 0;
				HWND f = GetFocus();
				for (HWND p = ::GetParent(f); p; p = ::GetParent(p))
				{
					if (p == Handle())
					{
						LastFocus = f;
					}
				}
			}
			break;
		}
		case WM_CREATE:
		{
			#ifdef SKIN_MAGIC
			SetWindowSkin(Handle(), "MainFrame");
			#endif
			
			Pour();
			OnCreate();

			if (NOT _Default)
			{
				_Default = FindControl(IDOK);
				if (_Default)
				{
					_Default->Invalidate();
				}
			}
			break;
		}
		case WM_QUERYENDSESSION:
	 	case WM_CLOSE:
		{
			bool QuitApp = false;
			if (OnRequestClose(MsgCode(Msg) == WM_QUERYENDSESSION))
			{
				delete this;
			}

			if (MsgCode(Msg) == WM_CLOSE)
			{
				return 0;
			}
			else
			{
				return QuitApp;
			}
			break;
		}
		case WM_DROPFILES:
		{
			HDROP hDrop = (HDROP) Msg->a;
			if (hDrop)
			{
				List<char> FileNames;
				int Count = DragQueryFile(hDrop, -1, NULL, 0);;
				char FileName[256];

				for (int i=0; i<Count; i++)
				{
					if (DragQueryFile(hDrop, i, FileName, sizeof(FileName)-1) > 0)
					{
						FileNames.Insert(LgiFromNativeCp(FileName));
					}
				}

				OnReceiveFiles(FileNames);
				FileNames.DeleteArrays();
			}
			break;
		}
		case M_COMMAND:
		{
			HWND OurWnd = Handle(); // copy onto the stack, because
									// we might lose the 'this' object in the
									// OnCommand handler which would delete
									// the memory containing the handle.

			Status = OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			if (NOT IsWindow(OurWnd))
			{
				// The window was deleted so break out now
				break;
			}
			// otherwise fall thru to the GView handler
		}
		default:
		{
			Status = GView::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

void GWindow::OnPosChange()
{
	Pour();
}

bool GWindow::RegisterHook(GView *Target, int EventType, int Priority)
{
	bool Status = false;
	
	if (Target AND EventType)
	{
		int i = d->GetHookIndex(Target, true);
		if (i >= 0)
		{
			d->Hooks[i].Flags = EventType;
			Status = true;
		}
	}
	
	return Status;
}

bool GWindow::UnregisterHook(GView *Target)
{
	int i = d->GetHookIndex(Target);
	if (i >= 0)
	{
		d->Hooks.DeleteAt(i);
		return true;
	}
	return false;
}

