#include <stdio.h>
#include "Lgi.h"

/////////////////////////////////////////////////////////////////////////////
class PE_Debug
{
public :
	PE_Debug();
	~PE_Debug();

	void ClearReport();
	void DumpDebugInfo( GStringPipe& dumpBuffer, const BYTE* caller, HINSTANCE hInstance );
	void Display();

private:
	// Report data
	enum { MAX_MODULENAME_LEN = 512, MAX_FILENAME_LEN = 256 };
	char latestModule[ MAX_MODULENAME_LEN ];
	char latestFile[ MAX_FILENAME_LEN ];

	// File mapping data
	HANDLE hFile;
	HANDLE hFileMapping;
	PIMAGE_DOS_HEADER fileBase;

	// Pointers to debug information
	PIMAGE_NT_HEADERS NT_Header;
	PIMAGE_COFF_SYMBOLS_HEADER COFFDebugInfo;
	PIMAGE_SYMBOL COFFSymbolTable;
	int COFFSymbolCount;
	const char* stringTable;

	void ClearFileCache();
	void ClearDebugPtrs();
	void MapFileInMemory( const char* module );
	void FindDebugInfo();
	void DumpSymbolInfo( GStringPipe& dumpBuffer, DWORD relativeAddress );
	void DumpLineNumber( GStringPipe& dumpBuffer, DWORD relativeAddress );
	PIMAGE_COFF_SYMBOLS_HEADER GetDebugHeader();
	PIMAGE_SECTION_HEADER SectionHeaderFromName( const char* name );
	const char* GetSymbolName( PIMAGE_SYMBOL sym );
};

// Add an offset to a pointer and cast to a given type; may be
// implemented as a template function but Visual C++ has some problems.
#define BasedPtr( type, ptr, ofs ) (type)( (DWORD)(ptr) + (DWORD)(ofs) )


PE_Debug :: PE_Debug()
{
	// Init file mapping cache
	hFileMapping = 0;
	hFile = INVALID_HANDLE_VALUE;
	fileBase = 0;
	ClearDebugPtrs();
	ClearReport();
}


PE_Debug :: ~PE_Debug()
{
	ClearFileCache();
}

void PE_Debug :: ClearReport()
{
	latestModule[ 0 ] = 0;
	latestFile[ 0 ] = 0;
}

void PE_Debug :: ClearDebugPtrs()
{
	NT_Header = NULL;
	COFFDebugInfo = NULL;
	COFFSymbolTable = NULL;
	COFFSymbolCount = 0;
	stringTable = NULL;
}

void PE_Debug::ClearFileCache()
{
	if (fileBase)
	{
		UnmapViewOfFile( fileBase );
		fileBase = 0;
	}
	
	if (hFileMapping != 0)
	{
		CloseHandle( hFileMapping );
		hFileMapping = 0;
	}
	
	if (hFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle( hFile );
		hFile = INVALID_HANDLE_VALUE;
	}
}

void PE_Debug :: DumpLineNumber( GStringPipe& dumpBuffer, DWORD relativeAddress )
{
	PIMAGE_LINENUMBER line = BasedPtr(	PIMAGE_LINENUMBER,
										COFFDebugInfo,
										COFFDebugInfo->LvaToFirstLinenumber);
	DWORD lineCount = COFFDebugInfo->NumberOfLinenumbers;
	const DWORD none = (DWORD)-1;
	DWORD maxAddr = 0;
	DWORD lineNum = none;
	
	for( DWORD i=0; i < lineCount; i++ )
	{
		if( line->Linenumber != 0 )  // A regular line number
		{
			// look for line with bigger address <= relativeAddress
			if( line->Type.VirtualAddress <= relativeAddress &&
				line->Type.VirtualAddress > maxAddr )
			{
				maxAddr = line->Type.VirtualAddress;
				lineNum = line->Linenumber;
			}
		}
		line++;
	}

	if( lineNum != none )
		dumpBuffer.Printf( "  line: %d\r\n", lineNum );
	else
		dumpBuffer.Printf( "  line: <unknown>\r\n" );
}


const char* PE_Debug :: GetSymbolName( PIMAGE_SYMBOL sym )
{
	const int NAME_MAX_LEN = 64;
	static char buf[ NAME_MAX_LEN ] = { 0 };
	if( sym->N.Name.Short != 0 )
	{
		strncpy( buf, (const char*)sym->N.ShortName, 8 );
		buf[ 8 ] = 0;
	}
	else
	{
		strncpy( buf, stringTable + sym->N.Name.Long, NAME_MAX_LEN );
		buf[ NAME_MAX_LEN - 1 ] = 0;
	}
	return( buf );
}


void PE_Debug :: DumpSymbolInfo( GStringPipe& dumpBuffer, DWORD relativeAddress )
{
	// Variables to keep track of function symbols
	PIMAGE_SYMBOL currentSym = COFFSymbolTable;
	PIMAGE_SYMBOL fnSymbol = NULL;
	DWORD maxFnAddress = 0;

	// Variables to keep track of file symbols
	PIMAGE_SYMBOL fileSymbol = NULL;
	PIMAGE_SYMBOL latestFileSymbol = NULL;
	for( int i = 0; i < COFFSymbolCount; i++ )
	{
		// Look for .text section where relativeAddress belongs to.
		// Keep track of the filename the .text section belongs to.
		if( currentSym->StorageClass == IMAGE_SYM_CLASS_FILE )
			latestFileSymbol = currentSym;
		
		// Borland uses "CODE" instead of the standard ".text" entry
		// Microsoft uses sections that only _begin_ with .text
		const char* symName = GetSymbolName( currentSym );
		if( strnicmp( symName, ".text", 5 ) == 0 || 
			strcmpi( symName, "CODE" ) == 0 )
		if( currentSym->Value <= relativeAddress )
		{
			PIMAGE_AUX_SYMBOL auxSym = (PIMAGE_AUX_SYMBOL)(currentSym + 1);
			if( currentSym->Value + auxSym->Section.Length >= relativeAddress )
				fileSymbol = latestFileSymbol;
		}
	
		// Look for the function with biggest address <= relativeAddress
		BOOL isFunction = ISFCN( currentSym->Type ); // Type == 0x20, See WINNT.H
		if( isFunction && 
			( currentSym->StorageClass == IMAGE_SYM_CLASS_EXTERNAL ||
			currentSym->StorageClass == IMAGE_SYM_CLASS_STATIC ) )
		{
			if( currentSym->Value <= relativeAddress &&
				currentSym->Value > maxFnAddress )
			{
				maxFnAddress = currentSym->Value;
				fnSymbol = currentSym;
			}
		}
		
		// Advance counters, skip aux symbols
		i += currentSym->NumberOfAuxSymbols;
		currentSym += currentSym->NumberOfAuxSymbols;
		currentSym++;
	}

	// dump symbolic info if found
	if( fileSymbol )
	{
		const char* auxSym = (const char*)(fileSymbol + 1);
		if( strcmpi( latestFile, auxSym ) )
		{
			strcpy( latestFile, auxSym );
			dumpBuffer.Printf( "  file: %s\r\n", auxSym );    
		}
	}
	else
	{
		latestFile[ 0 ] = 0;
		dumpBuffer.Printf( "  file: unknown\r\n" );    
	}
	
	if( fnSymbol )
		dumpBuffer.Printf( "    function: %s", GetSymbolName( fnSymbol ) );
	else
		dumpBuffer.Printf( "    function: <unknown>" );
}

PIMAGE_SECTION_HEADER PE_Debug::SectionHeaderFromName( const char* name )
{
	LgiTrace("\tSectionHeaderFromName('%s')", name);

	PIMAGE_SECTION_HEADER section = IMAGE_FIRST_SECTION( NT_Header );
	for( unsigned i = 0; i < NT_Header->FileHeader.NumberOfSections; i++ )
	{
		LgiTrace("\tname[%i]='%s'", i, (const char*)section->Name);
		if( strnicmp( (const char*)section->Name, name, IMAGE_SIZEOF_SHORT_NAME ) == 0 )
			return( section );
		else
			section++;
	}

	return 0;
}

PIMAGE_COFF_SYMBOLS_HEADER PE_Debug::GetDebugHeader()
{
	// Some files have a wrong entry in the COFF header, so
	// first check if the debug info exists at all

	/*
	if( NT_Header->FileHeader.PointerToSymbolTable == 0 )
	{
		LgiTrace("No symbol table");
		return( 0 );
	}
	*/

	DWORD debugDirRVA = NT_Header->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress;
	if( debugDirRVA == 0 )
	{
		LgiTrace("No Virtual Address");
		return( 0 );
	}

	// The following values must be calculated differently for MS/Borland files
	PIMAGE_DEBUG_DIRECTORY debugDir;
	DWORD size;

	// Borland files have the debug directory at the beginning of a .debug section
	PIMAGE_SECTION_HEADER debugHeader = SectionHeaderFromName( ".debug" );
	if( debugHeader && debugHeader->VirtualAddress == debugDirRVA )
	{
		debugDir = (PIMAGE_DEBUG_DIRECTORY)(debugHeader->PointerToRawData + (DWORD)fileBase);
		size = NT_Header->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size *
				sizeof( IMAGE_DEBUG_DIRECTORY );
	}
	else
	// Microsoft debug directory is in the .rdata section
	{
		debugHeader = SectionHeaderFromName( ".rdata" );
		if( debugHeader == 0 )
		{
			LgiTrace("No .rdata section");
			return( 0 );
		}
		
		size = NT_Header->OptionalHeader.
		DataDirectory[ IMAGE_DIRECTORY_ENTRY_DEBUG ].Size;
		DWORD offsetInto_rdata = debugDirRVA - debugHeader->VirtualAddress;
		debugDir = BasedPtr(PIMAGE_DEBUG_DIRECTORY, fileBase, debugHeader->PointerToRawData + offsetInto_rdata);
	}

	// look for CodeView debug info
	DWORD debugFormats = size / sizeof( IMAGE_DEBUG_DIRECTORY );
	LgiTrace("Debug formats = %i (Debug type=%i)", debugFormats, IMAGE_DEBUG_TYPE_COFF);
	for( DWORD i = 0; i < debugFormats; i++ )
	{
		LgiTrace("Format[%i]=%i (%08.8X)", i, debugDir->Type, debugDir->Type);
		if( debugDir->Type == IMAGE_DEBUG_TYPE_COFF )
		{
			LgiTrace("Debug info found..");
			return( (PIMAGE_COFF_SYMBOLS_HEADER)((DWORD)fileBase + debugDir->PointerToRawData) );
		}
		else
			debugDir++;
	}
	
	return( NULL );
}


void PE_Debug::FindDebugInfo()
{
	ClearDebugPtrs();

	// Put everything into a try/catch in case the file has wrong fields
	try
	{
		// Verify that fileBase is a valid pointer to a DOS header
		if( fileBase->e_magic == IMAGE_DOS_SIGNATURE )
		{
			// Get a pointer to the PE header
			NT_Header = BasedPtr( PIMAGE_NT_HEADERS, fileBase, fileBase->e_lfanew );

LgiTrace("::FindDebugInfo() 0 NT_Header=%p", NT_Header);

			
			// Verify that NT_Header is a valid pointer to a NT header
			if( NT_Header->Signature == IMAGE_NT_SIGNATURE )
			{
				// Get a pointer to the debug header if any
				COFFDebugInfo = GetDebugHeader();

LgiTrace("::FindDebugInfo() 1 COFFDebugInfo=%p", COFFDebugInfo);
				
				// Get a pointer to the symbol table and retrieve the number of symbols
				if (NT_Header->FileHeader.PointerToSymbolTable)
					COFFSymbolTable = BasedPtr( PIMAGE_SYMBOL, fileBase, NT_Header->FileHeader.PointerToSymbolTable );

LgiTrace("::FindDebugInfo() 2 COFFSymbolTable=%p", COFFSymbolTable);

				COFFSymbolCount = NT_Header->FileHeader.NumberOfSymbols;

LgiTrace("::FindDebugInfo() 3 COFFSymbolCount=%i", COFFSymbolCount);
					
				// The string table starts right after the symbol table
				stringTable = (const char*)(COFFSymbolTable + COFFSymbolCount); 

LgiTrace("::FindDebugInfo() 4 stringTable=%p", stringTable);
			}
		}
	}
	catch( ... )
	{
		// Header wrong, do nothing
		LgiTrace("Exeception: PE_Debug::FindDebugInfo()");
	}
}

void PE_Debug :: MapFileInMemory( const char* module )
{
	ClearFileCache();
	hFile = CreateFile( module,
						GENERIC_READ,
						FILE_SHARE_READ,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						0);

	if( hFile != INVALID_HANDLE_VALUE )
	{
		hFileMapping = CreateFileMapping( hFile, NULL, PAGE_READONLY, 0, 0, NULL );
		if( hFileMapping != 0 )
			fileBase = (PIMAGE_DOS_HEADER)MapViewOfFile( hFileMapping, FILE_MAP_READ, 0, 0, 0 );
	}
	// NB: open files/mapping are closed later in ClearFileCache
}

void PE_Debug :: DumpDebugInfo( GStringPipe& dumpBuffer, const BYTE* caller, HINSTANCE hInstance )
{
	// Avoid to open, map and looking for debug header/symbol table
	// by caching the latest and comparing the actual module with
	// the latest one.
	static char module[ MAX_MODULENAME_LEN ] = { 0 };
	GetModuleFileName( hInstance, module, MAX_MODULENAME_LEN );
	if( strcmpi( latestModule, module ) )  // New module
	{
		strcpy( latestModule, module );
		dumpBuffer.Printf( "Module: %s\r\n", module );
		MapFileInMemory( module );
		FindDebugInfo();
	}

	if( fileBase )
	{
		// Put everything into a try/catch in case the file has wrong fields
		try
		{
			DWORD relativeAddress = caller - (BYTE*)hInstance;
			
			// Dump symbolic information and line number if available
			if( COFFSymbolCount != 0 && COFFSymbolTable != NULL )
			{
				DumpSymbolInfo( dumpBuffer, relativeAddress );
				if( COFFDebugInfo )
					DumpLineNumber( dumpBuffer, relativeAddress );
			}
			else
			{
				dumpBuffer.Printf( "  no debug information\r\n" );
				dumpBuffer.Printf( "    address: %8X\r\n", caller );
			}
		}
		catch( ... )
		{
			// Header wrong, do nothing
		}
	}
	else
	{
		dumpBuffer.Printf( "  module not accessible\r\n" );
		dumpBuffer.Printf( "    address: %8X\r\n", caller );
	}
}

//////////////////////////////////////////////////////////////////////////
void StackDump(GStringPipe &dumpBuffer, DWORD EBP, DWORD EIP )
{
	// The structure of the stack frames is the following:
	// EBP -> parent stack frame EBP
	//        return address for this call ( = caller )
	// The chain can be navigated iteratively; the
	// initial value of EBP is given by the parameter.
	// If EIP != 0, then before we dump the call stack, we
	// dump the same information for EIP. Basically, we
	// consider EIP as the first element of the call stack.
	static PE_Debug PE_debug;
	BOOL EBP_OK = TRUE;
	do
	{
		// Check if EBP is a good address
		// I'm expecting a standard stack frame, so
		// EPB must be aligned on a DWORD address
		// and it should be possible to read 8 bytes
		// starting at it (next EBP, caller).
		if( (DWORD)EBP & 3 )
			EBP_OK = FALSE;

		if( EBP_OK && IsBadReadPtr( (void*)EBP, 8 ) )
			EBP_OK = FALSE;

		if( EBP_OK )
		{
			BYTE* caller = EIP ? (BYTE*)EIP : *((BYTE**)EBP + 1);
			EBP = EIP ? EBP : *(DWORD*)EBP;
			
			if( EIP )
				EIP = 0; // So it is considered just once
		
			// Get the instance handle of the module where caller belongs to
			MEMORY_BASIC_INFORMATION mbi;
			VirtualQuery( caller, &mbi, sizeof( mbi ) );
			
			// The instance handle is equal to the allocation base in Win32
			HINSTANCE hInstance = (HINSTANCE)mbi.AllocationBase;
			
			// If EBP is valid, hInstance is not 0
			if( hInstance == 0 )
				EBP_OK = FALSE;
			else
				PE_debug.DumpDebugInfo( dumpBuffer, caller, hInstance );
		}
		else
			break; // End of the call chain
	}
	while( TRUE );
	
	PE_debug.ClearReport();  // Prepare for future calls
}

void DumpExceptCallsStack(GStringPipe &dumpBuffer, DWORD EBP, DWORD EIP)
{
	char* separator = "------------------------------------------------------------------\r\n";
	dumpBuffer.Printf("\r\nCall stack:\r\n");
	dumpBuffer.Printf(separator);
	StackDump(dumpBuffer, EBP, EIP);
	dumpBuffer.Printf(separator);
}


#define CODE_DESCR( c ) { c, #c }

static 
struct ExceptionCodeDescr
{
	DWORD code;
	const char* descr;
}
exceptionTable[] =
{
	CODE_DESCR(EXCEPTION_ACCESS_VIOLATION),
	CODE_DESCR(EXCEPTION_ARRAY_BOUNDS_EXCEEDED),
	CODE_DESCR(EXCEPTION_BREAKPOINT),
	CODE_DESCR(EXCEPTION_DATATYPE_MISALIGNMENT),
	CODE_DESCR(EXCEPTION_FLT_DENORMAL_OPERAND),
	CODE_DESCR(EXCEPTION_FLT_DIVIDE_BY_ZERO),
	CODE_DESCR(EXCEPTION_FLT_INEXACT_RESULT),
	CODE_DESCR(EXCEPTION_FLT_INVALID_OPERATION),
	CODE_DESCR(EXCEPTION_FLT_OVERFLOW),
	CODE_DESCR(EXCEPTION_FLT_STACK_CHECK),
	CODE_DESCR(EXCEPTION_FLT_UNDERFLOW),
	CODE_DESCR(EXCEPTION_GUARD_PAGE),
	CODE_DESCR(EXCEPTION_ILLEGAL_INSTRUCTION),
	CODE_DESCR(EXCEPTION_IN_PAGE_ERROR),
	CODE_DESCR(EXCEPTION_INT_DIVIDE_BY_ZERO),
	CODE_DESCR(EXCEPTION_INT_OVERFLOW),
	CODE_DESCR(EXCEPTION_INVALID_DISPOSITION),
	CODE_DESCR(EXCEPTION_NONCONTINUABLE_EXCEPTION),
	CODE_DESCR(EXCEPTION_PRIV_INSTRUCTION),
	CODE_DESCR(EXCEPTION_SINGLE_STEP),
	CODE_DESCR(EXCEPTION_STACK_OVERFLOW)
	// might be extended with compiler-specific knowledge of
	// "C++ Exception" code[s]
};

void DumpExceptionDescr(GStringPipe &dumpBuffer, DWORD code)
{
	const char* descr = NULL;
	int n = sizeof( exceptionTable ) / sizeof( ExceptionCodeDescr );

	for (int i = 0; i < n; i++)
	{
		if (exceptionTable[i].code == code)
		{
			descr = exceptionTable[i].descr;
			break;
		}
	}
	
	if (descr)
	{
		dumpBuffer.Printf("  Description: %s\r\n", descr);
	}
	else
	{
		dumpBuffer.Printf("  Description unavailable\r\n");
	}
}

LONG __stdcall GApp::_ExceptionFilter(LPEXCEPTION_POINTERS e)
{
	EnterCriticalSection(&StackTraceSync);

	LONG res = 0;
	DWORD exceptionCode = e->ExceptionRecord->ExceptionCode;

	Buf.Empty();
	Buf.Printf("Unhandled Exception\r\n  Code: %x\r\n", exceptionCode);

	DumpExceptionDescr(Buf, exceptionCode);
	DumpExceptCallsStack(Buf, e->ContextRecord->Ebp, e->ContextRecord->Eip);

	char *Msg = Buf.NewStr();
	if (Msg)
	{
		int Result = LgiMsg(AppWnd, Msg, "Error: Exception", MB_YESNOCANCEL);
		switch (Result)
		{
			case IDYES: // Abort
				res = EXCEPTION_EXECUTE_HANDLER;
				break;
			case IDNO: // Debug
				res = EXCEPTION_CONTINUE_SEARCH;
				break;
			default:
				// Not reached, cannot use assert!
				MessageBeep(0);
		}
	}

	LeaveCriticalSection(&StackTraceSync);

	return res;
}

